#include "Env.Cell.h"
#include "Env.World.h"
#include "Env.Settings.h"

#include "Engine.OpenGL.h"

namespace app {
  namespace env {

    const IString CellNGInput_s[ _cngiCount ] = {
      "Health over 320",
      "Health over 160",
      "Health over 80",
      "Health over 40",
      "Health over 20",
      "Health over 10",

      "Energy over 320",
      "Energy over 160",
      "Energy over 80",
      "Energy over 40",
      "Energy over 20",
      "Energy over 10",

      "Cell on left",
      "Cell on top",
      "Cell on right",
      "Cell on bottom",

      "Food on left",
      "Food on top",
      "Food on right",
      "Food on bottom",

      "Wall on left",
      "Wall on top",
      "Wall on right",
      "Wall on bottom",

      "Something on left (2 dist)",
      "Something on top (2 dist)",
      "Something on right (2 dist)",
      "Something on bottom (2 dist)",

      "Tick",
      "Always on"
    };
    const Vec2World CellNGOutputDir_vec[ 4 ] = {
      Vec2World( -1, 0 ),
      Vec2World( 0, -1 ),
      Vec2World( 1, 0 ),
      Vec2World( 0, 1 )
    };
    const IString CellNGOutput_s[ _cngoCount ] = {
      "Direction #1",
      "Direction #2",

      "Action #1",
      "Action #2",
      "Hold"
      /*"Move left",
      "Move top",
      "Move right",
      "Move bottom",

      "Reproduce left",
      "Reproduce top",
      "Reproduce right",
      "Reproduce bottom",

      "Sex. rep. left",
      "Sex. rep. top",
      "Sex. rep. right",
      "Sex. rep. bottom",*/
    };

    // Cell
    void Cell::create( Vec2World IPos ) {
      Health = World->EnvSettings.CellHealth;
      Energy = World->EnvSettings.StartingEnergy;
      PrevPos = IPos;
      Pos = IPos;
      Destroyed = false;
      Sync = false;

      WorldObject &Obj = World->get( Pos );
      
      ASSERT_THROW( Obj.Type != woNone, "createCellToNonEmptyPosition", "pos", String().pushInt( Pos.X ).pushChar( ':' ).pushInt( Pos.Y ).istr() );

      Obj.Type = woCell;
      Obj.Pos = PosInContainer;
      
      distanceTraveled = 0;
      cellsEaten = 0;
      foodEaten = 0;
      reproducedTimes = 0;
      sexReproducedTimes = 0;
      generation = 1;
    }
    void Cell::destroy() {
      if( Destroyed ) return;

      Health = 0;
      Destroyed = true;
      World->get( Pos ).Type = woNone;

      World->CellEmptyList.pushItem( PosInContainer );
    }

    void Cell::step() {
      if( World->Sync == Sync )
        return;
      else
        Sync = World->Sync;

      PrevPos = Pos;

      inputCheck();
      Brain.step();

      /*rangeCheck( cdLeft,   Vec2World( Pos.X - 2, Pos.Y ) );
      rangeCheck( cdRight,  Vec2World( Pos.X + 2, Pos.Y ) );
      rangeCheck( cdUp,     Vec2World( Pos.X , Pos.Y - 2 ) );
      rangeCheck( cdDown,   Vec2World( Pos.X, Pos.Y + 2 ) );*/

      unsigned char dir = Brain.output( cngoDir1 ) * 2 + Brain.output( cngoDir2 );
      unsigned char actiont = Brain.output( cngoAction1 ) + Brain.output( cngoAction2 ) * 2 + Brain.output( cngoAction3 ) * 4;
      Vec2World dirVec = CellNGOutputDir_vec[ dir ];

      action( dir, actiont );

      if( --Energy <= 0 )
        Health = 0;

      if( --Health <= 0 )
        destroy();
    }

    void Cell::inputCheck() {
      generalCheck();

      directionalCheck( cdLeft );
      directionalCheck( cdRight );
      directionalCheck( cdUp );
      directionalCheck( cdDown );
    }
    void Cell::generalCheck() {
      Brain.input( cngiHealthOver320 ) = ( Health > 320 ) ? 1 : 0;
      Brain.input( cngiHealthOver160 ) = ( Health > 160 ) ? 1 : 0;
      Brain.input( cngiHealthOver80 ) = ( Health > 80 ) ? 1 : 0;
      Brain.input( cngiHealthOver40 ) = ( Health > 40 ) ? 1 : 0;
      Brain.input( cngiHealthOver20 ) = ( Health > 20 ) ? 1 : 0;
      Brain.input( cngiHealthOver10 ) = ( Health > 10 ) ? 1 : 0;

      Brain.input( cngiEnergyOver320 ) = ( Energy > 320 ) ? 1 : 0;
      Brain.input( cngiEnergyOver160 ) = ( Energy > 160 ) ? 1 : 0;
      Brain.input( cngiEnergyOver80 ) = ( Energy > 80 ) ? 1 : 0;
      Brain.input( cngiEnergyOver40 ) = ( Energy > 40 ) ? 1 : 0;
      Brain.input( cngiEnergyOver20 ) = ( Energy > 20 ) ? 1 : 0;
      Brain.input( cngiEnergyOver10 ) = ( Energy > 10 ) ? 1 : 0;

      Brain.input( cngiTick ) = ( World->Sync ) ? 1 : 0;
      Brain.input( cngiAlwaysOn ) = 1;
    }
    void Cell::directionalCheck( NG_t Dir ) {
      Vec2World TPos = Pos + CellNGOutputDir_vec[ Dir ];
      auto Obj = World->get( TPos ).Type;

      Brain.input( cngiCellOnLeft + Dir ) = ( Obj == woCell ) ? 1 : 0;
      Brain.input( cngiFoodOnLeft + Dir ) = ( Obj == woFood ) ? 1 : 0;
      Brain.input( cngiWallOnLeft + Dir ) = ( Obj == woWall ) ? 1 : 0;

      // Range check
      TPos = Pos + CellNGOutputDir_vec[ Dir ] * 2;
      Obj = World->get( TPos ).Type;

      Brain.input( cngiSomethingFarLeft + Dir ) = ( Obj != woNone ) ? 1 : 0;
    }
    void Cell::action( NG_t Dir, unsigned char action ) {
      env::Settings &Settings = World->EnvSettings;

      Vec2World TPos = Pos + CellNGOutputDir_vec[ Dir ];
      WorldObject &ObjT = World->get( TPos );
      auto Obj = ObjT.Type;

      // Sexual reproduction
      if( action == cngoaSex ) {
        if( Obj != woCell || Energy < Settings.EnergyForSex ) {
          Energy -= Settings.EnergyWrongAction;
          return;
        }

        Vec2World TPos2 = Pos * 2 - TPos;
        WorldObject_e Obj2 = World->get( TPos2 ).Type;

        if( Obj2 == woNone ) {
          Energy -= Settings.EnergyForSex;
          Cell &C = World->createCell( TPos2 );
          C.Brain.mutate( Brain, World->CellList[ ObjT.Pos ].Brain );
          C.generation = generation + 1;

          sexReproducedTimes ++;
        }

        return;
      }

      // Reproduction
      if( action == cngoaReproduce ) {
        if( Obj != woNone || Energy < Settings.EnergyForReproduction ) {
          Energy -= Settings.EnergyWrongAction;
          return;
        }

        Energy -= Settings.EnergyForReproduction;
        Cell &C = World->createCell( TPos );
        C.Brain.mutate( Brain, World->Generator.roll() );

        C.generation = generation + 1;
        reproducedTimes ++;

        return;
      }

      // Movement
      if( action == cngoaMove ) {
        Energy -= Settings.EnergyForMovement;

        if( Obj == woNone && Energy > 0 ) {
          move( TPos );

        } else if( Obj == woFood ) {
          Energy += World->FoodList[ ObjT.Pos ].Energy;

          if( Energy >= Settings.EnergyOverfeed )
            Health = 0;

          World->FoodList[ ObjT.Pos ].destroy();

          foodEaten ++;
          move( Pos );
        }

        distanceTraveled ++;

        return;
      }

      // Kill
      if( action == cngoaKill ) {
        if( Obj != woCell ) {
          Energy -= Settings.EnergyWrongAction;
          return;
        }

        Cell &c = World->CellList[ ObjT.Pos ];
        bool won = ( World->Generator.roll( ( Energy + Health * 2 ) * 3 + c.Energy + c.Health * 2 ) ) < ( Energy + Health*2 ) * 3;
        if( won ) {
          Energy += c.Energy;
          Health += c.Health / 2;
          World->CellList[ ObjT.Pos ].destroy();
          if( Energy >= Settings.EnergyOverfeed )
            Health = 0;

          cellsEaten ++;

        } else {
          c.Energy = mMin< CellHealth_t >( c.Energy + Energy / 2, Settings.EnergyOverfeed );
          c.Health += Health / 4;
          Health = 0;
        }

        return;
      }
    }

    bool Cell::move( Vec2World TargetPos ) {
      WorldObject& Tar = World->get( TargetPos );

      if( Tar.Type != woNone )
        return false;

      Tar.Type = woCell;
      Tar.Pos = PosInContainer;

      World->get( Pos ).Type = woNone;
      Pos = TargetPos;

      return true;
    }

    // CellDrawData
    CellDrawData::CellDrawData( Cell &Source ) {
      Pos = Source.Pos;
      PrevPos = Source.PrevPos;
      Energy = ( UInt8 ) ( Source.Energy / mMax( 1, ( World->EnvSettings.EnergyOverfeed / 8 ) ) );
      Health = ( UInt8 ) mLimit( ( Source.Health * 127 ) / World->EnvSettings.CellHealth, 0, 127 );
      cellsEaten = Source.cellsEaten;
      posInCont = Source.PosInContainer;
    }

  }
}