#pragma once

#include "Engine.Stream.h"

#include "Env.NeuralGrid.h"
#include "Env.Types.h"

namespace app {
  namespace env {

    enum CellDirection_e {
      cdLeft,
      cdUp,
      cdRight,
      cdDown
    };

    enum CellNGInput_e {
      cngiHealthOver320,
      cngiHealthOver160,
      cngiHealthOver80,
      cngiHealthOver40,
      cngiHealthOver20,
      cngiHealthOver10,

      cngiEnergyOver320,
      cngiEnergyOver160,
      cngiEnergyOver80,
      cngiEnergyOver40,
      cngiEnergyOver20,
      cngiEnergyOver10,

      cngiCellOnLeft,
      cngiCellOnUp,
      cngiCellOnRight,
      cnhiCellOnDown,

      cngiFoodOnLeft,
      cngiFoodOnUp,
      cngiFoodOnRight,
      cngiFoodOnDown,

      cngiWallOnLeft,
      cngiWallOnUp,
      cngiWallOnRight,
      cngiWallOnDown,

      cngiSomethingFarLeft,
      cngiSomethingFarUp,
      cngiSomethingFarRight,
      cngiSomethingFarDown,

      cngiTick,
      cngiAlwaysOn,

      _cngiCount
    };
    extern const IString CellNGInput_s[ _cngiCount ];

    enum CellNGOutput_e {
      /*cngoMoveLeft,
      cngoMoveUp,
      cngoMoveRight,
      cngoMoveDown,

      cngoReproduceLeft,
      cngoReproduceUp,
      cngoReproduceRight,
      cngoReproduceDown,

      cngoSexLeft,
      cngoSexUp,
      cngoSexRight,
      cngoSexDown,*/

      cngoDir1,
      cngoDir2,

      cngoAction1,
      cngoAction2,
      cngoAction3,

      _cngoCount
    };
    enum CellNGOutputAction_e {
      cngoaMove,
      cngoaReproduce,
      cngoaKill,
      cngoaSex,
    };
    extern const Vec2World CellNGOutputDir_vec[ 4 ];
    extern const IString CellNGOutput_s[ _cngoCount ];

    class Cell {

    public:
      CellHealth_t Energy, Health;
      NeuralGrid Brain;
      Vec2World PrevPos, Pos;
      WorldPos_t PosInContainer;
      bool Destroyed;

      bool Sync; // For update sync

    public:
      SizeT distanceTraveled, cellsEaten, foodEaten, reproducedTimes, sexReproducedTimes, generation;

    public:
      void create( Vec2World Pos );
      void destroy();

    public:
      void step();   

    public:
      void directionalCheck( NG_t Dir );
      void generalCheck();
      void inputCheck();

      void action( NG_t dir, unsigned char action );

      bool move( Vec2World TargetPos );

    };

    struct CellDrawData {
      Vec2World PrevPos, Pos;
      UInt8 Health, Energy;
      SizeT cellsEaten, posInCont;

      CellDrawData( Cell &Source );
      CellDrawData() {};
    };

  }
}