#include "Env.NeuralGrid.h"
#include "Env.World.h"
#include "Env.Settings.h"
#include "Env.Cell.h"

namespace app {
  namespace env {

    NG_value_t ActivationFunction( NG_treshold_t In ) {
      return ( In > 0 ) ? 1 : 0;
    }

    // NeuralGrid
    NeuralGrid::NeuralGrid() {

    }
    NeuralGrid::~NeuralGrid() {

    }

    NG_value_t& NeuralGrid::input( NG_t Pos ) {
      return Grid[ 0 ][ Pos ].Value;
    }
    NG_value_t& NeuralGrid::output( NG_t Pos ) {
      return Grid[ NG_depth - 1 ][ Pos ].Value;
    }

    void NeuralGrid::step() {

      NG_treshold_t Sum;
      NG_t dmo = 0; // depth minus one
       NeuralGridCell *C;

      // Computation
      for( NG_t depth = 1; depth < NG_depth; depth ++ ) {
        for( NG_t cell = 0; cell < NG_width; cell ++ ) {

          C = &Grid[ depth ][ cell ];

          Sum = 0;
          for( NG_t a = 0; a < NG_width; a ++ )
            Sum += ( NG_treshold_t ) Grid[ dmo ][ a ].Value * ( NG_treshold_t ) C->Weight[ a ];
          
          Sum -= C->Treshold;
          C->Value = ActivationFunction( Sum );

        }     
        dmo ++;
      }

      // Input reset
      /*for( NG_t a = 0; a < NG_width; a ++ )
        Grid[ 0 ][ a ].Value = 0;*/

      
      // Cell loopback
      for( NG_t a = _cngiCount; a < NG_width; a ++ )
        Grid[ 0 ][ a ].Value = Grid[ NG_depth - 1 ][ a ].Value;

    }

    void NeuralGrid::mutate( const NeuralGrid &Source, RandGen_t Seed ) {
      RandGen Rand( Seed );

      auto WMCh = World->EnvSettings.WeightMutationChance;
      auto TMCh = World->EnvSettings.TresholdMutationChance;
      double mutDepthLevMult = ( double ) World->EnvSettings.mutationLevelMultiplier / 100.0;

      memcpy( this, &Source, sizeof( NeuralGrid ) );

      for( NG_t depth = 1; depth < NG_depth; depth ++ ) {
        for( NG_t cell = 0; cell < NG_width; cell ++ ) {

          NeuralGridCell &C = Grid[ depth ][ cell ];

          if( Rand.roll< SizeT >( mMax< SizeT >( 1, 1000 * pow( mutDepthLevMult, NG_depth - depth - 1 ) ) ) < TMCh ) {
            if( Rand.roll( 100 ) < 50 && C.Treshold < NG_treshold_limit )
              C.Treshold += 1;
            else if( C.Treshold > -NG_treshold_limit )
              C.Treshold -= 1;
          }

          for( NG_t a = 0; a < NG_width; a ++ ) {
            if( Rand.roll< SizeT >( mMax< SizeT >( 1, 1000 * pow( mutDepthLevMult, NG_depth - depth - 1 ) ) ) < WMCh ) {
              if( Rand.roll( 100 ) < 50 && C.Weight[ a ] < NG_weight_limit )
                C.Weight[ a ] += 1;
              else if( C.Weight[ a ] > -NG_weight_limit )
                C.Weight[ a ] -= 1;
            }
          }
        }
      }
    }
    void NeuralGrid::mutate( const NeuralGrid &Source, const NeuralGrid &Source2 ) {
      for( NG_t depth = 1; depth < NG_depth; depth ++ ) {
        for( NG_t cell = 0; cell < NG_width; cell ++ ) {

          NeuralGridCell &C = Grid[ depth ][ cell ];
          const NeuralGridCell &SC = Source.Grid[ depth ][ cell ];
          const NeuralGridCell &SC2 = Source2.Grid[ depth ][ cell ];

          C.Treshold = SC.Treshold / 2 + SC2.Treshold / 2;

          for( NG_t a = 0; a < NG_width; a ++ )
            C.Weight[ a ] = SC.Weight[ a ] / 2 + SC2.Weight[ a ] / 2;
        }
      }
    }

    const String DefaultBrainFunctionListNames[ DefaultBrainFunctionListCount ] = {
      "One layer",
      "Random",
      "Custom fixed preset",
    };
    void ( *DefaultBrainFunctionList[ DefaultBrainFunctionListCount ] )( NeuralGrid &Target, RandGen_t Seed ) = {
      // One layer
      [] ( NeuralGrid &Target, RandGen_t Seed ) {
        RandGen Rand( Seed );

        for( NG_t a = 0; a < NG_width; a ++ )
          Target.Grid[ 0 ][ a ].Value = 0;

        for( NG_t depth = 1; depth < NG_depth - 1; depth ++ ) {
          for( NG_t cell = 0; cell < NG_width; cell ++ ) {

            NeuralGridCell &C = Target.Grid[ depth ][ cell ];

            C.Treshold = 9;

            for( NG_t a = 0; a < NG_width; a ++ )
              C.Weight[ a ] = 0; 

            C.Weight[ cell ] = 10;
          }
        }

        for( NG_t depth = NG_depth - 1; depth < NG_depth; depth ++ ) {
          for( NG_t cell = 0; cell < NG_width; cell ++ ) {

            NeuralGridCell &C = Target.Grid[ depth ][ cell ];
            C.Treshold = Rand.roll( NG_width );

            for( NG_t a = 0; a < NG_width; a ++ )
              C.Weight[ a ] = ( Rand.roll( 100 ) > 95 ) ? 10 : 0;
          }
        }
      },
      // Complete random
      [] ( NeuralGrid &Target, RandGen_t Seed ) {
        RandGen Rand( Seed );

        for( NG_t a = 0; a < NG_width; a ++ )
          Target.Grid[ 0 ][ a ].Value = 0;

        for( NG_t depth = 0; depth < NG_depth; depth ++ ) {
          for( NG_t cell = 0; cell < NG_width; cell ++ ) {

            NeuralGridCell &C = Target.Grid[ depth ][ cell ];

            C.Treshold = Rand.roll( NG_treshold_limit );

            for( NG_t a = 0; a < NG_width; a ++ )
              C.Weight[ a ] = Rand.roll( NG_weight_limit ); 
          }
        }
      },
      // From brain data
      [] ( NeuralGrid &Target, RandGen_t Seed ) {
        Target = World->EnvSettings.defaultBrain;
      }
    };

  }
}