#pragma once

#include "Env.World.h"
#include "Env.Settings.h"
#include "Engine.File.h"
#include "Application.h"
#include "Scene.Main.h"

namespace app {
  namespace env {

    // World_t
    World_t::World_t() : Thread( "World" ) {
      Size = 0;
      Area = 0;
      Generation = 0;

      DrawSyncWatch.reset();

      FoodScore = 0;

      Generator.Seed = rand();
    }
    World_t::~World_t() {
      free( Area );
    }

    void World_t::onThrdStep() {
      ManipulationMutex.lock();
      step();
      ManipulationMutex.unlock();
    }

    void World_t::setSize( Vec2World NewSize ) {
      if( Size == NewSize )
        return;

      /*if( Keep && NewSize.X * NewSize.Y > 0 && Size.X * Size.Y > 0 ) {
        WorldObject* OldArea = Area;
        Area = ( WorldObject* ) malloc( NewSize.X * NewSize.Y * sizeof( WorldObject ) );

        for( WorldPos_t a = 0; a < mMin( Size.Y, NewSize.Y ); a ++ )
          memcpy( Area + NewSize.X * a, OldArea + Size.X * a, mMin( Size.X, NewSize.X ) * sizeof( WorldObject* ) );

        free( OldArea );

      } else {*/
        CellList.clear();
        CellEmptyList.clear();

        FoodList.clear();
        FoodEmptyList.clear();

        WallList.clear();
        WallEmptyList.clear();

        Area = ( WorldObject* ) realloc( Area, NewSize.X * NewSize.Y * sizeof( WorldObject ) );
        memset( Area, 0, NewSize.X * NewSize.Y * sizeof( WorldObject ) );

      //}

      if( CellList.Capacity < ( SizeT ) NewSize.X * NewSize.Y )
      CellList.allocate( NewSize.X * NewSize.Y - CellList.Capacity );
      Size = NewSize;
    }

    WorldObject& World_t::get( Vec2World &Pos ) {
      ASSERT_THROW( Size.X * Size.Y == 0, "zeroWorldSize" );

      if( Pos.X < 0 )
        Pos.X += Size.X;
      else if( Pos.X >= Size.X )
        Pos.X -= Size.X;

      if( Pos.Y < 0 )
        Pos.Y += Size.Y;
      else if( Pos.Y >= Size.Y )
        Pos.Y -= Size.Y;

      return Area[ Pos.Y * Size.X + Pos.X ];
    }

    void World_t::step() {
      if( Size.X * Size.Y == 0 ) return;

      foreach_a( CellList ) {
        if( CellList[ a ].Health ) {
          ASSERT_THROW( CellList[ a ].Health <= 0, "as" );
          CellList[ a ].step();
        }
      }

      if( CellList.Capacity < CellList.Size * 2 )
        CellList.allocate( CellList.Size );

      checkForGeneration();
      Steps ++;

      Sync = !Sync;

      if( DrawSyncWatch.get() > 30 ) {
        DrawMutex.lock();
        updateDrawBuffer();
        DrawMutex.unlock();

        SyncTime = time::get();

        DrawSyncWatch.reset();
      }

      if( Steps == EnvSettings.stepBreakpoint ) {
        // Yeah I don't like this either
        scn::onWorkspace_PauseClick( *global->ScnMain->cPause );
      }
    }

    void World_t::checkForGeneration() {
      if( CellEmptyList.Size == CellList.Size ) {
        if( EnvSettings.StartCells == 0 ) {
          if( CellEmptyList.Size ) {
            CellEmptyList.clear();
            CellList.clear();
            Generation ++;
          }
        } else {
          CellEmptyList.clear();
          CellList.clear();

          clear();

          for( SizeT a = 0; a < EnvSettings.StartCells; a ++ )
            spawnCell( 0 );

          for( SizeT a = 0; a < EnvSettings.StartingFood; a ++ )
            spawnFood();

          Generation ++;
          Steps = 0;

          generationSave.clear();
          save( generationSave, false );
        }        
      }

      FoodScore += 100;
      while( EnvSettings.FoodInterval && FoodScore > EnvSettings.FoodInterval ) {
        spawnFood();
        FoodScore -= EnvSettings.FoodInterval;
      }
      if( FoodScore > 100 * 100 )
        FoodScore = 0;
    }
    void World_t::spawnCell( unsigned int retries ) {
      Vec2World Pos( Generator.roll( Size.X ), Generator.roll( Size.Y ) );
      WorldObject &o = get( Pos );
      if( o.Type == woNone || o.Type == woFood || retries > 50 ) {
        if( o.Type != woNone )
          destroyObjectAtPos( Pos );

        Cell &C = createCell( Pos );
         DefaultBrainFunctionList[ EnvSettings.defaultBrainFunctionIndex ]( C.Brain, Generator.roll() );
        //C.Brain.randomize( Generator.roll() );

      } else spawnCell( retries + 1 );
    }
    void World_t::updateDrawBuffer() {
      CellDrawList.clear();
      foreach_a( CellList ) {
        if( CellList[ a ].Health ) {
          CellDrawList.pushItem( CellList[ a ] );
        }
      }

      FoodDrawList.clear();
      foreach_a( FoodList ) {
        if( FoodList[ a ].Energy ) {
          FoodDrawList.pushItem( FoodList[ a ].Pos );
        }
      }

      WallDrawList.clear();
      foreach_a( WallList ) {
        if( WallList[ a ] != Size ) {
          WallDrawList.pushItem( WallList[ a ] );
        }
      }
    }

    void World_t::spawnFood() {
      Vec2World Pos( Generator.roll( Size.X ), Generator.roll( Size.Y ) );

      if( get( Pos ).Type == woNone )
        createFood( Pos );
    }

    Cell& World_t::createCell( Vec2World Pos ) {
      Cell *C = 0;

      if( CellEmptyList.Size ) {
        C = &CellList[ CellEmptyList[ --CellEmptyList.Size ] ];
        C->PosInContainer = CellEmptyList[ CellEmptyList.Size ];

      } else {
        CellList.pushItem();
        C = &CellList[ CellList.Size - 1 ];
        C->PosInContainer = CellList.Size - 1;
      }

      C->create( Pos );

      return *C;
    }
    Food& World_t::createFood( Vec2World Pos ) {
      Food *C = 0;

      if( FoodEmptyList.Size ) {
        C = &FoodList[ FoodEmptyList[ --FoodEmptyList.Size ] ];
        C->PosInContainer = FoodEmptyList[ FoodEmptyList.Size ];

      } else {
        FoodList.pushItem();
        C = &FoodList[ FoodList.Size - 1 ];
        C->PosInContainer = FoodList.Size - 1;
      }

      C->create( Pos );

      return *C;
    }
    void World_t::createWall( Vec2World Pos ) {
      Wall *W = 0;
      SizeT PosInContainer;

      if( WallEmptyList.Size ) {
        W = &WallList[ WallEmptyList[ --WallEmptyList.Size ] ];
        PosInContainer = WallEmptyList[ WallEmptyList.Size ];

      } else {
        WallList.pushItem();
        W = &WallList[ WallList.Size - 1 ];
        PosInContainer = WallList.Size - 1;
      }

      WorldObject &Obj = World->get( Pos );

      if( Obj.Type != woNone )
        int a = 5;
      ASSERT_THROW( Obj.Type != woNone, "createWallToNonEmptyPosition", "pos", String().pushInt( Pos.X ).pushChar( ':' ).pushInt( Pos.Y ).istr() );

      Obj.Type = woWall;
      Obj.Pos = PosInContainer;

      *W = Pos;
    }

    void World_t::createObjectAtPos( Vec2World pos, WorldObject_e tp ) {
      switch( tp ) {

      case woNone:
        return;

      case woCell:
        createCell( pos );
        break;

      case woWall:
        createWall( pos );
        break;

      case woFood:
        createFood( pos );
        break;

      }
    }
    void World_t::destroyObjectAtPos( Vec2World Pos ) {
      WorldObject& T = get( Pos );

      switch( T.Type ) {

      case woNone:
        break;


      case woCell:
        CellList[ T.Pos ].destroy();
        break;

      case woFood:
        FoodList[ T.Pos ].destroy();
        break;

      case woWall:
        WallList[ T.Pos ] = Size; // set to over-maximum to let it know the spot is free
        WallEmptyList.pushItem( T.Pos );
        T.Type = woNone;
        break;

      }

    }
    void World_t::clear() {
      CellList.clear();
      CellEmptyList.clear();

      FoodList.clear();
      FoodEmptyList.clear();

      for( WorldPos_t a = 0; a < Size.X * Size.Y; a ++ ) {
        if( Area[ a ].Type != woWall )
          Area[ a ].Type = woNone;
      }
    }
    void World_t::clear( WorldObject_e tp ) {
      if( tp == woWall ) {
        foreach_a( WallList ) {
          if( WallList[ a ] != Size ) get( WallList[ a ] ).Type = woNone;
        }

        WallList.clear();
        WallEmptyList.clear();

      } else if( tp == woCell ) {
        foreach_a( CellList ) {
          if( CellList[ a ].Health ) get( CellList[ a ].Pos ).Type = woNone;
        }

        CellList.clear();
        CellEmptyList.clear();

      }  else if( tp == woFood ) {
        foreach_a( FoodList ) {
          if( FoodList[ a ].Energy ) get( FoodList[ a ].Pos ).Type = woNone;
        }

        FoodList.clear();
        FoodEmptyList.clear();
      }
    }

    void World_t::save( Stream &str, bool includeGenerationStart /* = true */ ) {
      // Env
      str.pushSimple( Size );
      str.pushSimple( Generator );
      str.pushSimple( Sync );

      str.pushSimple( FoodScore );
      str.pushSimple( Generation );
      str.pushSimple( Steps );

      // Walls
      str.pushSimple< SizeT >( WallList.Size );
      for( SizeT a = 0; a < WallList.Size; a ++ ) {
        str.pushSimple( WallList[ a ] );
      }
      str.pushSimple< SizeT >( WallEmptyList.Size );
      for( SizeT a = 0; a < WallEmptyList.Size; a ++ ) {
        str.pushSimple( WallEmptyList[ a ] );
      }

      // Cells
      str.pushSimple< SizeT >( CellList.Size );
      for( SizeT a = 0; a < CellList.Size; a ++ ) {
        ASSERT_THROW( CellList[ a ].PosInContainer != a, "worldDesync" );
        str.pushSimple( CellList[ a ] );
      }
      str.pushSimple< SizeT >( CellEmptyList.Size );
      for( SizeT a = 0; a < CellEmptyList.Size; a ++ ) {
        str.pushSimple( CellEmptyList[ a ] );
      }

      // Food
      str.pushSimple< SizeT >( FoodList.Size );
      for( SizeT a = 0; a < FoodList.Size; a ++ ) {
        ASSERT_THROW( FoodList[ a ].PosInContainer != a, "worldDesync" );
        str.pushSimple( FoodList[ a ] );
      }
      str.pushSimple< SizeT >( FoodEmptyList.Size );
      for( SizeT a = 0; a < FoodEmptyList.Size; a ++ ) {
        str.pushSimple( FoodEmptyList[ a ] );
      }

      if( includeGenerationStart )
        str.push( generationSave );
    }
    void World_t::load( Stream &str, bool includeGenerationStart /* = true */ ) {
      // Env
      Vec2World WSz;

      int a = 0;
      str.pullSimple( WSz );
      str.pullSimple( Generator );
      str.pullSimple( Sync );

      str.pullSimple( FoodScore );
      str.pullSimple( Generation );
      str.pullSimple( Steps );

      global->ScnMain->cWorldSize->cListBox.SelectedItem = ( WSz.X == 32 ) ? 0 : ( ( WSz.X == 64 ) ? 1 : 2 );

      SizeT s;
      WorldObject *wo;

      setSize( WSz );

      // Clear world
      Vec2World v;
      for( v.X = 0; v.X < Size.X; v.X ++ )
        for( v.Y = 0; v.Y < Size.Y; v.Y ++ ) {
          get( v ).Type = woNone;
          get( v ).Pos = 0;
        }

      // Walls
      str.pullSimple< SizeT >( s );
      WallList.clear();
      WallList.allocate( s );
      WallList.Size = s;
      for( SizeT a = 0; a < s; a ++ ) {
        str.pullSimple( WallList[ a ] );

        if( WallList[ a ].oneGreaterOrEqual( Size ) ) continue;

        wo = &get( WallList[ a ] );
        wo->Type = woWall;
        wo->Pos = a;
      }

      str.pullSimple< SizeT >( s );
      WallEmptyList.clear();
      WallEmptyList.allocate( s );
      WallEmptyList.Size = s;
      for( SizeT a = 0; a < s; a ++ )
        str.pullSimple( WallEmptyList[ a ] );

      // Cells
      str.pullSimple< SizeT >( s );
      CellList.clear();
      CellList.allocate( s );
      CellList.Size = s;
      for( SizeT a = 0; a < s; a ++ ) {
        Cell &c = CellList[ a ];

        str.pullSimple( c );
        c.PosInContainer = a;

        if( c.Health <= 0 || c.Pos.oneGreaterOrEqual( Size ) ) continue;

        wo = &get( c.Pos );
        wo->Type = woCell;
        wo->Pos = a;
      }
      str.pullSimple< SizeT >( s );
      CellEmptyList.clear();
      CellEmptyList.allocate( s );
      CellEmptyList.Size = s;
      for( SizeT a = 0; a < s; a ++ )
        str.pullSimple( CellEmptyList[ a ] );

      // Food
      str.pullSimple< SizeT >( s );
      FoodList.clear();
      FoodList.allocate( s );
      FoodList.Size = s;
      for( SizeT a = 0; a < s; a ++ ) {
        Food &f = FoodList[ a ];

        str.pullSimple( f );
        f.PosInContainer = a;

        if( f.Energy <= 0 || f.Pos.oneGreaterOrEqual( Size ) ) continue;

        wo = &get( f.Pos );
        wo->Type = woFood;
        wo->Pos = a;
      }
      str.pullSimple< SizeT >( s );
      FoodEmptyList.clear();
      FoodEmptyList.allocate( s );
      FoodEmptyList.Size = s;
      for( SizeT a = 0; a < s; a ++ )
        str.pullSimple( FoodEmptyList[ a ] );

      if( includeGenerationStart )
        str.pull( generationSave );
    }

    World_t* World;

  }
}