#pragma once

#include "Engine.Vector.h"
#include "Engine.ZeroBaseContainer.h"
#include "Engine.Thread.h"
#include "Engine.Time.h"

#include "Env.Cell.h"
#include "Env.Types.h"
#include "Env.Food.h"
#include "Env.Types.h"
#include "Env.Settings.h"

namespace app {
  namespace env {

    enum WorldObject_e {
      woNone,
      woCell,
      woFood,
      woWall
    };

    struct  WorldObject {
      WorldObject_e Type;
      WorldPos_t Pos;
    };
    /*struct WorldStats {
      SizeT totalCells, totalMovement, totalEat, totalKill, totalReproduce;
      SizeT steps;
    };*/

    class World_t : public Thread {

    public:
      Vec2World Size;
      WorldObject *Area;
      Mutex ManipulationMutex;

    public:
      StopWatch DrawSyncWatch;
      Mutex DrawMutex;

    public:
      Container< Cell > CellList;
      Container< WorldPos_t > CellEmptyList; // optimalization for faster finding empty cell slots
      Container< CellDrawData > CellDrawList;

    public:
      Container< Food > FoodList;
      Container< WorldPos_t > FoodEmptyList;
      Container< FoodDrawData > FoodDrawList;
      SizeT FoodScore;

    public:
      Container< Wall > WallList;
      Container< WorldPos_t > WallEmptyList;
      Container< WallDrawData > WallDrawList;

    public:
      Settings EnvSettings;
      SizeT Generation, Steps;
      RandGen Generator;
      time::type SyncTime;
      Stream generationSave;

      //WorldStats stats;

      bool Sync; // Sync var

    public:
      World_t();
      ~World_t();

    public:
      virtual void onThrdStep();

    public:
      void setSize( Vec2World NewSize );

      WorldObject& get( Vec2World &Pos );

      void step();
      inline void checkForGeneration();
      inline void updateDrawBuffer();

    public:
      void spawnFood();
      void spawnCell( unsigned int retries );

      Cell& createCell( Vec2World Pos );
      Food& createFood( Vec2World Pos );
      void createWall( Vec2World Pos );

      void createObjectAtPos( Vec2World pos, WorldObject_e tp );
      void destroyObjectAtPos( Vec2World Pos );
      void clear();
      void clear( WorldObject_e tp );

    public:
      void save( Stream &str, bool includeGenerationStart = true );
      void load( Stream &str, bool includeGenerationStart = true );

    };

    extern World_t* World;

  }
}