#include "GUI.EnvSettings.h"
#include "GUI.Window.h"
#include "GUI.Button.h"
#include "GUI.CustomDrawListBox.h"
#include "GUI.ComboBox.h"

#include "Resource.Font.h"

#include "Engine.Global.h"
#include "Engine.Application.h"
#include "Engine.File.h"
#include "Engine.GUI.Skin.h"
#include "Engine.Rectangle.h"

#include "Env.World.h"
#include "Env.Settings.h"

namespace app {
  namespace gui {

    using namespace env;

    class EEnvSettingsProperty;

    class EnvSettingsPropertyList {

    public:
      EnvSettingsPropertyList() {

      }
      ~EnvSettingsPropertyList();

    public:
      Container< EEnvSettingsProperty* > Properties;

    };

    class EEnvSettingsProperty {

    public:
      EEnvSettingsProperty() {};
      virtual ~EEnvSettingsProperty() {};

    public:
      virtual void drawItem( CustomDrawListBox &ListBox, Vec2Screen Pos, ElementState_e State ) = 0;

    };

    template< typename T >
    class EnvSettingsProperty : public EEnvSettingsProperty {

    public:
      T &Value;
      double Min, Max, OrigValue;
      String Name;
      bool Slide;

    public:
      EnvSettingsProperty( T& Value, double Min, double Max, const IString &IName, EnvSettingsPropertyList *List ) : Value( Value ), Min( Min ), Max( Max ) {
        List->Properties.pushItem( this );
        Slide = false;

        Name.pushCChar( "<> " ).pushString( IName );
      }
      virtual ~EnvSettingsProperty() {

      }

    public:
      virtual void drawItem( CustomDrawListBox &ListBox, Vec2Screen Pos, ElementState_e State ) {

        if( Slide ) {
          World->ManipulationMutex.lock();
          Value = ( T ) mLimit< double >(
            Value + pow( abs( global_s.Cursor->RelPos.X ), 2.0 ) * ( ( global_s.Cursor->RelPos.X > 0 ) ? 1 : -1 ),
            Min,
            Max );
          World->ManipulationMutex.unlock();
        }

        if( Slide && global_s.Cursor->PrimaryInteraction->isRelease() )
          Slide = false;
        else if(
          !Slide
          && State == esHover
          && global_s.Cursor->PrimaryInteraction->isPress()
          //&& Rec2Screen( Pos, Vec2Screen( ListBox.RealSize.X, ListBox.ElementSize ) ).contains( global_s.Cursor->Pos )
          ) {
          Slide = true;
          OrigValue = ( double ) Value;
        }

        if( State == esHover && global_s.Cursor->SecondaryInteraction->isPress() ) {
          World->ManipulationMutex.lock();
          Value = Min;
          World->ManipulationMutex.unlock();
        }

        ListBox.Skn->drawLabel( Vec2Screen( Pos.X + 5, Pos.Y + ListBox.ElementSize / 2 ), Name, res::Font::valCenter );
        ListBox.Skn->drawLabel( Vec2Screen( Pos.X + ListBox.RealSize.X - 15, Pos.Y + ListBox.ElementSize / 2 ), String().pushFloat( ( float ) Value ), res::Font::valCenter | res::Font::halRight );
      }

    };
    class EnvSettingsPropertySection : public EEnvSettingsProperty {

    public:
      String Name;

    public:
      EnvSettingsPropertySection( const IString& Name, EnvSettingsPropertyList *List ) : Name( Name ) {
        List->Properties.pushItem( this );
      }
      virtual ~EnvSettingsPropertySection() {
      
      }

    public:
      virtual void drawItem( CustomDrawListBox &ListBox, Vec2Screen Pos, ElementState_e State ) {
        ListBox.Skn->drawPanel( Pos, Vec2Screen( ListBox.RealSize.X, ListBox.ElementSize ), Skin::ptPanel, esNormal );
        ListBox.Skn->drawLabel( Vec2Screen( Pos.X + 5, Pos.Y + ListBox.ElementSize / 2 ), Name, res::Font::valCenter );
      }

    };
    class EnvSettingsPropertyDefBrain : public EEnvSettingsProperty {

    public:
      EnvSettingsPropertyDefBrain( EnvSettingsPropertyList *List ) {
        List->Properties.pushItem( this );
      }
      virtual ~EnvSettingsPropertyDefBrain() {

      }

    public:
      virtual void drawItem( CustomDrawListBox &ListBox, Vec2Screen Pos, ElementState_e State ) {
        //ListBox.Skn->drawPanel( Pos, Vec2Screen( ListBox.RealSize.X, ListBox.ElementSize ), Skin::ptPanel, esNormal );
        if( State == esHover
          && global_s.Cursor->PrimaryInteraction->isPress()
          && Rec2Screen( Pos, Vec2Screen( ListBox.RealSize.X, ListBox.ElementSize ) ).contains( global_s.Cursor->Pos ) ) {
            World->ManipulationMutex.lock();
            World->EnvSettings.defaultBrainFunctionIndex ++;
            if( World->EnvSettings.defaultBrainFunctionIndex >= DefaultBrainFunctionListCount ) World->EnvSettings.defaultBrainFunctionIndex = 0;
            World->ManipulationMutex.unlock();
        }

        ListBox.Skn->drawLabel( Vec2Screen( Pos.X + 5, Pos.Y + ListBox.ElementSize / 2 ), String( "o  Default: " ).pushString( DefaultBrainFunctionListNames[ World->EnvSettings.defaultBrainFunctionIndex ] ), res::Font::valCenter );
      }

    };

    // blah I know
    EnvSettingsPropertyList::~EnvSettingsPropertyList() {
      foreach_a( Properties )
        delete Properties[ a ];
    }

    void onEnvSettings_ListBoxItemDraw( CustomDrawListBox &ListBox, UInt16 Item, Vec2Screen Pos, ElementState_e State ) {
      ( ( EnvSettingsPropertyList* ) ListBox.EventData )->Properties[ Item ]->drawItem( ListBox, Pos, State );
    }
    void onEnvSettings_WindowClose( Window &Win ) {
      env::Settings &Set = *( ( env::Settings* ) Win.EventData );
      Set.WindowOpen = false;

      findInGuiDecl( Win, "ListBox", CustomDrawListBox ) {
        delete ( EnvSettingsPropertyList* ) E->EventData;
      }
    }

    gui::ComboBox *cBox;
    void onEnvSet_save( gui::Button &btn ) {
      Stream str;
      World->ManipulationMutex.lock();
      World->EnvSettings.save( str );
      World->ManipulationMutex.unlock();
      saveFile( String( "../save/envpreset_" ).pushInt( cBox->cListBox.SelectedItem ).pushCChar( ".bin" ), str );
    }
    void onEnvSet_load( gui::Button &btn ) {
      String fn = String( "../save/envpreset_" ).pushInt( cBox->cListBox.SelectedItem ).pushCChar( ".bin" );
      if( !fileExists( fn ) ) return;

      Stream str;
      readFile( fn, str );
      World->ManipulationMutex.lock();
      World->EnvSettings.load( str );
      World->ManipulationMutex.unlock();
    }

    void openEnvSettings( env::Settings &Set ) {
      if( Set.WindowOpen )
        return;

      Set.WindowOpen = true;
      Element *W = global_s.Application->CurrentScene->GUI.loadFromFile( "../cfg/gui/EnvSettings.gui" );
      ASSERT_RETURN( !W, "envSettingsGuiError" );

      findInGuiNoDecl( *W, "EnvSaveBox", gui::ComboBox, cBox ) {}

      findInGuiDecl( *W, "EnvSave", gui::Button ) {
        E->onClick = onEnvSet_save;
      }
      findInGuiDecl( *W, "EnvLoad", gui::Button ) {
        E->onClick = onEnvSet_load;
      }

      EnvSettingsPropertyList *L = new EnvSettingsPropertyList;
      // Properties
      new EnvSettingsPropertySection( "Health & energy factors", L );
      new EnvSettingsProperty< CellHealth_t >( Set.CellHealth, 1, 16000, "Initial health", L );
      new EnvSettingsProperty< CellHealth_t >( Set.StartingEnergy, 1, 16000, "Initial energy", L );
      new EnvSettingsProperty< CellHealth_t >( Set.EnergyForReproduction, 1, 16000, "Energy for reproduction", L );
      new EnvSettingsProperty< CellHealth_t >( Set.EnergyForSex, 1, 16000, "Energy for sex. rep.", L );
      new EnvSettingsProperty< CellHealth_t >( Set.EnergyForMovement, 1, 16000, "Energy for movement", L );
      new EnvSettingsProperty< CellHealth_t >( Set.EnergyWrongAction, 1, 16000, "Energy for wrong action", L );

      new EnvSettingsPropertySection( "Brain", L );
      new EnvSettingsPropertyDefBrain( L );
      new EnvSettingsProperty< UInt8 >( Set.WeightMutationChance, 0, 1000, "Weight mut. chance (%%)", L );
      new EnvSettingsProperty< UInt8 >( Set.TresholdMutationChance, 0, 1000, "Tres. mut. chance (%%)", L );
      new EnvSettingsProperty< SizeT >( Set.mutationLevelMultiplier, 0, 1000, "Depth multiplier (%)", L );

      new EnvSettingsPropertySection( "Food", L );
      new EnvSettingsProperty< CellHealth_t >( Set.EnergyFromFood, 1, 16000, "Energy from food", L );
      new EnvSettingsProperty< CellHealth_t >( Set.EnergyOverfeed, 1, 16000, "Energy overfeed", L );
      new EnvSettingsProperty< SizeT >( Set.FoodInterval, 0, 10000, "Food spawn interval", L );
      new EnvSettingsProperty< SizeT >( Set.StartingFood, 0, World->Size.X * World->Size.Y - 1, "Initial food count", L );

      new EnvSettingsPropertySection( "Simulation", L );
      new EnvSettingsProperty< SizeT >( Set.StartCells, 0, 100, "Adam count", L );
      new EnvSettingsProperty< SizeT >( Set.stepBreakpoint, 0, 10000000, "Breakpoint after", L );

      // GUI
      findInGuiDecl( *W, "EnvSettings", Window ) {
        E->EventData = &Set;
        E->onDestroy = onEnvSettings_WindowClose;
      }

      findInGuiDecl( *W, "ListBox", CustomDrawListBox ) {
        E->onItemDraw = onEnvSettings_ListBoxItemDraw;
        E->EventData = L;

        E->ElementCount = L->Properties.Size;
        E->ElementSize = 18;

      } else delete L;

      W->recalc();
    }

  }
}