#include "GUI.NeuralEditor.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"
#include "Engine.Application.h"
#include "Engine.OpenGL.h"
#include "Engine.Display.h"
#include "Engine.Rectangle.h"
#include "Engine.File.h"

#include "Resource.Font.h"

#include "Env.NeuralGrid.h"
#include "Env.World.h"
#include "Env.Settings.h"

#include "Application.h"
#include "Scene.Main.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    using namespace env;

    NeuralEditor *neuralEditor = NULL;

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< NeuralEditor >( "NeuralEditor" );
    }

    // NeuralEditor
    NeuralEditor::NeuralEditor() {
      CellPos = 0; //-1;
      SelCell = -2;
      Scroll = Vec2Screen( 150, 70 );

      ASSERT( neuralEditor != NULL, "neuralEditorAlreadyOpened" );
      neuralEditor = this;
    }
    NeuralEditor::~NeuralEditor() {
      neuralEditor = NULL;
    }

    void NeuralEditor::init( cfx::Node &Source ) {
      Element::init( Source );
      IContent::init( Source );

      findInGuiNoDecl( *this, "NeuSaveBox", ComboBox, cSaveSel ) {
        cSaveSel->cListBox.EventData = this;
        cSaveSel->cListBox.onChange = []( gui::CustomDrawListBox &lb )  {
          NeuralEditor *E = ( NeuralEditor* ) lb.EventData;
          E->cSave->Enabled = lb.SelectedItem > DefaultBrainFunctionListCount - 2;
        };
      }

      findInGuiNoDecl( *this, "NeuLoad", Button, cLoad ) {
        cLoad->EventData = this;
        cLoad->onClick = [] ( gui::Button &btn ) {
          NeuralEditor *E = ( NeuralEditor* ) btn.EventData;

          if( E->cSaveSel->cListBox.SelectedItem > DefaultBrainFunctionListCount - 2 ) {
            Stream str;
            String fn = String( "../save/neuralpreset_" ).pushInt( E->cSaveSel->cListBox.SelectedItem ).pushCChar( ".bin" );

            if( !fileExists( fn ) ) return;
            readFile( fn, str );

            World->ManipulationMutex.lock();
            str.pullSimple( World->CellList[ E->CellPos ].Brain );
            World->CellList[ E->CellPos ].step();
            World->ManipulationMutex.unlock();

          } else {
            World->ManipulationMutex.lock();
            DefaultBrainFunctionList[ E->cSaveSel->cListBox.SelectedItem ]( World->CellList[ E->CellPos ].Brain, World->Generator.roll() );
            World->ManipulationMutex.unlock();
          }

          env::Cell &c = World->CellList[ E->CellPos ];
          c.inputCheck();
          c.Brain.step();
        };
      }
      findInGuiNoDecl( *this, "NeuSave", Button, cSave ) {
        cSave->EventData = this;
        cSave->onClick = [] ( gui::Button &btn ) {
          NeuralEditor *E = ( NeuralEditor* ) btn.EventData;
          Stream str;
          str.pushSimple( E->TempCell.Brain );
          saveFile( String( "../save/neuralpreset_" ).pushInt( E->cSaveSel->cListBox.SelectedItem ).pushCChar( ".bin" ), str );
        };
      }
      findInGuiNoDecl( *this, "NeuUpload", Button, cUpl ) {
        cUpl->EventData = this;
        cUpl->onClick = [] ( gui::Button &btn ) {
          NeuralEditor *E = ( NeuralEditor* ) btn.EventData;

          if( E->cSaveSel->cListBox.SelectedItem > DefaultBrainFunctionListCount - 2 ) {
            Stream str;
            String fn = String( "../save/neuralpreset_" ).pushInt( E->cSaveSel->cListBox.SelectedItem ).pushCChar( ".bin" );

            if( !fileExists( fn ) ) return;
            readFile( fn, str );

            World->ManipulationMutex.lock();
            str.pullSimple( World->EnvSettings.defaultBrain );
            World->EnvSettings.defaultBrainFunctionIndex = DefaultBrainFunctionListCount - 1;
            World->ManipulationMutex.unlock();

          } else {
            World->ManipulationMutex.lock();
            World->EnvSettings.defaultBrainFunctionIndex = E->cSaveSel->cListBox.SelectedItem;
            World->ManipulationMutex.unlock();
          }
        };
      }
      findInGuiNoDecl( *this, "NeuThisUpload", Button, cThisUpl ) {
        cThisUpl->EventData = this;
        cThisUpl->onClick = [] ( gui::Button &btn ) {
          NeuralEditor *E = ( NeuralEditor* ) btn.EventData;

          World->ManipulationMutex.lock();
          World->EnvSettings.defaultBrain = World->CellList[ E->CellPos ].Brain;
          World->EnvSettings.defaultBrainFunctionIndex = DefaultBrainFunctionListCount - 1;
          World->ManipulationMutex.unlock();
        };
      }

      findInGuiNoDecl( *this, "NeuScroll", ScrollBar, cSBar ) {}

      Button *cNF;
      findInGuiNoDecl( *this, "NeuFind", Button, cNF ) {
        cNF->onClick = [] ( gui::Button &btn ) {
          global->ScnMain->ViewTranslate = neuralEditor->TempCell.Pos.cast< TypeScreen >() * 16 - global_s.Display->Resolution / 2;
        };
      }

      SyncWatch.reset();
      ForceSync = true;
      SelCell = -1;
    }

    void NeuralEditor::recalc() {
      Element::recalc(); 
      IContent::recalc();
    }

    void NeuralEditor::draw() {
      Element::draw();

      Skn->drawPanel( AbsolutePos, RealSize, Skin::ptPanel, State );   

      if( CellPos != -1 ) {

        if( SyncWatch.get() >= 1000/20 || ForceSync ) {
          ForceSync = false;
          SyncWatch.reset();

          World->ManipulationMutex.lock();
          TempCell = World->CellList[ CellPos ];
          World->ManipulationMutex.unlock();

          if( TempCell.Health <= 0 || CellPos >= World->CellList.Size ) {
            CellPos ++;
            if( CellPos >= World->CellList.Size )
              CellPos = 0;
            ForceSync = true;
          }

          if( SelCell != -2 )
            computeNeuralWeights();
        }
        
        /*if( cMove && cMove->State == esClicked ) {
          Scroll.Y += global_s.Cursor->RelPos.Y * 10;
          //Scroll.X = mLimit( Scroll.X, -100, 200 );
          Scroll.Y = mLimit( Scroll.Y, -500, 70 );
        }*/
        cSBar->Max = ( Float32 ) max( 0,  NG_width * 16 - RealSize.Y + 120 );
        cSBar->BarSize = ( Float32 ) RealSize.Y;
        Scroll.Y = 60 - ( TypeScreen ) cSBar->Value;
        Scroll.X = RealSize.X / 2 - 16 * 2;

        glPushScissor( AbsolutePos, RealSize );

        Vec2Screen BasePos;
        Color Col;
        double w;
        for( NG_t x = 0; x < NG_width; x ++ ) {
          for( NG_t y = 0; y < NG_depth; y ++ ) {
            NeuralGridCell &Cell = TempCell.Brain.Grid[ y ][ x ];
            BasePos.assign( AbsolutePos + Scroll + Vec2Screen( y * 20, x * 16 ) );
            

            if( y == 0 && x < _cngiCount )
              Skn->drawLabel( BasePos +  Vec2Screen( -2, 3 ), CellNGInput_s[ x ], res::Font::halRight );
            else if( y == NG_depth - 1 && x < _cngoCount )
              Skn->drawLabel( BasePos +  Vec2Screen( 18, 3 ), CellNGOutput_s[ x ], 0 );

            if( SelCell.Y != -1 && y < SelCell.Y ) {
              w = NeuralWeight[ x ][ y ];
              glDrawRectangleOutline( BasePos + 2, 12, Color( ( w < 0 ) ? 0 : 255, ( w < 0 ) ? 0 : 255, ( w < 0 ) ? 0 : 255, ( UInt8 )( abs( w * 255.0 ) ) ) );
            }

            auto v = 128 + Cell.Value * 127;
            Col.assign( 255 - v, v, 128, 255 );
            if( SelCell == Vec2Screen( x, y ) )
              Skn->drawLabel( BasePos + 8, String().pushInt( Cell.Treshold ), res::Font::halCenter | res::Font::valCenter | res::Font::efShadow, Col );
            else if( SelCell.Y - 1 == y )
              Skn->drawLabel( BasePos + 8, String().pushInt( TempCell.Brain.Grid[ SelCell.Y ][ SelCell.X ].Weight[ x ] ), res::Font::halCenter | res::Font::valCenter | res::Font::efShadow, Col );
            else
              glDrawRectangle( BasePos + 4, 8, Col );

          }
        }

        BasePos = global_s.Cursor->Pos - AbsolutePos - Scroll;
        if( ( State == esHover || State == esClicked ) && Rec2Screen( 0, Vec2Screen( NG_depth * 20, NG_width * 16 ) ).contains( BasePos ) ) {
          BasePos = Vec2Screen( BasePos.Y / 16, BasePos.X / 20 );

          if( State == esClicked && BasePos.Y >= 1 )
            SelCell.assign( BasePos );
          else if( SelCell == BasePos ) {
            if( global_s.Cursor->ScrollUp->isPress() && World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Treshold < NG_treshold_limit ) {
              World->ManipulationMutex.lock();
              World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Treshold += 1;
              World->ManipulationMutex.unlock();

            } else if( global_s.Cursor->ScrollDown->isPress() && World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Treshold > -NG_treshold_limit ) {
              World->ManipulationMutex.lock();
              World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Treshold -= 1;
              World->ManipulationMutex.unlock();
            } else
              cSBar->doWheelControl();
          }
          else if( SelCell.Y - 1 == BasePos.Y ) {
            if( global_s.Cursor->ScrollUp->isPress() && World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Weight[ BasePos.X ] < NG_weight_limit ){
              World->ManipulationMutex.lock();
              World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Weight[ BasePos.X ] += 1;
              World->ManipulationMutex.unlock();

            } else if( global_s.Cursor->ScrollDown->isPress() && World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Weight[ BasePos.X ] > -NG_weight_limit ) {
              World->ManipulationMutex.lock();
              World->CellList[ CellPos ].Brain.Grid[ SelCell.Y ][ SelCell.X ].Weight[ BasePos.X ] -= 1;
              World->ManipulationMutex.unlock();
            } else
              cSBar->doWheelControl();
          } else
            cSBar->doWheelControl();

        } else if( State == esClicked )
          SelCell.assign( - 1 );
        else if( State == esHover )
          cSBar->doWheelControl();

        glPopScissor();
      }

      IContent::draw();

      Skn->drawLabel( AbsolutePos + Vec2Screen( 25, 5 ), String( "Health" ), res::Font::halCenter );
      Skn->drawLabel( AbsolutePos + Vec2Screen( 25, 18 ), String().pushInt( TempCell.Health ), res::Font::halCenter );

      Skn->drawLabel( AbsolutePos + Vec2Screen( 75, 5 ), String( "Energy" ), res::Font::halCenter );
      Skn->drawLabel( AbsolutePos + Vec2Screen( 75, 18 ), String().pushInt( TempCell.Energy ), res::Font::halCenter );

      Skn->drawLabel( AbsolutePos + Vec2Screen( 125, 5 ), String( "Gen." ), res::Font::halCenter );
      Skn->drawLabel( AbsolutePos + Vec2Screen( 125, 18 ), String().pushInt( TempCell.generation ), res::Font::halCenter );

      Skn->drawLabel( AbsolutePos + Vec2Screen( 175, 5 ), String( "Traveled" ), res::Font::halCenter );
      Skn->drawLabel( AbsolutePos + Vec2Screen( 175, 18 ), String().pushInt( TempCell.distanceTraveled ), res::Font::halCenter );

      Skn->drawLabel( AbsolutePos + Vec2Screen( 225, 5 ), String( "Eaten" ), res::Font::halCenter );
      Skn->drawLabel( AbsolutePos + Vec2Screen( 225, 18 ), String().pushInt( TempCell.foodEaten ), res::Font::halCenter );

      Skn->drawLabel( AbsolutePos + Vec2Screen( 275, 5 ), String( "Killed" ), res::Font::halCenter );
      Skn->drawLabel( AbsolutePos + Vec2Screen( 275, 18 ), String().pushInt( TempCell.cellsEaten ), res::Font::halCenter );
    }

    void NeuralEditor::computeNeuralWeights() {

      NG_t x, y, p;
      for( x = 0; x < NG_width; x ++ ) {
        for( y = 0; y < NG_depth; y ++ ) {
          NeuralWeight[ x ][ y ] = 0;
        }
      }

      NeuralGridCell &C = TempCell.Brain.Grid[ SelCell.Y ][ SelCell.X ];
      double Tres = abs( ( double ) C.Treshold );
      if( Tres == 0 ) Tres = 0.0001;
      for( x = 0; x < NG_width; x ++ ) {
        NeuralWeight[ x ][ SelCell.Y - 1 ] = mLimit( ( double )( C.Weight[ x ] ) / Tres, -1.0, 1.0 );
      }

      double s;
      for( y = SelCell.Y - 2; y >= 0; y -- ) {
        for( x = 0; x < NG_width; x ++ ) {
          for( p = 0; p < NG_width; p ++ ) {
            Tres = TempCell.Brain.Grid[ y + 1 ][ p ].Treshold;
            if( Tres == 0 ) Tres = 0.0001;
            s = NeuralWeight[ p ][ y + 1 ] * mLimit( ( double ) TempCell.Brain.Grid[ y + 1 ][ p ].Weight[ x ] / abs( ( double ) Tres ), -1.0, 1.0 );
            NeuralWeight[ x ][ y ] += s;
          }
          NeuralWeight[ x ][ y ] = mLimit( NeuralWeight[ x ][ y ], -1.0, 1.0 );
        }
      }

    }

    interfaceRoot( NeuralEditor,
      interfaceEntry( Element )
      );

  }

}

#endif