#include "Scene.Main.h"

#include "Engine.GUI.Skin.h"
#include "Engine.File.h"

#include "GUI.EnvSettings.h"
#include "GUI.Button.h"
#include "GUI.Panel.h"
#include "GUI.NeuralEditor.h"

#include "Toolkit.Display.h"
#include "Application.h"
#include "Resource.Font.h"

#include "Env.World.h"
#include "Env.Cell.h"
#include "Env.Settings.h"

#include "WallPresets.h"

namespace app {
  namespace scn {

    using namespace env;

    static const env::WorldObject_e ToolObjType[] = {
      env::woWall,
      env::woFood,
      env::woCell
    };

    void onWorkspace_EnvSettingsClick( gui::Button &Btn ) {
      gui::openEnvSettings( World->EnvSettings );
    }
    void onWorkspace_MoveHold( gui::Button &Btn ) {
      (( Main* ) Btn.EventData)->ViewTranslate -= global_s.Cursor->RelPos * 3;
    }

    void onWorkspace_PlayClick( gui::Button &Btn ) {
      findInGuiDecl( *Btn.GRoot, "Pause", gui::Button )
        E->Enabled = true;

      Btn.Enabled = false;

      World->ManipulationMutex.lock();
      findInGuiDecl( *Btn.GRoot, "FPS", gui::ComboBox )
        World->Performance.limitFPS( E->Items[ E->cListBox.SelectedItem ]->toInt() );
      World->ManipulationMutex.unlock();
    }
    void onWorkspace_PauseClick( gui::Button &Btn ) {
      findInGuiDecl( *Btn.GRoot, "Play", gui::Button )
        E->Enabled = true;

      Btn.Enabled = false;
      World->ManipulationMutex.lock();
      World->Performance.limitFPS( 0 );
      World->ManipulationMutex.unlock();
    }
    void onWorkspace_StepClick( gui::Button &Btn ) {
      findInGuiDecl( *Btn.GRoot, "Pause", gui::Button )
        E->Enabled = false;
      findInGuiDecl( *Btn.GRoot, "Play", gui::Button )
        E->Enabled = true;

      World->ManipulationMutex.lock();
      World->Performance.limitFPS( 0 );
      World->Performance.oneStep();
      World->ManipulationMutex.unlock();
    }

    void onWorkspace_NormalSpeedClick( gui::Button &Btn ) {
      findInGuiDecl( *Btn.GRoot, "FPS", gui::ComboBox ) {
        E->cListBox.SelectedItem = 4; // Hardcoded
        E->cListBox.PrevSelectedItem = E->cListBox.SelectedItem;
        E->cListBox.onChange( E->cListBox );
      }
    }
    void onWorkspace_HighSpeedClick( gui::Button &Btn ) {
      findInGuiDecl( *Btn.GRoot, "FPS", gui::ComboBox ) {
        E->cListBox.SelectedItem = 11; // Hardcoded
        E->cListBox.PrevSelectedItem = E->cListBox.SelectedItem;
        E->cListBox.onChange( E->cListBox );
      }
    }

    void onWorkspace_FPSChange( gui::CustomDrawListBox &LB ) {
      findInGuiDecl( *LB.GRoot, "Pause", gui::Button )
        E->Enabled = true;
      findInGuiDecl( *LB.GRoot, "Play", gui::Button )
        E->Enabled = false;

      World->ManipulationMutex.lock();
      findInGuiDecl( *LB.GRoot, "FPS", gui::ComboBox )
        World->Performance.limitFPS( E->Items[ E->cListBox.SelectedItem ]->toInt() );
      World->ManipulationMutex.unlock();
    }
    void onWorkspace_ToolChange( gui::Button &btn ) {
      global->ScnMain->cToolC->Enabled = true;
      global->ScnMain->cToolF->Enabled = true;
      global->ScnMain->cToolW->Enabled = true;

      btn.Enabled = false;

      if( btn.Caption == "Wall" )
        global->ScnMain->DrawType = woWall;
      else if( btn.Caption == "Cell" )
        global->ScnMain->DrawType = woCell;
      else
        global->ScnMain->DrawType = woFood;
    }

    void onWorkspace_ToolsClick( gui::Button &Btn ) {
      findInGuiDecl( *Btn.GRoot, "ToolPanel", gui::Panel )
        E->Visible = !E->Visible;
    }
    void onWorkspace_ClearSClick( gui::Button &Btn ) {
      World->ManipulationMutex.lock();
      if( Btn.Name == "ClearAll" ) {
        World->clear( env::woWall );
        World->clear();
      }
      else if( Btn.Name == "ClearWalls" ) World->clear( env::woWall );
      else if( Btn.Name == "ClearFood" ) World->clear( env::woFood );
      else if( Btn.Name == "ClearCells" ) World->clear( env::woCell );

      World->ManipulationMutex.unlock();
    }

    #define EXPERIMENT_COUNT 4
    void onWorkspace_save( gui::Button &Btn ) {
      Main *scn = global->ScnMain;

      World->ManipulationMutex.lock();
      Stream str;
      World->EnvSettings.save( str );
      World->save( str );
      World->ManipulationMutex.unlock();

      saveFile( String( "../save/save_" ).pushInt( scn->cSave->cListBox.SelectedItem - EXPERIMENT_COUNT ).pushCChar( ".bin" ), str );
    }
    void onWorkspace_load( gui::Button &Btn ) {
      Main *scn = global->ScnMain;
      String Fn;

      if( scn->cSave->cListBox.SelectedItem < EXPERIMENT_COUNT )
        Fn = String( "../save/exp_" ).pushInt( scn->cSave->cListBox.SelectedItem + 1 ).pushCChar( ".bin" );
      else
        Fn = String( "../save/save_" ).pushInt( scn->cSave->cListBox.SelectedItem - EXPERIMENT_COUNT ).pushCChar( ".bin" );

      if( !fileExists( Fn ) ) return;

      Stream str;
      readFile( Fn, str );

      World->ManipulationMutex.lock();
      World->EnvSettings.load( str );
      World->load( str );
      World->ManipulationMutex.unlock();
    }
    void onWorkspace_saveBoxChange( gui::CustomDrawListBox &LB ) {
      Main *scn = global->ScnMain;

      findInGuiDecl( scn->GUI, "SaveSave", gui::Button )
        E->Enabled = LB.SelectedItem >= EXPERIMENT_COUNT;
    }

    void onWorkspace_WorldSizeChange( gui::CustomDrawListBox &LB ) {
      World->ManipulationMutex.lock();
      World->setSize( LB.SelectedItem == 0 ? 32 : ( LB.SelectedItem == 1 ? 64 : 128 ) );
      World->ManipulationMutex.unlock();
    }

    Main::Main() : Scene( using2DDraw ) {
      World = new World_t;

      Drag = false;
      DrawMode = dmNone;
      DrawType = woWall;
    }
    Main::~Main() {
      Stream str;
      /*World->EnvSettings.save( Str );
      saveFile( "settings.bin", Str );*/
      World->EnvSettings.save( str );
      World->save( str );
      saveFile( "../save/session.bin", str );

      World->waitForStop();
      delete World;
    }

    void Main::onLoadResources() {
      GUI.loadFromFile( "../cfg/gui/workspace.gui" );
      //GUI.loadFromFile( "../cfg/gui/neuralEditor.gui" );

      wallPresetsInit( *this );

      findInGuiDecl( GUI, "EnvSettings", gui::Button )
        E->onClick = onWorkspace_EnvSettingsClick;

      findInGuiDecl( GUI, "Normalspeed", gui::Button )
        E->onClick = onWorkspace_NormalSpeedClick;
      findInGuiDecl( GUI, "Highspeed", gui::Button )
        E->onClick = onWorkspace_HighSpeedClick;

      findInGuiDecl( GUI, "ShowPanel", gui::Button )
        E->onClick = onWorkspace_ToolsClick;

      findInGuiNoDecl( GUI, "Play", gui::Button, cPlay ) {
        cPlay->onClick = onWorkspace_PlayClick;
        cPlay->EventData = this;
      }
      findInGuiNoDecl( GUI, "Pause", gui::Button, cPause ) {
        cPause->onClick = onWorkspace_PauseClick;
        cPause->EventData = this;
      }
      findInGuiNoDecl( GUI, "Step", gui::Button, cStep ) {
        cStep->onClick = onWorkspace_StepClick;
        cStep->EventData = this;
      }

      findInGuiNoDecl( GUI, "FPS", gui::ComboBox, cSpeed ) {
        cSpeed->cListBox.onChange = onWorkspace_FPSChange;
        World->waitForRun();
        World->Performance.limitFPS( cSpeed->Items[ cSpeed->cListBox.SelectedItem ]->toInt() );
      }

      findInGuiNoDecl( GUI, "ToolWall", gui::Button, cToolW )
        cToolW->onClick = onWorkspace_ToolChange;
      findInGuiNoDecl( GUI, "ToolCell", gui::Button, cToolC )
        cToolC->onClick = onWorkspace_ToolChange;
      findInGuiNoDecl( GUI, "ToolFood", gui::Button, cToolF )
        cToolF->onClick = onWorkspace_ToolChange;

      findInGuiDecl( GUI, "ClearAll", gui::Button )
        E->onClick = onWorkspace_ClearSClick;
      findInGuiDecl( GUI, "ClearFood", gui::Button )
        E->onClick = onWorkspace_ClearSClick;
      findInGuiDecl( GUI, "ClearCells", gui::Button )
        E->onClick = onWorkspace_ClearSClick;
      findInGuiDecl( GUI, "ClearWalls", gui::Button )
        E->onClick = onWorkspace_ClearSClick;

      findInGuiDecl( GUI, "SaveLoad", gui::Button )
        E->onClick = onWorkspace_load;
      findInGuiDecl( GUI, "SaveSave", gui::Button )
        E->onClick = onWorkspace_save;
      findInGuiNoDecl( GUI, "SaveBox", gui::ComboBox, cSave )
        cSave->cListBox.onChange = onWorkspace_saveBoxChange;

      findInGuiNoDecl( GUI, "WorldSize", gui::ComboBox, cWorldSize )
        cWorldSize->cListBox.onChange = onWorkspace_WorldSizeChange;

      findInGuiDecl( GUI, "GenStart", gui::Button ) {
        E->onClick = [] ( gui::Button &btn ) {
          World->ManipulationMutex.lock();
          World->generationSave.Reader = 0;
          World->load( World->generationSave, false );
          World->ManipulationMutex.unlock();
        };
      }
      findInGuiDecl( GUI, "ResGenCnt", gui::Button ) {
        E->onClick = [] ( gui::Button &btn ) {
          World->Generation = 1;
        };
      }

      Stream str;
      readFile( "../save/session.bin", str );

      World->EnvSettings.load( str );
      World->load( str );

      World->run();
    }
    void Main::onDraw2D() {    
      if( World->Performance.MinPing == -1 )
        World->updateDrawBuffer();

      World->DrawMutex.lock();
      CellDrawList.clear().pushContainer( World->CellDrawList );
      FoodDrawList.clear().pushContainer( World->FoodDrawList );
      WallDrawList.clear().pushContainer( World->WallDrawList );
      World->DrawMutex.unlock();

      auto T = time::get();
      float AnimProg = mLimit<float>( ( float )( T - World->SyncTime ) / ( float ) World->Performance.getPing(), 0, 1 );

      // View translate
      if( !Drag && global_s.Cursor->SecondaryInteraction->isPress() && global_s.Cursor->ObjectOnCursor == 0 ) {
        Drag = true;

        // Cell info
        Vec2World Pos = ( ( ViewTranslate + global_s.Cursor->Pos ) / 16 ).cast< WorldPos_t >();

        if( Pos.everyLess( World->Size ) && Pos.everyGreaterOrEqual( 0 ) ) {

          World->ManipulationMutex.lock();
          WorldObject &o = World->get( Pos );
          World->ManipulationMutex.unlock();

          if( o.Type == woCell ) {
            if( !gui::neuralEditor ) GUI.loadFromFile( "../cfg/gui/neuralEditor.gui" );
            gui::neuralEditor->CellPos = o.Pos;
          }
        }
      }
      if( Drag ) {
        ViewTranslate -= global_s.Cursor->RelPos * 3;
        ViewTranslate.X = mLimit( ViewTranslate.X, 300 - global_s.Display->Resolution.X, World->Size.X * 16 - 300 );
        ViewTranslate.Y = mLimit( ViewTranslate.Y, 300 - global_s.Display->Resolution.Y, World->Size.Y * 16 - 300 );

        if( global_s.Cursor->SecondaryInteraction->isRelease() )
          Drag = false;
      }

      // Handle editing
      if( global_s.Cursor->PrimaryInteraction->isPress() && global_s.Cursor->ObjectOnCursor == 0 ) {
        Vec2World Pos = ( ( ViewTranslate + global_s.Cursor->Pos ) / 16 ).cast< WorldPos_t >();
        if( Pos.everyLess( World->Size ) && Pos.everyGreaterOrEqual( 0 ) ) {
          World->ManipulationMutex.lock();
          WorldObject &o = World->get( Pos );
          DrawMode = ( o.Type == DrawType ) ? dmDestroy : dmCreate;
          World->ManipulationMutex.unlock();

        }
      } else if( DrawMode != dmNone && global_s.Cursor->ObjectOnCursor == 0 ) {
        Vec2World Pos = ( ( ViewTranslate + global_s.Cursor->Pos ) / 16 ).cast< WorldPos_t >();

        if( Pos.everyLess( World->Size ) && Pos.everyGreaterOrEqual( 0 ) ) {
          World->ManipulationMutex.lock();
          WorldObject &o = World->get( Pos );
          if( o.Type == DrawType ) {
            if( DrawMode == dmDestroy )
              World->destroyObjectAtPos( Pos );

          } else if( DrawMode == dmCreate ) {
            if( o.Type != woNone )
              World->destroyObjectAtPos( Pos );

            World->createObjectAtPos( Pos, DrawType );
          }

          World->ManipulationMutex.unlock();
        }
        if( global_s.Cursor->PrimaryInteraction->isRelease() )
          DrawMode = dmNone;
      }


      // Draw env
      Vec2< float > P, PP, RP;
      CellDrawData c;
      SizeT watchedCellPIC = ( gui::neuralEditor ) ? gui::neuralEditor->TempCell.PosInContainer : -1;
      glDrawRectangleOutline( ViewTranslate * -1, World->Size.cast< TypeScreen >() * 16, Color( 180, 180, 180 ) );
      foreach_a( CellDrawList ) {
        c = CellDrawList[ a ];

        P = c.Pos.cast< float >();
        PP = c.PrevPos.cast< float >();
        if( ( P - PP ).mAbs().oneGreater( Vec2< float >( 1, 1 ) ) )
          RP = P;
        else
         RP = Vec2< float >( P.X * AnimProg + PP.X * ( 1 - AnimProg ), P.Y * AnimProg + PP.Y * ( 1 - AnimProg ) );

        glDrawRectangle( ( RP * 16 ).cast< TypeScreen >() + 6 - c.Energy / 2 - ViewTranslate, 4 + c.Energy, Color( 128 + CellDrawList[ a ].Health, 64, 64 + c.cellsEaten * 20 ) );

        if( c.posInCont == watchedCellPIC ) {
          glDrawRectangleOutline( ( RP * 16 ).cast< TypeScreen >() - ViewTranslate + 4 - c.Energy / 2, 8 + c.Energy, Color( 255, 200, 0 ) );
          //glDrawRectangleOutline( ( RP * 16 ).cast< TypeScreen >() - ViewTranslate + 0 - c.Energy / 2, 16 + c.Energy, Color( 255, 200, 0 ) );
        }
      }
      foreach_a( FoodDrawList ) {
        glDrawRectangle( FoodDrawList[ a ].cast< TypeScreen >() * 16 + 6 - ViewTranslate, 4, Color( 64, 255, 64 ) );
      }
      foreach_a( WallDrawList ) {
        glDrawRectangle( WallDrawList[ a ].cast< TypeScreen >() * 16 + 1 - ViewTranslate, 14, Color( 128, 128, 128 ) );
      }

      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( global_s.Display->Resolution.X - 5, 30 ),  String( "Create/destroy objects     LMB" ), res::Font::halRight );
      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( global_s.Display->Resolution.X - 5, 42 ),  String( "Move view/Cell info     RMB" ), res::Font::halRight );
    }
    void Main::onPostDraw2D() {
      //global_s.gui.SkinList[ 0 ]->drawLabel( 600, String( "FPS: " ).pushInt( global_s.Application->Performance.getFPS() ).pushCChar( "; Generation: " ).pushInt( World->Generation ).pushCChar( "; Thrd fps: " ).pushInt( World->Performance.getFPS() ), 0 );
      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( 110, 6 ), String( "FPS: " ).pushInt( global_s.Application->Performance.getFPS() ), 0 );
      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( 160, 6 ), String( "TFPS: " ).pushInt( World->Performance.getFPS() ), 0 );
      World->ManipulationMutex.lock();
      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( 260, 6 ),  String( "Generation: " ).pushInt( World->Generation ), 0 );
      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( 360, 6 ),  String( "Steps: " ).pushInt( World->Steps ), 0 );
      global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( 460, 6 ),  String( "Cells: " ).pushInt( World->CellList.Size - World->CellEmptyList.Size ), 0 );
      World->ManipulationMutex.unlock();

      //global_s.gui.SkinList[ 0 ]->drawLabel(  Vec2Screen( 5, 6 ),  String( "RMB: Move view" ), 0 );
    }

  }
}