#pragma once

#include "WallPresets.h"
#include "Scene.Main.h"
#include "GUI.ComboBox.h"
#include "GUI.ListBox.h"
#include "Env.World.h"
#include "Env.Types.h"
#include "Engine.File.h"

namespace app {

  using namespace env;

  gui::ComboBox *cBox;
  gui::Button *cSBtn;

  void wallPresetSaveToFile( SizeT index ) {
    Stream str;

    World->ManipulationMutex.lock();
    Container< Wall > cont = World->WallList;
    World->ManipulationMutex.unlock();

    foreach_a( cont ) str.pushSimple( cont[ a ] );
    saveFile( String().pushCChar( "../save/wallpreset_" ).pushInt( index ).pushCChar( ".bin" ), str );
  }
  void wallPresetLoadFromFile( SizeT index ) {
    Stream str;

    String fn = String().pushCChar( "../save/wallpreset_" ).pushInt( index ).pushCChar( ".bin" );
    if( !fileExists( fn ) ) return;
    
      readFile( fn, str );

    Vec2World w;
    for( SizeT a = 0; a < str.Size / sizeof( Vec2World ); a ++ ) {
      str.pullSimple( w );
      World->destroyObjectAtPos( w );
      World->createWall( w );
    }
  }

  typedef bool ( *wpFunc )( Vec2World, RandGen_t );

  bool wpEdges( Vec2World ps, RandGen_t seed ) {
    if( ps.X == 0 || ps.Y == 0 )
      return true;

    if( ps.X == World->Size.X - 1 || ps.Y == World->Size.Y - 1 )
      return true;

    return false;
  }
  bool wpRooms( Vec2World ps, RandGen_t seed ) {
    return
      ( ( 1 + ps.X + ( ps.Y / 6 ) * 3 ) % 6 < 3 && ps.Y % 6 == 0 ) ||
      ( ( 1 + ps.Y + ( ps.X / 6 ) * 3 ) % 6 < 3 && ps.X % 6 == 0 );
  }
  bool wpMaze( Vec2World ps, RandGen_t seed ) {
    return
      ( ps.X % 2 == 0 && ps.Y % 2 == 0 ) ||
      ( ( ps.X % 2 == 0 || ps.Y % 2 == 0 ) && mRand( 20 ) == 0 );
  }
  bool wpMazeMed( Vec2World ps, RandGen_t seed ) {
    return
      ( ps.X % 3 == 0 && ps.Y % 3 == 0 ) ||
      ( ( ps.X % 3 == 0 || ps.Y % 3 == 0 ) && randFromSeed( randFromSeed( randFromSeed( seed + ( ps.Y % 3 == 0 ) ) + ps.X / 3 ) + ps.Y / 3 ) % 10 == 0 );
  }
  bool wpMazeBig( Vec2World ps, RandGen_t seed ) {
    return
      ( ps.X % 4 == 0 && ps.Y % 4 == 0 ) ||
      ( ( ps.X % 4 == 0 || ps.Y % 4 == 0 ) && randFromSeed( randFromSeed( randFromSeed( seed + ( ps.Y % 4 == 0 ) ) + ps.X / 4 ) + ps.Y / 4 ) % 10 == 0 );
  }
  static const bool wpChinaPat[ 14 ][ 14 ] = {
    { 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0 },
    { 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0 },
    { 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1 },
    { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 },
    { 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0 },
    { 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 },
    { 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1 },
    { 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0 }, // Center
    { 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 1 },
    { 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 },
    { 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0 },
    { 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0 },
    { 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1 },
    { 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0 },
  };
  static const Vec2World wpChinaPatSz( 14 );
  bool wpChina( Vec2World ps, RandGen_t seed ) {
    return wpChinaPat[ ps.X % wpChinaPatSz.X ][ ps.Y % wpChinaPatSz.Y ];
  }

  static const wpFunc wpList[] = {
    &wpEdges, &wpRooms, &wpMaze, &wpMazeMed, &wpMazeBig, &wpChina
  };
  static const int wpListS = 6;

  void loadWallPreset( gui::Button &Btn ) {
    if( cBox->cListBox.SelectedItem < wpListS ) {
      wpFunc f = wpList[ cBox->cListBox.SelectedItem ];

      World->ManipulationMutex.lock();
      RandGen_t rand = World->Generator.roll();

      if( Btn.Name != "WallLoadAdd" )
        World->clear( env::woWall );
      
      Vec2World ps;
      for( ps.X = 0; ps.X < World->Size.X; ps.X ++ ) {
        for( ps.Y = 0; ps.Y < World->Size.Y; ps.Y ++ ) {
          if( f( ps, rand ) ) {
            World->destroyObjectAtPos( ps );
            World->createWall( ps );
          }
        }
      }

      World->ManipulationMutex.unlock();
    } else {
      World->ManipulationMutex.lock();
      if( Btn.Name != "WallLoadAdd" )
        World->clear( env::woWall );

      wallPresetLoadFromFile( cBox->cListBox.SelectedItem - wpListS );
      World->ManipulationMutex.unlock();
    }
  }
  void saveWallPreset( gui::Button &Btn ) {
    if( cBox->cListBox.SelectedItem < wpListS ) return;
    wallPresetSaveToFile( cBox->cListBox.SelectedItem - wpListS );
  }
  void wallBoxChange( gui::CustomDrawListBox &LB ) {
    if( cSBtn )
      cSBtn->Enabled = LB.SelectedItem >= wpListS;
  }

  void wallPresetsInit( scn::Main &Scn ) {
    findInGuiDecl( Scn.GUI, "WallLoad", gui::Button ) {
      E->onClick = &loadWallPreset;
    }
    findInGuiDecl( Scn.GUI, "WallLoadAdd", gui::Button ) {
      E->onClick = &loadWallPreset;
    }
    findInGuiNoDecl( Scn.GUI, "WallSave", gui::Button, cSBtn ) {
      cSBtn->onClick = &saveWallPreset;
    }
    findInGuiNoDecl( Scn.GUI, "WallBox", gui::ComboBox, cBox ) {
      cBox->cListBox.onChange = &wallBoxChange;
    }
  }

}