#pragma once

#include "Engine.String.h"

namespace app {

  enum eCfxProperty {
    cfxName,
    cfxValue,
    cfxComment
  };
  enum eCfxMergeFlag {
    cfxmOverwriteEmpty        = 1 << 0,

    cfxmNodesByName           = 0,
    cfxmNodesByPos            = 1 << 1,

    cfxmArgsByName            = 1 << 2,
    cfxmArgsByPos             = 0,

    cfxmDuplicate             = 1 << 3,
  };

  namespace cfx {

    String safeString( const String &Source );

    class Argument {

    public:
      String Name, Value;

    public:
      Argument();
      Argument( const IString &Name, const IString &Value = "" );

    public:
      String& operator>( eCfxProperty Property );

    public:
      void formatSave( String &Target, bool Prefix );

      void merge( Argument &Source, UInt8 Flags );

      Argument* duplicate();

    };

    class Node {

    public:
      String Name, Value, Comment;
      Container< Node* > Nodes;
      Container< Argument* > Arguments;

    public:
      Node();
      Node( const IString &Name, const IString &Value = String(), const IString &Comment = String() );

      ~Node();

    public:
      Node& operator>>( const char *Name );
      Node& operator>>( const IString &Name );
      Node& operator>>( const SizeT Pos );
      Node& N( const IString &Name );

      String& operator>( const char *Name );
      String& operator>( const IString &Name );
      String& operator>( const SizeT Pos );
      String& NV( const IString &Name );

      String& operator>( eCfxProperty Property );

      Argument& operator||( const char *Name );
      Argument& operator||( const IString &Name );
      Argument& operator||( const SizeT Pos );
      Argument& A( const IString &Name );

      String& operator|( const char *Name );
      String& operator|( const IString &Name );
      String& operator|( const SizeT Pos );
      String& AV( const IString &Name );

    public:
      bool operator>=( const char *Name );
      bool operator>=( const IString &Name );
      bool operator>=( const SizeT Pos );

      bool operator|=( const char *Name );
      bool operator|=( const IString &Name );
      bool operator|=( const SizeT Pos );

    public:
      Node& merge( Node &Source, UInt8 Flags );

      Node* duplicate();

      void clearUnalloc();

      void formatSave( String &Target, const IString &Prefix = String() );

      void debugPrintTree();

    };

    class Document : public Node {

    private:
      Container< Node* > Tree;
      SizeT Offset, Pos, Start, Blank, Line;
      Node *N;
      Argument *A;
      char Ch;
      bool Spaces;

    public:
      void load( const char *File );
      void load( const IString &File );

      void loadString( const IString &Source );

      void save( const char *File );
      void save( const IString &File );
      void save();

    private:
      bool parseNode();
      bool parseOffset();
      void parseName();
      void parseValue();
      void parseArgument();
      void parseArgumentValue();
      void parseComment();

    };

  }

}