#pragma once

#include "Engine.Math.h"
#include "Engine.Types.h"

#include <stdio.h>
#include <stdlib.h>

#define foreach_a( Stor ) for( SizeT a = 0; a < ( Stor ).Size; a ++ )
#define foreach_b( Stor ) for( SizeT b = 0; b < ( Stor ).Size; b ++ )

namespace app {

  template< typename T >
  class Container {

  public:
    T* Ptr;
    SizeT Size, Capacity;

  public:
    Container() {
      Size = 0;
      Capacity = 2;

      Ptr = ( T* ) malloc( Capacity * sizeof( T ) );
    }
    Container( const Container< T > &Source, SizeT Start, SizeT Count = -1 ) {
      Size = mMin< SizeT >( Count, Source.Size - mMin< SizeT >( Start, Source.Size - 1 ) );
      Capacity = 2;
      while ( Capacity < Size )
        Capacity *= 2;

      Ptr = ( T* ) malloc( Capacity * sizeof( T ) );
      memcpy( Ptr, Source.Ptr + Start, Size * sizeof( T ) );
    }
    Container( const Container< T > &Source ) {
      Size = Source.Size;
      Capacity = Source.Capacity;

      Ptr = ( T* ) malloc( Capacity * sizeof( T ) );
      memcpy( Ptr, Source.Ptr, Size * sizeof( T ) );
    }
    Container( const T *Source, SizeT Count ) {
      Size = Count;
      Capacity = 2;
      while ( Capacity < Size )
        Capacity *= 2;

      Ptr = ( T* ) malloc( Capacity * sizeof( T ) );
      memcpy( Ptr, Source, Size * sizeof( T ) );
    }

    ~Container() {
      free( Ptr );
    }

  public:
    Container< T >& pushItem() {
      allocate( 1 );
      Size++;
      return *this;
    }
    Container< T >& pushItem( const T &Source ) {
      allocate( 1 );
      Ptr[ Size++ ] = Source;
      
      return *this;
    }
    Container< T >& pushContainer( const Container< T > &Source, SizeT Start = 0, SizeT Count = -1 ) {
      if( !Source.Size )
        return *this;

      Count = mMin< SizeT >( Count, Source.Size - mMin< SizeT >( Start, Source.Size - 1 ) );
      allocate( Count );

      memcpy( Ptr + Size, Source.Ptr + Start, Count * sizeof( T ) );
      Size += Count;

      return *this;
    }

    Container< T >& clear() {
      Size = 0;
      return *this;
    }
    Container< T >& clearUnalloc() {
      Size = 0;
      Capacity = 2;

      Ptr = ( T* ) realloc( Ptr, Capacity * sizeof( T ) );

      return *this;
    }

    Container< T >& removeItem( const T &Source ) {
      for( SizeT a = 0; a < Size; a ++ ) {
        if( Ptr[ a ] == Source ) {
          memmove( Ptr + a, Ptr + a + 1, ( Size - a - 1 ) * sizeof( T ) );
          Size --;
          return *this;
        }

      }

      return *this;
    }
    SizeT removeItemRetPos( const T &Source ) {
      for( SizeT a = 0; a < Size; a ++ ) {
        if( Ptr[ a ] == Source ) {
          memmove( Ptr + a, Ptr + a + 1, ( Size - a - 1 ) * sizeof( T ) );
          Size --;
          return a;
        }

      }

      return -1;
    }
    Container< T >& removeItemPos( SizeT Pos ) {
      memmove( Ptr + Pos, Ptr + Pos + 1, ( Size - Pos - 1 ) * sizeof( T ) );
      Size --;

      return *this;
    }

  public:
    void allocate( SizeT Count ) {
      if ( Capacity >= Size + Count )
        return;

      while ( Capacity < Size + Count )
        Capacity *= 2;

      Ptr = ( T* ) realloc( Ptr, Capacity * sizeof( T ) );
    }

  public:
    T first( const T Default ) {
      if( Size == 0 )
        return Default;

      return Ptr[ 0 ];
    }
    T last( const T Default ) {
      if( Size == 0 )
        return Default;

      return Ptr[ Size - 1 ];
    }

  public:
    bool operator==( const Container< T > &Source ) const {
      if ( Source.Size != Size )
        return false;

      return memcmp( Ptr, Source.Ptr, sizeof( T ) * Size ) == 0;
    }
    bool operator!=( const Container< T > &Source ) const {
      if ( Source.Size != Size )
        return true;

      return memcmp( Ptr, Source.Ptr, sizeof( T ) * Size ) != 0;
    }

    void operator=( const Container< T > &Source ) {
      clear();
      pushContainer( Source );
    }

    T& operator[]( SizeT Pos ) {
      return Ptr[ Pos ];
    }
    const T& operator[]( SizeT Pos ) const {
      return Ptr[ Pos ];
    }

  };

}