#include "Engine.Controls.h"
#include "Toolkit.Display.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace ctrl {

    float PressLevel;

    hook( hookAppPreInit ) {
      linkEngineVariable( "Controls" > "PressLevel",    PressLevel,     float );

      global_s.Controls = new Manager();
      global_s.Controls->init();
    }
    hook( hookAppDestroy ) {
      global_s.Controls->destroy();
      delete global_s.Controls;
    }
    hook( hookAppPreStep ) {
      global_s.Controls->step();
    }

    // InputRecord
    InputRecord::InputRecord( const String &Name ) : ID( Name ) {
      Value = 0;
      PrevValue = 0;
      Handled = false;

      global_s.Controls->Records.pushItem( this );
    }
    InputRecord::~InputRecord() {

    }

    void InputRecord::step() {
      PrevValue = Value;
      Handled = false;
    }

    // Device
    Device::Device() {

    }
    Device::~Device() {

    }

    void Device::init( cfx::Node &Source ) {
      Name.pushString( Source.Value );
      onInit( Source );
    }

    // Manager
    Manager::Manager() {

    }
    Manager::~Manager() {
      
    }

    void Manager::init() {
      HookManager.call( hookRegisterInputDevices );

      cfx::Node &N = global_s.Application->Settings >> "Controls" >> "Devices";
      Device *D;

      foreach_a( N.Nodes ) {
        cfx::Node &I = *N.Nodes[ a ];
        D = DeviceTypes.create( I.Name );
        ASSERT_CONTINUE( !D, "inputDeviceNotFound", "name", &I.Name );

        Devices.pushItem( D );
        D->init( I );
      }

      /*foreach_a( Records )
        printf( "input: %s \n", Records[ a ]->ID.toCChar() );*/
    }
    void Manager::destroy() {
      foreach_a( Records )
        delete Records[ a ];

      foreach_a( Devices )
        delete Devices[ a ];
    }
    void Manager::step() {
      if( global_s.Application->CurrentScene )
        global_s.Application->CurrentScene->Controls.step();

      foreach_a( Records )
        Records[ a ]->step();

      foreach_a( Devices )
        Devices[ a ]->onStep();
    }
    void Manager::handle( SDL_Event &Event ) {
      foreach_a( Devices ) {
        if( Devices[ a ]->onHandle( Event ) )
          break;
      }
    }

    InputRecord* Manager::find( const String &Name ) {
      foreach_a( Records ) {
        if( Records[ a ]->ID == Name )
          return Records[ a ];
      }

      return 0;
    }

    // Input
    Input::Input( cfx::Node &Source ) {
      StoreValue = false;

      Value = 0;
      PrevValue = 0;

      Link = global_s.Controls->find( Source.Value );
      ASSERT_RETURN( !Link, "inputNotFound", "input", &Source.Value, "bind", &Source.Name );

      InputRecord *R;

      foreach_a( Source.Arguments ) {
        String &S = ( Source || a ).Name;

        if( S == "store" )
          StoreValue = true;
        else {
          R = global_s.Controls->find( S );
          ASSERT_RETURN( !R, "inputNotFound", "input", &S, "bind", &Source.Name );
          Operators.pushItem( R );
        }
      }
    }
    Input::~Input() {
      
    }

    void Input::step() {
      if( !Link )
        return;

      PrevValue = Value;

      if( Link->Value == Link->PrevValue )
        return;

      bool R = true;
      if( Link->Handled ) {
        R = false;

      } else {
        foreach_a( Operators ) {
          if( Operators[ a ]->Value < 0.5f ) {
            R = false;
            break;
          }
        }
      }

      if( R ) {
        Link->Handled = true;
        Value = Link->Value;

      } else if( !StoreValue )
        Value = 0.0f;
    }
    void Input::reset() {
      PrevValue = 0;
      Value = 0;
    }

    bool Input::isPress() {
      return Value >= PressLevel && PrevValue < PressLevel;
    }
    bool Input::isRelease() {
      return Value < PressLevel && PrevValue >= PressLevel;
    }
    bool Input::isState() {
      return Value >= PressLevel;
    }

    // InputList
    InputList::InputList() {
          
    }
    InputList::~InputList() {
      foreach_a( List ) {
        foreach_b( *List[ a ] ) {
          delete List[ a ]->Ptr[ b ];
        }

        delete List[ a ];
      }
    }

    Input* InputList::create( const String &Name ) {
      cfx::Node &Source = global_s.Application->Settings >> "Controls" >> "Binds" >> Name;

      Input *I = new Input( Source );

      while( List.Size <= I->Operators.Size )
        List.pushItem( new IList() );

      List[ I->Operators.Size ]->pushItem( I );

      return I;
    }

    void InputList::step() {
      Container< IList* > &TList = global_s.Controls->GlobalControls.List;
      SizeT M1 = mMax< SizeT >( List.Size, TList.Size );

      for( SizeT a = M1; a > 0; a -- ) {
        if( List.Size >= a ) {
          foreach_b( *List[ a - 1 ] ) {
            List[ a - 1 ]->Ptr[ b ]->step();
          }
        }

        if( TList.Size >= a ) {
          foreach_b( *TList[ a - 1 ] ) {
            TList[ a - 1 ]->Ptr[ b ]->step();
          }
        }
      }
    }
    
  }

}

#endif