#pragma once

#include "Engine.Dedicated.h"
#include "Engine.String.h"
#include "Engine.Confix.h"
#include "Engine.TypeContainer.h"
#include "Engine.SDL.Header.h"

#define registerDevice( name ) global_s.Controls->DeviceTypes.add< name >( #name )

#ifdef MODULE_CLIENT

namespace app {
  
  namespace ctrl {

    class InputRecord {

    public:
      String ID;
      Float32 Value, PrevValue;
      bool Handled;

    public:
      InputRecord( const String &Name );
      ~InputRecord();

    public:
      void step();

    };

    class Device {

    public:
      String Name;

    public:
      Device();
      virtual ~Device();

    public:
      void init( cfx::Node &Source );

    public:
      virtual void onInit( cfx::Node &Source ) {};

      virtual void onStep() {};
      virtual bool onHandle( SDL_Event &Event ) = 0;

    };

    class Input {

    public:
      InputRecord *Link;
      Container< InputRecord* > Operators;
      bool StoreValue;
      Float32 Value, PrevValue;

    public:
      Input( cfx::Node &Source );
      ~Input();

    public:
      void step();
      void reset();

    public:
      bool isPress();
      bool isRelease();
      bool isState();

    };

    class InputList {

    public:
      typedef Container< Input* > IList;

    public:
      Container< IList* > List;

    public:
      InputList();
      ~InputList();

    public:
      Input* create( const String &Name );

      void step();

    };

    class Manager {

    public:
      Container< InputRecord* > Records;
      Container< Device* > Devices;
      TypeContainer< Device > DeviceTypes;

    public:
      InputList GlobalControls;

    public:
      Manager();
      ~Manager();

    public:
      void init();
      void destroy();
      void step();
      void handle( SDL_Event &Event );

      InputRecord* find( const String &Name );

    };

  }

}

#endif