#include "Engine.Cursor.h"
#include "Toolkit.Display.h"

namespace app {

  hook( hookAppInit ) {
    global_s.Cursor = new CursorData;
    global_s.Cursor->init();
  }
  hook( hookAppDestroy ) {
    global_s.Cursor->destroy();
    delete global_s.Cursor;
  }
  hook( hookAppPostStep ) {
    global_s.Cursor->step();
  }

  // IDraggedObject
  IDraggedObject::IDraggedObject() {

  }
  IDraggedObject::~IDraggedObject() {
    global_s.Cursor->DraggedObject = 0;
  }

  // CursorData
  CursorData::CursorData() {

  }
  CursorData::~CursorData() {
    if( DraggedObject )
      delete DraggedObject;
  }

  void CursorData::init() {
    ObjectOnCursor = 0;
    ClickedObject = 0;
    PrevClickedObject = 0;
    DraggedObject = 0;

    PrimaryInteractionPos = -1;

    PrimaryInteraction = global_s.Controls->GlobalControls.create( "PrimaryInteraction" );
    SecondaryInteraction = global_s.Controls->GlobalControls.create( "SecondaryInteraction" );

    ScrollUp = global_s.Controls->GlobalControls.create( "ScrollUp" );
    ScrollDown = global_s.Controls->GlobalControls.create( "ScrollDown" );
  }
  void CursorData::destroy() {

  }
  void CursorData::step() {
    RelPos = Pos - PrevPos;
    PrevPos = Pos;

    PrevClickedObject = ClickedObject;

    if( ObjectOnCursorStack.Size )
      ObjectOnCursor = ObjectOnCursorStack[ ObjectOnCursorStack.Size - 1 ];
    else
      ObjectOnCursor = 0;

    for( SizeT a = ObjectOnCursorStack.Size; a-- != 0; ) {
      if( ObjectOnCursorStack[ a ]->onCursorStackStep() ) break;
    }

    if( PrimaryInteraction->isPress() ) {
      PrimaryInteractionPos = Pos;

    } else if ( PrimaryInteraction->isRelease() ) {
      PrimaryInteractionPos = -1;
      ClickedObject = 0;

      if( DraggedObject ) {

        for( SizeT a = ObjectOnCursorStack.Size; a-- != 0; ) {
          if( ObjectOnCursorStack[ a ]->onDragObjectDrop( *DraggedObject ) ) break;
        }

        delete DraggedObject;   
      }
    }

    ObjectOnCursorStack.clear();
  }
  void CursorData::draw() {
    if( DraggedObject )
      DraggedObject->onDraw();
  }

  // CursorTarget
  interfaceRoot( CursorTarget, );

}