#include "Engine.Dedicated.h"

#ifdef MODULE_CLIENT

#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.Global.h"
#include "Engine.Variables.h"
#include "Engine.Exception.h"
#include "Engine.OpenGL.h"
#include "Engine.Controls.h"

#include <SDL/SDL.h>

namespace app {

  hook( hookAppInit ) {

    global_s.Display = new RenderWindow();
    global_s.Display->init();

  }
  hook( hookAppDestroy ) {

    global_s.Display->destroy();
    delete global_s.Display;

  }
  hook( hookAppStep ) {

    global_s.Display->step();

  }
  hook( hookException ) {
    SDL_Quit();
  }

  // RenderWindow
  void RenderWindow::init() {
    linkVariable( "Graphics" > "Resolution",    Resolution,     Vec2Screen  );
    linkVariable( "Graphics" > "Position",      Position,       Vec2Screen  );
    linkVariable( "Graphics" > "Fullscreen",    Fullscren,      bool        );
    linkVariable( "Graphics" > "Console",       Console,        bool        );
    linkVariable( "Graphics" > "VSync",         VSync,          bool        );
    linkVariable( "Graphics" > "AntiAliasing",  AntiAliasing,   UInt8       );

    linkEngineVariable( "Texture" > "MaxTextureSize",    MaxTextureSize,    TypeScreen );

    ASSERT_THROW( Resolution.X < 10 || Resolution.Y < 10, "invalidResolution", "W", String().pushInt( Resolution.X ).istr(), "H", String().pushInt( Resolution.Y ).istr() );

    PreviousResolution = Resolution;

    Flags = SDL_OPENGL |  SDL_RESIZABLE | SDL_HWSURFACE | SDL_ANYFORMAT | SDL_DOUBLEBUF;

    if ( Fullscren )
      Flags |= SDL_FULLSCREEN;

    char * Cap = global_s.Application->WindowCaption.toCChar();
    /*Window = SDL_CreateWindow( Cap, ( int ) Position.X, ( int) Position.Y, ( int ) Resolution.X, ( int ) Resolution.Y, Flags );
    delete Cap;
    ASSERT_THROW( !Window, "windowCreateFail", "error", &IString( SDL_GetError() ) );

    Context = SDL_GL_CreateContext( Window );
    ASSERT_THROW( !Window, "windowCreateFail", "error", &IString( SDL_GetError() ) );*/

    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 8 );

    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
    SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 32 );

    SDL_GL_SetAttribute( SDL_GL_ACCUM_RED_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_ACCUM_GREEN_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_ACCUM_BLUE_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_ACCUM_ALPHA_SIZE, 8 );

    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    SDL_GL_SetAttribute( SDL_GL_MULTISAMPLEBUFFERS, 1 );
    SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, AntiAliasing );


    Window = SDL_SetVideoMode( ( int ) Resolution.X, ( int ) Resolution.Y, 0, Flags );

    SDL_WM_SetCaption( Cap, 0 );
    free( Cap );

    Flags = IMG_INIT_JPG | IMG_INIT_PNG;
    ASSERT_THROW( IMG_Init( Flags ) != Flags, "SDL_imageInitFail", "error", &IString( IMG_GetError() ) );

    ASSERT_THROW( TTF_Init() != 0, "SDL_ttfInitFail", "error", &IString( TTF_GetError() ) );

    ASSERT_THROW( !Window, "windowCreateFail", "error", &IString( SDL_GetError() ) );

    GLenum GlErr = glewInit();
	  ASSERT_THROW( GlErr != GLEW_OK, "glewInitFail", "error", &IString( ( const char* ) glewGetErrorString( GlErr ) ) );

    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glDepthFunc( GL_LEQUAL );

    SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 1 );

    updateSettings();

    if( !glewIsSupported( "GL_VERSION_2_0" ) )
      printf( "WARNING: OpenGL 2.0 not available, for advanced rendering using ARB functions when available... \n" );

    HookManager.call( hookPostSDL_OPENGLInit );
  }
  void RenderWindow::destroy() {
    /*SDL_GL_DeleteContext( Context );
    SDL_DestroyWindow( Window );*/

    IMG_Quit();
    TTF_Quit();
  }
  void RenderWindow::step() {
    while ( SDL_PollEvent( &Event ) ) {

      switch( Event.type ) {

      case SDL_QUIT:
        global_s.Application->stop();
        break;

      case SDL_VIDEORESIZE:
        Resolution.assign( Event.resize.w, Event.resize.h );
        updateResolution();
        break;

      default:
        global_s.Controls->handle( Event );
        break;

      }

    }

    if ( !global_s.Application->CurrentScene )
      return;

    glClearColor( 0.1f, 0.1f, 0.1f, 1.0f );
    glClear( GL_COLOR_BUFFER_BIT );
    glCalls( 2 );
    glBeginFrame();

    if ( global_s.Application->CurrentScene->Flags & Scene::using3DDraw ) {
      ASSERT_THROW( true, "TODO: 3D DRAW" );

      //global_s.Application->CurrentScene->onDraw3D();
      // Let's skip this now
    }

    if ( global_s.Application->CurrentScene->Flags & Scene::using2DDraw ) {
      glLoadIdentity();
      glMatrixMode( GL_PROJECTION );
      glLoadIdentity();
      glOrtho( 0, Resolution.X, Resolution.Y, 0, 1, -1 );
      glMatrixMode( GL_MODELVIEW );
      glCalls( 5 );

      global_s.Application->CurrentScene->onDraw2D();
      global_s.Application->CurrentScene->GUI.draw();
      global_s.Application->CurrentScene->onPostDraw2D();
    }

    global_s.Cursor->draw();

    //SDL_GL_SwapWindow( Window );
    glEndFrame();

    SDL_GL_SwapBuffers();
  }
  void RenderWindow::parseWindowEvent() {
    /*switch( Event.window.event ) {
    
    case SDL_WINDOWEVENT_MOVED:
      Position.assign( Event.window.data1, Event.window.data2 );
      break;

    case SDL_WINDOWEVENT_RESIZED:
      Resolution.assign( Event.window.data1, Event.window.data2 );
      updateResolution();
      break;

    case SDL_WINDOWEVENT_CLOSE:
      global_s.Application->stop();
      break;

    }*/
  }

  void RenderWindow::updateSettings() {
    if ( global_s.Display->Console )
      consoleShow();
    else
      consoleHide();

    if( AntiAliasing > 0 ) {
      glEnable( GL_MULTISAMPLE );
      glEnable( GL_POLYGON_SMOOTH );
      glEnable( GL_LINE_SMOOTH );

      glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
      glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST );

      //SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, AntiAliasing );
    } else {
      glDisable( GL_MULTISAMPLE );
      glDisable( GL_POLYGON_SMOOTH );
      glDisable( GL_LINE_SMOOTH );

    }

    typedef void (APIENTRY * WGLSWAPINTERVALEXT) (int);

    WGLSWAPINTERVALEXT wglSwapIntervalEXT = (WGLSWAPINTERVALEXT) 
      wglGetProcAddress("wglSwapIntervalEXT");
    if (wglSwapIntervalEXT)
    {
      wglSwapIntervalEXT( VSync ); 
    }
  }
  void RenderWindow::updateResolution() {
    glViewport( 0, 0, Resolution.X, Resolution.Y );

    HookManager.call( hookResolutionChange );
    PreviousResolution = Resolution;
  }


}


#endif