#include "Engine.FrameBuffer.h"
#include "Engine.Global.h"

namespace app {

  static const GLenum eColorAttachments[] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3 };

  FrameBuffer::FrameBuffer() {

  }
  FrameBuffer::~FrameBuffer() {
  
  }

  void FrameBuffer::create( const Vec2Screen Sz, UInt8 Att /* = atDepthBuffer */ ) {
    if( glewIsSupported( "GL_VERSION_2_0" ) )
      ARB = false;
    else if( GLEW_EXT_framebuffer_object )
      ARB = true;
    else
      ASSERT_THROW( true, "noFBOSupport" );

    Size = Sz;
    Attachments = Att;
    DrawBuffers = 0;

    if( ARB ) {
      glGenFramebuffersEXT( 1, &ID );

      bind();

      if ( Attachments & atDepthBuffer ) {
        glGenRenderbuffersEXT( 1, &DepthBuffer );

        glBindRenderbufferEXT( GL_RENDERBUFFER, DepthBuffer );
        glRenderbufferStorageEXT( GL_RENDERBUFFER, GL_DEPTH_COMPONENT, ( GLsizei ) Size.X, ( GLsizei ) Size.Y );
        glBindRenderbufferEXT( GL_RENDERBUFFER, 0 );

        glFramebufferRenderbufferEXT( GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, DepthBuffer );

        glCalls( 5 );
      }

    } else {
      glGenFramebuffers( 1, &ID );

      bind();

      if ( Attachments & atDepthBuffer ) {
        glGenRenderbuffers( 1, &DepthBuffer );

        glBindRenderbuffer( GL_RENDERBUFFER, DepthBuffer );
        glRenderbufferStorage( GL_RENDERBUFFER, GL_DEPTH_COMPONENT, ( GLsizei ) Size.X, ( GLsizei ) Size.Y );
        glBindRenderbuffer( GL_RENDERBUFFER, 0 );

        glFramebufferRenderbuffer( GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, DepthBuffer );

        glCalls( 5 );
      }
    }

    unbind();

    glCalls( 1 );
  }
  void FrameBuffer::destroy() {
    if( ARB ) {
      glDeleteFramebuffersEXT( 1, &ID );

      if ( Attachments & atDepthBuffer ) {
        glDeleteRenderbuffersEXT( 1, &DepthBuffer );

        glCalls( 1 );
      }

    } else {
      glDeleteFramebuffers( 1, &ID );

      if ( Attachments & atDepthBuffer ) {
        glDeleteRenderbuffers( 1, &DepthBuffer );

        glCalls( 1 );
      }
    }

    glCalls( 1 );
  }

  void FrameBuffer::bind() {
    if( ARB ) {
      glBindFramebufferEXT( GL_DRAW_FRAMEBUFFER, ID );

      glDrawBuffersARB( DrawBuffers, eColorAttachments );

    } else {
      glBindFramebuffer( GL_DRAW_FRAMEBUFFER, ID );

      glDrawBuffers( DrawBuffers, eColorAttachments );
    }

    glViewport( 0, 0, Size.X, Size.Y );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0.0, Size.X, 0.0, Size.Y, -1.0, 1.0 ); 
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();


    glCalls( 8 );

    if( QueueBinds.Size ) {
      foreach_a( QueueBinds ) {
        if( QueueBinds[ a ].Tex->Tex_OGL != -1 ) {     
          if( ARB )
            glFramebufferTexture2DEXT( GL_FRAMEBUFFER, eColorAttachments[ QueueBinds[ a ].Pos ], GL_TEXTURE_2D, QueueBinds[ a ].Tex->Tex_OGL, 0 );
          else
            glFramebufferTexture2D( GL_FRAMEBUFFER, eColorAttachments[ QueueBinds[ a ].Pos ], GL_TEXTURE_2D, QueueBinds[ a ].Tex->Tex_OGL, 0 );
          glCalls( 1 );

          QueueBinds.removeItemPos( a );
          a --;
        }

      }
    }
  }
  void FrameBuffer::unbind() {
    if( ARB )
      glBindFramebufferEXT( GL_DRAW_FRAMEBUFFER, 0 );
    else
      glBindFramebuffer( GL_DRAW_FRAMEBUFFER, 0 );

    glViewport( 0, 0, global_s.Display->Resolution.X, global_s.Display->Resolution.Y );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho( 0, global_s.Display->Resolution.X, global_s.Display->Resolution.Y, 0, 1, -1 );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
   
    glCalls( 7 );
  }

  void FrameBuffer::bindTexture( res::Texture *Tex, UInt8 Pos /* = 0 */ ) {
    ASSERT_RETURN( Tex->Size != Size, "framebufferTextureSizeMisMatch" );

    if ( DrawBuffers <= Pos ) 
      DrawBuffers = Pos + 1;

    if( Tex->Tex_OGL != -1 ) {
      bind();
      if( ARB )
        glFramebufferTexture2DEXT( GL_FRAMEBUFFER, eColorAttachments[ Pos ], GL_TEXTURE_2D, Tex->Tex_OGL, 0 );
      else
        glFramebufferTexture2D( GL_FRAMEBUFFER, eColorAttachments[ Pos ], GL_TEXTURE_2D, Tex->Tex_OGL, 0 );
      unbind();
      glCalls( 1 );

    } else {
      QueueBind Q;
      Q.Tex = Tex;
      Q.Pos = Pos;
      QueueBinds.pushItem( Q );
    }

  }

}