#pragma once

#include "Engine.Dedicated.h"
#include "Engine.Container.h"
#include "Engine.GLQuad.h"
#include "Engine.OpenGL.Header.h"
#include "Engine.Vector.h"
#include "Engine.Color.h"

#ifdef MODULE_CLIENT

namespace app {

  extern UInt8 GL_TexArray;

  template< typename T >
  class GLBuffer2 {

  public:
    T *Ptr;
    SizeT Size, Capacity;

  public:
    GLBuffer2() {
      Size = 0;
      Capacity = 16;
      Ptr = ( T* ) malloc( sizeof( T ) * Capacity );
    }
    ~GLBuffer2() {
      free( Ptr );
    }

  public:
    void allocate( SizeT Cap ) {
      Cap += Size;

      if ( Capacity >= Cap )
        return;

      while( Capacity < Cap )
        Capacity *= 2;

      Ptr = ( T* ) realloc( Ptr, sizeof( T ) * Capacity );
    }
    void clear() {
      Size = 0;
    }

  public:
    //void upload();
    void uploadVertex();
    void uploadTexture();
    void uploadColor();

  public:
    void pushRect( Vec2< T > Pos, Vec2< T > ISize ) {
      allocate( 12 );

      Ptr[ Size++ ] = Pos.X;
      Ptr[ Size++ ] = Pos.Y;

      Ptr[ Size++ ] = Pos.X + ISize.X;
      Ptr[ Size++ ] = Pos.Y;

      Ptr[ Size++ ] = Pos.X + ISize.X;
      Ptr[ Size++ ] = Pos.Y + ISize.Y;


      Ptr[ Size++ ] = Pos.X + ISize.X;
      Ptr[ Size++ ] = Pos.Y + ISize.Y;

      Ptr[ Size++ ] = Pos.X;
      Ptr[ Size++ ] = Pos.Y + ISize.Y;

      Ptr[ Size++ ] = Pos.X;
      Ptr[ Size++ ] = Pos.Y;
    }
    void pushRectRotated4Abs( Vec2< T > Pos, Vec2< T > ISize, UInt8 Rot ) {
      allocate( 12 );

      switch( Rot ) {
      
      case 0:
        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.X;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.X;
        Ptr[ Size++ ] = Pos.Y + ISize.Y;


        Ptr[ Size++ ] = Pos.X + ISize.X;
        Ptr[ Size++ ] = Pos.Y + ISize.Y;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y + ISize.Y;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y;
        break;

      case 1:
        Ptr[ Size++ ] = Pos.X + ISize.Y;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.Y;
        Ptr[ Size++ ] = Pos.Y + ISize.X;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y + ISize.X;


        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y + ISize.X;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.Y;
        Ptr[ Size++ ] = Pos.Y;
        break;

      case 2:
        Ptr[ Size++ ] = Pos.X + ISize.X;
        Ptr[ Size++ ] = Pos.Y + ISize.Y;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y + ISize.Y;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y;


        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.X;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.X;
        Ptr[ Size++ ] = Pos.Y + ISize.Y;
        break;

      case 3:
        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y + ISize.X;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.Y;
        Ptr[ Size++ ] = Pos.Y;


        Ptr[ Size++ ] = Pos.X + ISize.Y;
        Ptr[ Size++ ] = Pos.Y;

        Ptr[ Size++ ] = Pos.X + ISize.Y;
        Ptr[ Size++ ] = Pos.Y + ISize.X;

        Ptr[ Size++ ] = Pos.X;
        Ptr[ Size++ ] = Pos.Y + ISize.X;
        break;

      default:
        ASSERT_THROW( true, "invalidRotationValue" );
        break;
      
      }
    }

    void pushQuad( const GLQuad2< T > &Quad ) {
      allocate( 12 );
      memcpy( Ptr + Size, Quad.Ptr, sizeof( T ) * 12 );
      Size += 12;
    }

    void pushStripRect( const Vec2< T > &Pos, const Vec2< T > &ISize ) {
      allocate( 8 );

      Ptr[ Size++ ] = Pos.X + ISize.X;
      Ptr[ Size++ ] = Pos.Y;

      Ptr[ Size++ ] = Pos.X;
      Ptr[ Size++ ] = Pos.Y;

      Ptr[ Size++ ] = Pos.X + ISize.X;
      Ptr[ Size++ ] = Pos.Y + ISize.Y;

      Ptr[ Size++ ] = Pos.X;
      Ptr[ Size++ ] = Pos.Y + ISize.Y;
    }

    void pushLineRect( const Vec2< T > &Pos, const Vec2< T > &ISize ) {
      pushRect( Pos + Vec2< T >( 1, 0 ),            Vec2< T >( ISize.X - 1, 1 ) );
      pushRect( Pos + Vec2< T >( ISize.X - 1, 1 ),  Vec2< T >( 1, ISize.Y - 1 ) );
      pushRect( Pos + Vec2< T >( 0, ISize.Y - 1 ),  Vec2< T >( ISize.X - 1, 1 ) );
      pushRect( Pos,                                Vec2< T >( 1, ISize.Y - 1 ) );
    }

    void pushPoint( const Vec2< T > Pos ) {
      allocate( 2 );

      Ptr[ Size++ ] = Pos.X;
      Ptr[ Size++ ] = Pos.Y;
    }

    void pushColorRect( Color C ) {
      allocate( 24 );

      for( int a = 0; a < 6; a ++ ) {
        Ptr[ Size++ ] = C.R;
        Ptr[ Size++ ] = C.G;
        Ptr[ Size++ ] = C.B;
        Ptr[ Size++ ] = C.A;
      }
    }
    void pushColorNTimes( Color C, SizeT N ) {
      allocate( N * 4 );

      while( N-- ) {
        Ptr[ Size++ ] = C.R;
        Ptr[ Size++ ] = C.G;
        Ptr[ Size++ ] = C.B;
        Ptr[ Size++ ] = C.A;
      }
    }

    void pushBuffer( GLBuffer2< T > &Buf ) {
      allocate( Buf.Size );

      memcpy( Ptr + Size, Buf.Ptr, Buf.Size );
      Size += Buf.Size;
    }

  public:
    void rotateLastNPoints( SizeT N, const Vec2< T > Center, Float32 Angle ) {
      Float32 Sin = sin( Angle ), Cos = cos( Angle );
      Vec2< Float32 > L;

      for( SizeT a = Size - N * 2; a < Size; a += 2 ) {
        L.assign< T >( Vec2< T >( Ptr[ a ] - Center.X, Ptr[ a + 1 ] - Center.Y ) );
        Ptr[ a ] = Center.X + ( T ) mRound< Float32 >( L.X * Cos - L.Y * Sin );
        Ptr[ a + 1 ] = Center.Y + ( T ) mRound< Float32 >( L.X * Sin + L.Y * Cos );
      }

    }
    void rotateLastNPoints4( SizeT N, const Vec2< T > Center, UInt8 Angle ) {
      if( Angle == 0 )
        return;
      else if( Angle == 1 ) {

        Vec2< T > L;
        for( SizeT a = Size - N * 2; a < Size; a += 2 ) {
          L = Vec2Screen( Ptr[ a ], Ptr[ a + 1 ] ) - Center;
          Ptr[ a ] = Center.X - L.Y;
          Ptr[ a + 1 ] = Center.Y + L.X;
        }

      } else if( Angle == 2 ) {

        Vec2< T > L;
        for( SizeT a = Size - N * 2; a < Size; a += 2 ) {
          L = Vec2Screen( Ptr[ a ], Ptr[ a + 1 ] ) - Center;
          Ptr[ a ] = Center.X - L.X;
          Ptr[ a + 1 ] = Center.Y - L.Y;
        }

      } else if( Angle == 3 ) {

        Vec2< T > L;
        for( SizeT a = Size - N * 2; a < Size; a += 2 ) {
          L = Vec2Screen( Ptr[ a ], Ptr[ a + 1 ] ) - Center;
          Ptr[ a ] = Center.X + L.Y;
          Ptr[ a + 1 ] = Center.Y - L.X;
        }

      }
    }

    void cycleLastNTriangles2D( SizeT N ) {
      Vec2< T > Buf;
      for( SizeT a = Size - N * 6; a < Size; a += 6 ) {
        Buf.assign( Ptr[ a ], Ptr[ a + 1 ] );

        Ptr[ a ] = Ptr[ a + 2 ];
        Ptr[ a + 1 ] = Ptr[ a + 3 ];

        Ptr[ a + 2 ] = Ptr[ a + 4 ];
        Ptr[ a + 3 ] = Ptr[ a + 5 ];

        Ptr[ a + 4 ] = Buf.X;
        Ptr[ a + 5 ] = Buf.Y;
      }
    }

  };

  typedef GLBuffer2< TypeScreen > GLBuffer2Screen;
  typedef GLBuffer2< TypeTex > GLBuffer2Tex;
  typedef GLBuffer2< UInt8 > GLBuffer2Col;


}

#endif
