#pragma once

#include "Engine.Vector.h"
#include "Engine.Math.h"

namespace app {

  template< typename T >
  class GLQuad2 {

  public:
    T Data[ 12 ];

  public:
    GLQuad2() {

    }
    GLQuad2( const Vec2< T > &Pos, const Vec2< T > &Size ) {
      Data[ 0 ] = Pos.X;
      Data[ 1 ] = Pos.Y;

      Data[ 2 ] = Pos.X + Size.X;
      Data[ 3 ] = Pos.Y;

      Data[ 4 ] = Pos.X + Size.X;
      Data[ 5 ] = Pos.Y + Size.Y;


      Data[ 6 ] = Pos.X + Size.X;
      Data[ 7 ] = Pos.Y + Size.Y;

      Data[ 8 ] = Pos.X;
      Data[ 9 ] = Pos.Y + Size.Y;

      Data[ 10 ] = Pos.X;
      Data[ 11 ] = Pos.Y;
    }
    
    ~GLQuad2() {

    }

  public:
    void assign( const Vec2< T > &Pos, const Vec2< T > &Size ) {
      Data[ 0 ] = Pos.X;
      Data[ 1 ] = Pos.Y;

      Data[ 2 ] = Pos.X + Size.X;
      Data[ 3 ] = Pos.Y;

      Data[ 4 ] = Pos.X + Size.X;
      Data[ 5 ] = Pos.Y + Size.Y;


      Data[ 6 ] = Pos.X + Size.X;
      Data[ 7 ] = Pos.Y + Size.Y;

      Data[ 8 ] = Pos.X;
      Data[ 9 ] = Pos.Y + Size.Y;

      Data[ 10 ] = Pos.X;
      Data[ 11 ] = Pos.Y;
    }

  };

  template< typename T >
  class GLQuad2Strip {

  public:
    T Data[ 8 ];

  public:
    GLQuad2Strip() {

    }
    GLQuad2Strip( const Vec2< T > &Pos, const Vec2< T > &Size ) {
      Data[ 0 ] = Pos.X + Size.X;
      Data[ 1 ] = Pos.Y;
      
      Data[ 2 ] = Pos.X;
      Data[ 3 ] = Pos.Y;

      Data[ 4 ] = Pos.X + Size.X;
      Data[ 5 ] = Pos.Y + Size.Y;

      Data[ 6 ] = Pos.X;
      Data[ 7 ] = Pos.Y + Size.Y;
    }

  public:
    void assign( const Vec2< T > &Pos, const Vec2< T > &Size ) {
      Data[ 0 ] = Pos.X + Size.X;
      Data[ 1 ] = Pos.Y;

      Data[ 2 ] = Pos.X;
      Data[ 3 ] = Pos.Y;

      Data[ 4 ] = Pos.X + Size.X;
      Data[ 5 ] = Pos.Y + Size.Y;

      Data[ 6 ] = Pos.X;
      Data[ 7 ] = Pos.Y + Size.Y;
    }
    void assign( const Vec2< T > LT, const Vec2< T > RT, const Vec2< T > LB, const Vec2< T > RB ) {
      Data[ 0 ] = RT.X;
      Data[ 1 ] = RT.Y;

      Data[ 2 ] = LT.X;
      Data[ 3 ] = LT.Y;

      Data[ 4 ] = RB.X;
      Data[ 5 ] = RB.Y;

      Data[ 6 ] = LB.X;
      Data[ 7 ] = LB.Y;
    }

    void rotate( const Vec2< T > Center, Float64 Angle ) {
      Float64 Sin = sin( Angle ), Cos = cos( Angle );
      Vec2< T > L;

      L.assign( Data[ 0 ], Data[ 1 ] );
      Data[ 0 ] = Center.X + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Cos - ( Float64 )( L.Y - Center.Y ) * Sin );
      Data[ 1 ] = Center.Y + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Sin + ( Float64 )( L.Y - Center.Y ) * Cos );

      L.assign( Data[ 2 ], Data[ 3 ] );
      Data[ 2 ] = Center.X + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Cos - ( Float64 )( L.Y - Center.Y ) * Sin );
      Data[ 3 ] = Center.Y + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Sin + ( Float64 )( L.Y - Center.Y ) * Cos );

      L.assign( Data[ 4 ], Data[ 5 ] );
      Data[ 4 ] = Center.X + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Cos - ( Float64 )( L.Y - Center.Y ) * Sin );
      Data[ 5 ] = Center.Y + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Sin + ( Float64 )( L.Y - Center.Y ) * Cos );

      L.assign( Data[ 6 ], Data[ 7 ] );
      Data[ 6 ] = Center.X + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Cos - ( Float64 )( L.Y - Center.Y ) * Sin );
      Data[ 7 ] = Center.Y + ( T ) mRound< Float64 >( ( Float64 )( L.X - Center.X ) * Sin + ( Float64 )( L.Y - Center.Y ) * Cos );
    }

  };

  typedef GLQuad2< TypeScreen > GLQuad2Screen;
  typedef GLQuad2< TypeTex > GLQuad2Tex;

  typedef GLQuad2Strip< TypeScreen > GLQuad2StripScreen;
  typedef GLQuad2Strip< TypeTex > GLQuad2StripTex;

}