#include "Engine.GUI.Skin.h"
#include "Engine.String.Cast.h"
#include "Engine.Hooks.h"
#include "Engine.Global.h"

#include "Resource.Font.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookAppInit ) {
      HookManager.call( hookRegisterGUISkins );

      Skin::loadSkins();
    }
    hook( hookAppDestroy ) {
      Skin::unloadSkins();
    }

    // Skin
    Skin::Skin() {

    }
    Skin::~Skin() {

    }

    Skin* Skin::get( const IString &Name, Skin *Default /* = 0 */ ) {
      Container< Skin* > &SkinList = global_s.gui.SkinList;

      foreach_a( SkinList ) {
        if( SkinList[ a ]->Name == Name )
          return SkinList[ a ];
      }

      if( Default )
        return Default;

      return SkinList[ 0 ];
    }

    void Skin::loadSkins() {
      cfx::Document Fl;
       
      Skin* Sk;
      Fl.load( "../cfg/gui/skins.cfx" );

      foreach_a( Fl.Nodes ) {
        Sk = global_s.gui.SkinTypeList.create( Fl.Nodes[ a ]->Value );
        ASSERT_CONTINUE( !Sk, "skinTypeNotFound", "typeName", Fl.Nodes[ a ]->Value.istr() );

        cfx::Document D;
        D.load( String().pushCChar( "../cfg/gui/skin/" ).pushString( Fl.Nodes[ a ]->Name ).pushCChar( ".skin" ) );

        Sk->Name = Fl.Nodes[ a ]->Name;
        Sk->init( D );
        global_s.gui.SkinList.pushItem( Sk );
      }

      ASSERT_THROW( !global_s.gui.SkinList.Size, "noSkinLoaded" );
    }
    void Skin::unloadSkins() {
      foreach_a( global_s.gui.SkinList )
        delete global_s.gui.SkinList[ a ];

      global_s.gui.SkinList.clearUnalloc();
      global_s.gui.SkinTypeList.clear();
    }

    void Skin::init( cfx::Node &Source ) {
      // Button data
      stringTo< Vec2Screen >( Source > "DefaultButtonSize", DefaultButtonSize ); 

      // ListBox data
      cfx::Node &LB = Source >> "ListBox";
      ListBoxScrollBarSize = ( LB > "ScrollBarSize" ).toInt();
      ListBoxScrollBarSizeHover = ( LB > "ScrollBarSizeHover" ).toInt();

      ListBoxItemSize = ( LB > "ItemSize" ).toInt();

      ListBoxButtonSpeed = ( LB > "ButtonSpeed" ).toFloat();
      ListBoxWheelSpeed = ( LB > "WheelSpeed" ).toFloat();

      // Window data
      cfx::Node &Wnd = Source >> "Window";

      cfx::Node &WndHd = Wnd >> "Header";

      WindowHeaderSize = ( WndHd > "Size" ).toInt();
      WindowHeaderCaptionOffset = ( WndHd > "CaptionOffset" ).toInt();

      stringTo< Vec2Screen >( WndHd > "ButtonPos", WindowHeaderButtonPos ); 
      stringTo< Vec2Screen >( WndHd > "ButtonSize", WindowHeaderButtonSize ); 
      WindowHeaderButtonCaption = WndHd > "ButtonCaption";

      // Slider data
      cfx::Node &Sld = Source >> "Slider";

      SliderBarSize = ( Sld > "BarSize" ).toInt();
      stringTo< Vec2Screen >( Sld > "SliderSize", SliderSliderSize );

      // ComboBox data
      cfx::Node &Cbb = Source >> "ComboBox";
      ComboBoxMaxListBoxSize = ( Cbb > "MaxListBoxSize" ).toInt();
      ComboBoxMinListBoxSize = ( Cbb > "MinListBoxSize" ).toInt();
    }

    void Skin::drawButton( Vec2Screen Pos, Vec2Screen Size, const IString &Caption, ElementState_e State ) {
      drawPanel( Pos, Size, ptButton, State );
      drawLabel( Pos + Size / 2, Caption, res::Font::valCenter | res::Font::halCenter );
    }
    void Skin::drawProgressBar( Vec2Screen Pos, Vec2Screen Size, Float32 Value, Float32 Max, UInt8 Flags /* = 0 */ ) {
      drawPanel( Pos, Size, ptProgressBarBg, esNormal );

      if( Flags & pgfVertical ) {
        if( Value > 0 )
          drawPanel( Pos, Vec2Screen( ( TypeScreen )( Size.X, Value / Max * ( Float32 ) Size.Y ) ), ptProgressBarFg, esNormal );
      } else {
        if( Value > 0 )
          drawPanel( Pos, Vec2Screen( ( TypeScreen )( Value / Max * ( Float32 ) Size.X ), Size.Y ), ptProgressBarFg, esNormal );

        if( ( Flags & ( pgfValueLabel | pgfPercentLabel ) ) == ( pgfValueLabel | pgfPercentLabel ) )
          drawLabel( Pos + Size / 2, String().pushFloat( Value ).pushCChar( "/" ).pushFloat( Max ).pushCChar( " (" ).pushFloat( Value / Max * 100 ).pushCChar( "%)" ), res::Font::halCenter | res::Font::valCenter );
        else if( Flags & pgfValueLabel )
          drawLabel( Pos + Size / 2, String().pushFloat( Value ).pushCChar( "/" ).pushFloat( Max ), res::Font::halCenter | res::Font::valCenter );
        else if( Flags & pgfPercentLabel )
          drawLabel( Pos + Size / 2, String().pushFloat( Value / Max * 100 ).pushCChar( "%" ), res::Font::halCenter | res::Font::valCenter );

      }
    }

  }

}

#endif