#pragma once

#include "Engine.Dedicated.h"
#include "Engine.String.h"
#include "Engine.Confix.h"
#include "Engine.Vector.h"
#include "Engine.Color.h"

#include "Engine.GUI.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    class Element;

    class Skin {

    public:
      enum PanelType_e {
        ptButton,

        ptScrollBarButton,
        ptScrollBarBar,
        ptScrollBarBack,

        ptListBox,
        ptListBoxItem,

        ptPanel,

        ptImage,

        ptProgressBarBg,
        ptProgressBarFg,

        ptWindow,
        ptWindowHeader,
        ptWindowCloseButton,

        ptSliderBar,
        ptSliderSlider,

        ptComboBox,

        _ptLast
      };
      enum ProgressBarFlags_e {
        pgfPercentLabel   = 1 << 0,
        pgfValueLabel     = 1 << 1,
        pgfVertical       = 1 << 2,
      };

    public:
      String Name;

    public:
      Vec2Screen DefaultButtonSize;

      TypeScreen ListBoxScrollBarSize, ListBoxScrollBarSizeHover, ListBoxItemSize;
      Float32 ListBoxButtonSpeed, ListBoxWheelSpeed;

      TypeScreen WindowHeaderSize, WindowHeaderCaptionOffset;
      Vec2Screen WindowHeaderButtonPos, WindowHeaderButtonSize;
      String WindowHeaderButtonCaption;

      Vec2Screen SliderSliderSize;
      TypeScreen SliderBarSize;

      TypeScreen ComboBoxMaxListBoxSize, ComboBoxMinListBoxSize;

    public:
      Skin();
      virtual ~Skin();

    public:
      static Skin* get( const IString &Name, Skin *Default = 0 );

      static void loadSkins();
      static void unloadSkins();

    public:
      virtual void init( cfx::Node &Source );

    public:
      virtual Vec2Screen getLabelSize( const IString &Text ) = 0;

    public:
      virtual void drawPanel( Vec2Screen Pos, Vec2Screen Size, PanelType_e Panel, ElementState_e State, bool Vertical = false ) {};
      virtual void drawLabel( Vec2Screen Pos, const IString &Text, UInt8 Flags, Color Col = Color( 0, 0, 0, 0 ) ) = 0;

    public:
      virtual void drawButton( Vec2Screen Pos, Vec2Screen Size, const IString &Caption, ElementState_e State );
      virtual void drawProgressBar( Vec2Screen Pos, Vec2Screen Size, Float32 Value, Float32 Max, UInt8 Flags = 0 );

    };

  }

}

#endif