#include "Engine.GUI.h"
#include "Engine.String.Cast.h"
#include "Engine.Global.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"
#include "Engine.Application.h"

#include "GUI.Interface.Content.h"

#include "GUI.Root.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookAppInit ) {
      HookManager.call( hookRegisterGUIElements );
    }
    hook( hookResolutionChange ) {
      if( global_s.Application->CurrentScene )
        global_s.Application->CurrentScene->GUI.recalc();
    }

    // IElement
    IElement::IElement() {
      Flags = 0;
    }
    IElement::~IElement() {

    }

    void IElement::init( cfx::Node &Source ) {
      Flags = 0;
      cfx::Node *N;
      IString *S;

      // Name
      Name = Source.Value;

      // Pos
      N = &( Source >> "Position" );
      Pos.X = N->Value.toInt();
      Pos.Y = ( *N || ( SizeT ) 0 ).Name.toInt();

      // Size
      N = &( Source >> "Size" );
      Size.X = N->Value.toInt();
      if( N->Value.last() == '%' )
        Flags |= hszRelative;

      S = &( *N || (SizeT ) 0 ).Name;
      Size.Y = S->toInt();
      if( S->last() == '%' )
        Flags |= vszRelative;

      // Align

      N = &( Source >> "Align" );
      if( N->Value == "center" )
        Flags |= halCenter;
      else if( N->Value == "right" )
        Flags |= halRight;

      S = &( *N || ( SizeT ) 0 ).Name;
      if( *S == "center" )
        Flags |= valCenter;
      else if( *S == "bottom" )
        Flags |= valBottom;

      // Skin
      Skn = Skin::get( Source > "Skin" );

      // Enabled, Visible
      Enabled = ( Source > "Enabled" ) != "false";
      Visible = ( Source > "Visible" ) != "false";
    }
    IElement* IElement::find( const IString &El ) {
      if( Name == El )
        return this;
      else return 0;
    }

    void IElement::recalc() {

      // Size
      if( Flags & hszRelative )
        RealSize.X = ( Parent->RealSize.X * Size.X ) / 100;
      else
        RealSize.X = Size.X;

      if( Parent && RealSize.X < 0 )
        RealSize.X += Parent->RealSize.X;


      if( Flags & vszRelative )
        RealSize.Y = ( Parent->RealSize.Y * Size.Y ) / 100;
      else
        RealSize.Y = Size.Y;

      if( Parent && RealSize.Y < 0 )
        RealSize.Y += Parent->RealSize.Y;

      // Position
      if( Flags & halRight )
        RealPos.X = Pos.X + Parent->RealSize.X - RealSize.X;
      else if( Flags & halCenter )
        RealPos.X = Pos.X + ( Parent->RealSize.X - RealSize.X ) / 2;
      else
        RealPos.X = Pos.X;

      if( Flags & valBottom )
        RealPos.Y = Pos.Y + Parent->RealSize.Y - RealSize.Y;
      else if( Flags & valCenter )
        RealPos.Y = Pos.Y + ( Parent->RealSize.Y - RealSize.Y ) / 2;
      else
        RealPos.Y = Pos.Y;

    }
    bool IElement::onCursorStackStep() {
      if( global_s.Cursor->PrimaryInteraction->isPress() ) {
        global_s.Cursor->ClickedObject = this;
        focus();

        return true;
      }

      return false;
    };

    void IElement::draw() {
      if( global_s.Cursor->Pos.everyGreaterOrEqual( AbsolutePos ) && global_s.Cursor->Pos.everyLess( AbsolutePos + RealSize ) )
        global_s.Cursor->ObjectOnCursorStack.pushItem( this );

      if( !Enabled || !Visible )
        State = esDisabled;
      else if( global_s.Cursor->ClickedObject == this )
        State = esClicked;
      else if( global_s.Cursor->ObjectOnCursor == this && !global_s.Cursor->ClickedObject )
        State = esHover;
      else if( GRoot->FocusedElement == this )
        State = esFocused;
      else
        State = esNormal;
    }

    void IElement::focus() {
      GRoot->FocusedElement = this;
      Layer->focusLayer();
    }

    ElementState_e IElement::virtualSubElementState( Vec2Screen Ps, Vec2Screen Sz ) {

      if( State == esHover && global_s.Cursor->Pos.everyGreaterOrEqual( Ps ) && global_s.Cursor->Pos.everyLess( Ps + Sz ) )
         return esHover;
      else if( State == esClicked && global_s.Cursor->PrimaryInteractionPos.everyGreaterOrEqual( Ps ) && global_s.Cursor->PrimaryInteractionPos.everyLess( Ps + Sz ) )
        return esClicked;
      else if( !Enabled )
        return esDisabled;
      else
        return esNormal;
    }

    // Element
    interfaceRoot( Element,

      );

  }

}

#endif