#pragma once

#include "Engine.Dedicated.h"

#include "Engine.Vector.h"
#include "Engine.OpenGL.h"

namespace app {

  template< typename T >
  class Matrix {

  public:
    enum {
      Xx, Yx, Zx, Tx,
      Xy, Yy, Zy, Ty,
      Xz, Yz, Zz, Tz,
      Xt, Yt, Zt, Tt
    };

  public:
    T M[ 16 ];

  public:
    Matrix() {
      T M2[ 16 ] = { 1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1 };
      memcpy( M, M2, sizeof( T ) * 16 );
    }
    Matrix( T iXx, T iXy, T iXt, T iYx, T iYy, T iYt ) {
      T M2[ 16 ] = { iXx,iYx,0,0, iXy,iYy,0,0, 0,0,1,0, iXt,iYt,0,1 };
      memcpy( M, M2, sizeof( T ) * 16 );
    }

    ~Matrix() {

    }

  public:
    Vec2< T > apply( Vec2< T > Source ) {
      return Vec2< T >( Source.X * M[ Xx ] + Source.Y * M[ Xy ] + M[ Xt ], Source.X * M[ Yx ] + Source.Y * M[ Yy ] + M[ Yt ] );
    }
    Vec2< T > applyInverse2D( Vec2< T > Source ) {
      return inverse2D().apply( Source );
    }

    void applyMatrix2D( T Xx2, T Xy2, T Xt2, T Yx2, T Yy2, T Yt2 ) {
      T M2[ 16 ];
      memcpy( M2, M, sizeof( T ) * 16 );

      M[ Xx ] = Xx2 * M2[ Xx ] + Xy2 * M2[ Yx ];
      M[ Xy ] = Xx2 * M2[ Xy ] + Xy2 * M2[ Yy ];
      M[ Xt ] = Xx2 * M2[ Xt ] + Xy2 * M2[ Yt ] + Xt2;

      M[ Yx ] = Yx2 * M2[ Xx ] + Yy2 * M2[ Yx ];
      M[ Yy ] = Yx2 * M2[ Xy ] + Yy2 * M2[ Yy ];
      M[ Yt ] = Yx2 * M2[ Xt ] + Yy2 * M2[ Yt ] + Yt2;
    }

  public:
    void reset() {
      T M2[ 16 ] = { 1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1 };
      memcpy( M, M2, sizeof( T ) * 16 );
    }

    void translate2D( Vec2< T > Source ) {
      applyMatrix2D( 1,0,Source.X, 0,1,Source.Y );
    }
    void translate2DNegative( Vec2< T > Source ) {
      applyMatrix2D( 1,0,-Source.X, 0,1,-Source.Y );
    }

    void rotate2D( T Angle ) {
      T Sin = sin( Angle ), Cos = cos( Angle );
      applyMatrix2D( Cos,Sin,0, -Sin,Cos,0 );
    }
    void rotate2D( T Angle, Vec2< T > Center ) {
      translate2DNegative( Center );
      rotate2D( Angle );
      translate2D( Center );
    }

    void scale2D( T U ) {
      applyMatrix2D( U,0,0, 0,U,0 );
    }

  public:
    Matrix< T > inverse2D() {
      T K1 = M[ Xx ] * M[ Yy ] - M[ Xy ] * M[ Yx ];
      T K2 = M[ Yx ] * M[ Xy ] - M[ Xx ] * M[ Yy ];
      return Matrix< T >(
        M[ Yy ]/K1, -M[ Xy ]/K1, ( M[ Xy ]*M[ Yt ] - M[ Xt ]*M[ Yy ] )/K1,
        M[ Yx ]/K2, -M[ Xx ]/K2, ( M[ Yt ]*M[ Xx ] - M[ Yx ]*M[ Xt ] )/K2
        );
    }

#ifdef MODULE_CLIENT
    public:
      void use();
      void push() {
        glFinishDrawing();
        glPushMatrix();
        use();
        glCalls( 1 );
      }
      void pop() {
        glFinishDrawing();
        glPopMatrix();
        glCalls( 1 );
      }

#endif

  };

#ifdef MODULE_CLIENT

  void popMatrix();

#endif

  typedef Matrix< Float32 > Float32M;
  typedef Matrix< Float64 > Float64M;

}