#include "Toolkit.Display.h"
#include "Resource.Texture.h"

#ifdef MODULE_CLIENT

namespace app {

  // Data

  enum glDrawingMode_e {
    gldmColor,
    gldmTexture,
    gldmCustom,
    gldmNone
  };

  GLenum glPropertyValue[] = {
    GL_BLEND,
    GL_TEXTURE_2D,
    GL_CULL_FACE,
    GL_DEPTH_TEST,
    GL_COLOR_MATERIAL,
    GL_SCISSOR_TEST,

    GL_VERTEX_ARRAY,
    GL_TEXTURE_COORD_ARRAY,
    GL_COLOR_ARRAY,
  };
  bool glPropertyIsClientState[] = {
    false,
    false,
    false,
    false,
    false,
    false,

    true,
    true,
    true,
  };
  enum glPropertyState_e {
    gpsNone,
    gpsEnabled,
    gpsDisabled
  };

  glPropertyState_e glPropertyState[ _glLast ];

  // Variables

  GLBuffer2Screen glVertexBuffer;
  GLBuffer2Tex glTexCoordBuffer;
  GLBuffer2Col glColorBuffer;

  SizeT glCallCounter, glPrevCallCounter;
  GLuint glSelectedProgram, glSelectedTexture;

  glDrawingMode_e glDrawingMode = gldmCustom;

  Container< Rec2Screen > glScissorStack;

  // Core functions
  void glCalls( UInt8 Num ) {
    glCallCounter += Num;
  }
  SizeT glCalls() {
    return glPrevCallCounter;
  }
  void glFinishDrawing() {
    if( !glVertexBuffer.Size )
      return;

    if( glDrawingMode == gldmColor ) {
      glVertexBuffer.uploadVertex();
      glColorBuffer.uploadColor();

      glDrawArrays( GL_TRIANGLES, 0, glVertexBuffer.Size / 2 );

      glVertexBuffer.clear();
      glColorBuffer.clear();

    } else if( glDrawingMode == gldmTexture ) {
      if( glSelectedTexture == -1 )
        return;

      glVertexBuffer.uploadVertex();
      glTexCoordBuffer.uploadTexture();
      glColorBuffer.uploadColor();

      glDrawArrays( GL_TRIANGLES, 0, glVertexBuffer.Size / 2 );

      glVertexBuffer.clear();
      glTexCoordBuffer.clear();
      glColorBuffer.clear();

    } else if( glDrawingMode == gldmCustom ) {
      glVertexBuffer.clear();
      glTexCoordBuffer.clear();
      glColorBuffer.clear();

    }
  }

  void glOn( glProperty_e Property ) {
    if( glPropertyState[ Property ] == gpsEnabled )
      return;

    if( glPropertyIsClientState[ Property ] )
      glEnableClientState( glPropertyValue[ Property ] );
    else
      glEnable( glPropertyValue[ Property ] );
    glPropertyState[ Property ] = gpsEnabled;
  }
  void glOff( glProperty_e Property ) {
    if( glPropertyState[ Property ] == gpsDisabled )
      return;

    if( glPropertyIsClientState[ Property ] )
      glDisableClientState( glPropertyValue[ Property ] );
    else
      glDisable( glPropertyValue[ Property ] );
    glPropertyState[ Property ] = gpsDisabled;
  }

  void glProgram( GLuint Program ) {
    if ( glSelectedProgram == Program )
      return;

    if( glDrawingMode == gldmColor || glDrawingMode == gldmTexture )
      glFinishDrawing();

    glSelectedProgram = Program;
    glUseProgram( Program );
  }
  void glTexture( GLuint Texture ) {
    if ( glSelectedTexture == Texture )
      return;

    if( glDrawingMode == gldmTexture )
      glFinishDrawing();

    glSelectedTexture = Texture;
    glBindTexture( GL_TEXTURE_2D, Texture );
  }

  void glPushScissor( Vec2Screen Pos, Vec2Screen Size ) {
    glFinishDrawing();

    glOn( glpScissor );

    glScissor( Pos.X, global_s.Display->Resolution.Y - Pos.Y - Size.Y, Size.X, Size.Y );
    glScissorStack.pushItem( Rec2Screen( Pos, Size ) );

    glCalls( 1 );
  }
  void glPopScissor() {
    if( !glScissorStack.Size )
      return;

    glFinishDrawing();

    if( --glScissorStack.Size ) {
      Rec2Screen R = glScissorStack[ glScissorStack.Size - 1 ];
      glScissor( R.Pos.X, R.Pos.Y, R.Size.X, R.Size.Y );
      glCalls( 1 );
    } else {
      glOff( glpScissor );
    }

  }

  bool glIsInDrawArea( Vec2Screen Pos, Vec2Screen Size ) {
    if( !glScissorStack.Size )
      return Rec2Interstects< TypeScreen >( Pos, Size, 0, global_s.Display->Resolution );

    else
      return glScissorStack[ glScissorStack.Size - 1 ].intersects( Pos, Size );
  }
  bool glIsInDrawArea( Vec2Screen Pos ) {
    if( !glScissorStack.Size )
      return Pos.everyGreaterOrEqual( 0 ) && Pos.everyLess( global_s.Display->Resolution );

    else
      return glScissorStack[ glScissorStack.Size - 1 ].contains( Pos );
  }

  void glBeginFrame() {
    for( UInt16 a = 0; a < _glLast; a ++ )
      glPropertyState[ a ] = gpsNone;

    glSelectedProgram = -1;
    glSelectedTexture = -1;

    glPrevCallCounter = glCallCounter;
    glCallCounter = 0;

    glDrawingMode = gldmCustom;

    glOn( glpVertexArray );
    glOn( glpColorArray );
    glOn( glpBlend );

    glOff( glpTexture2D );
    glOff( glpTextureCoordArray );
  }
  void glEndFrame() {
    glFinishDrawing();
  }

  // Draw modes
  void glColorDrawMode() {
    if( glDrawingMode == gldmColor )
      return;

    glFinishDrawing();

    glOff( glpTexture2D );
    glOff( glpTextureCoordArray );

    glSelectedTexture = -1;
    glDrawingMode = gldmColor;
  }
  void glTextureDrawMode() {
    if( glDrawingMode == gldmTexture )
      return;

    glFinishDrawing();

    glOn( glpTexture2D );
    glOn( glpTextureCoordArray );

    glDrawingMode = gldmTexture;
  }

  void glCustomDrawMode() {
    if( glDrawingMode == gldmCustom )
      return;

    glFinishDrawing();

    glSelectedTexture = -1;
    glDrawingMode = gldmCustom;
  }
  void glEndCustomDrawMode() {
    if( glDrawingMode != gldmCustom )
      return;

    glOn( glpBlend );
    glOn( glpColorMaterial );
    glOn( glpColorArray );

    glOff( glpCullFace );
    glOff( glpDepthTest );

    glColor4ub( 255, 255, 255, 255 );
  }

  // Drawing functions
  void glDrawRectangle( Vec2Screen Pos, Vec2Screen Size, Color Col ) {
    glColorDrawMode();

    glVertexBuffer.pushRect( Pos, Size );
    glColorBuffer.pushColorRect( Col );
  }
  void glDrawRectangle( Vec2Screen Pos, Vec2Screen Size, Color ColLT, Color ColRT, Color ColLB, Color ColRB ) {
    glColorDrawMode();

    glVertexBuffer.pushRect( Pos, Size );

    glColorBuffer.pushColorNTimes( ColLT, 1 );
    glColorBuffer.pushColorNTimes( ColRT, 1 );
    glColorBuffer.pushColorNTimes( ColRB, 2 );
    glColorBuffer.pushColorNTimes( ColLB, 1 );
    glColorBuffer.pushColorNTimes( ColLT, 1 );
  }
  void glDrawRectangleOutline( Vec2Screen Pos, Vec2Screen Size, Color Col ) {
    glColorDrawMode();

    glVertexBuffer.pushLineRect( Pos, Size );

    // Line rect = 4 x rect
    glColorBuffer.pushColorRect( Col );
    glColorBuffer.pushColorRect( Col );
    glColorBuffer.pushColorRect( Col );
    glColorBuffer.pushColorRect( Col );

  }

  void glDrawBuffersContent( GLBuffer2Screen &SBuf, GLBuffer2Tex &TBuf, GLBuffer2Col &CBuf, res::Texture *Tex ) {
    if( Tex->Pass != res::Texture::passDone )
      return;

    glTextureDrawMode();
    glTexture( Tex->Tex_OGL );

    glVertexBuffer.pushBuffer( SBuf );
    glTexCoordBuffer.pushBuffer( TBuf );
    glColorBuffer.pushBuffer( CBuf );
  }
  void glDrawBuffersContent( GLBuffer2Screen &SBuf, GLBuffer2Tex &TBuf, res::Texture *Tex, Color Col /* = clWhite */ ) {
    if( Tex->Pass != res::Texture::passDone )
      return;

    glTextureDrawMode();
    glTexture( Tex->Tex_OGL );

    glVertexBuffer.pushBuffer( SBuf );
    glTexCoordBuffer.pushBuffer( TBuf );
    glColorBuffer.pushColorNTimes( Col, SBuf.Size / 2 );
  }

  void glDrawBuffersContent( GLBuffer2Screen &SBuf, GLBuffer2Col &CBuf ) {
    glColorDrawMode();

    glVertexBuffer.pushBuffer( SBuf );
    glColorBuffer.pushBuffer( CBuf );
  }
  void glDrawBuffersContent( GLBuffer2Screen &SBuf, Color Col /* = clWhite */ ) {
    glColorDrawMode();

    glVertexBuffer.pushBuffer( SBuf );
    glColorBuffer.pushColorNTimes( Col, SBuf.Size / 2 );
  }

}

#endif
