#pragma once

#include "Engine.Dedicated.h"
#include "Engine.OpenGL.Header.h"
#include "Engine.GLQuad.h"
#include "Engine.GLBuffer.h"
#include "Engine.Color.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace res {
    class Texture;
  }

  enum glProperty_e {
    glpBlend,
    glpTexture2D,
    glpCullFace,
    glpDepthTest,
    glpColorMaterial,
    glpScissor,

    glpVertexArray,
    glpTextureCoordArray,
    glpColorArray,

    _glLast
  };

  extern GLBuffer2Screen glVertexBuffer;
  extern GLBuffer2Tex glTexCoordBuffer;
  extern GLBuffer2Col glColorBuffer;

  // Core functions

  void glCalls( UInt8 Num );
  SizeT glCalls();

  void glTexture( GLuint Texture );
  void glProgram( GLuint Program );

  void glOn( glProperty_e Property );
  void glOff( glProperty_e Property );

  void glPushScissor( Vec2Screen Pos, Vec2Screen Size );
  void glPopScissor();

  bool glIsInDrawArea( Vec2Screen Pos, Vec2Screen Size );
  bool glIsInDrawArea( Vec2Screen Pos );

  void glBeginFrame();
  void glEndFrame();
  void glFinishDrawing();

  // Draw modes
  void glColorDrawMode();
  void glTextureDrawMode();

  void glCustomDrawMode();
  void glEndCustomDrawMode();

  // Drawing functions
  void glDrawRectangle( Vec2Screen Pos, Vec2Screen Size, Color Col );
  void glDrawRectangle( Vec2Screen Pos, Vec2Screen Size, Color ColLT, Color ColRT, Color ColLB, Color ColRB );
  void glDrawRectangleOutline( Vec2Screen Pos, Vec2Screen Size, Color Col );

  void glDrawBuffersContent( GLBuffer2Screen &SBuf, GLBuffer2Tex &TBuf, GLBuffer2Col &CBuf, res::Texture *Tex ); 
  void glDrawBuffersContent( GLBuffer2Screen &SBuf, GLBuffer2Tex &TBuf, res::Texture *Tex, Color Col = clWhite );

  void glDrawBuffersContent( GLBuffer2Screen &SBuf, GLBuffer2Col &CBuf ); 
  void glDrawBuffersContent( GLBuffer2Screen &SBuf, Color Col = clWhite );

}

#endif