#include "Engine.Resources.h"
#include "Engine.Global.h"
#include "Engine.Application.h"
#include "Engine.Hooks.h"

#ifdef MODULE_CLIENT

namespace app {

  hook( hookPostSDL_OPENGLInit ) {
    global_s.Resources = new ResourceManager;
    global_s.Resources->init();
  }
  hook( hookAppPreDestroy ) {
    global_s.Resources->destroy();
    delete global_s.Resources;
  }
  hook( hookAppPreStep ) {
    global_s.Resources->step();
  }

  // Resource
  Resource::Resource( UInt8 Passes, const String &Identificator ) : Pass( Passes ), Identificator( Identificator ) {
    global_s.Resources->GeneralList.pushItem( this );
  }
  Resource::~Resource() {

  }

  void Resource::progress( ePass Type ) {
    switch( Type ) {

    case pEnd:
      Pass --;

      break;

    case pThreaded:
      Pass --;

      global_s.Resources->WorkerMutex.lock();
      global_s.Resources->WorkerPendingList.pushItem( this );
      global_s.Resources->WorkerMutex.unlock();

      break;

    case pMain:
      Pass --;

      global_s.Resources->GeneralMutex.lock();
      global_s.Resources->PendingList.pushItem( this );
      global_s.Resources->GeneralMutex.unlock();

      break;

    }
  }

  void Resource::destroy() {
    global_s.Resources->GeneralList.removeItem( this );
    delete this;
  }

  // ResourceManager
  ResourceManager::ResourceManager() : Scene( Scene::using2DDraw ) {
    Loaded = true;

    Pos = 0;
    WorkerPos = 0;    
    WasLoading = false;
    ActiveWorkers = 0;
  }
  ResourceManager::~ResourceManager() {
    
  }

  void ResourceManager::workers( UInt8 Num ) {
    ThreadWorker *W;

    while ( Workers.Size < Num ) {
      W = new ThreadWorker( Workers.Size );
      W->run();
      Workers.pushItem( W );
    }

    while ( Workers.Size > Num ) {
      W = Workers[ --Workers.Size ];
      W->waitForStop();
      delete W;
    }
  }

  bool ResourceManager::loading() {
    WorkerMutex.lock();
    bool R = ( Pos != PendingList.Size ) || ( WorkerPos != WorkerPendingList.Size ) || ( ActiveWorkers > 0 );
    WorkerMutex.unlock();

    return R;
  }

  void ResourceManager::loadStandard() {
    if ( !loading() )
      return;

    NextScene = global_s.Application->CurrentScene;
    global_s.Application->CurrentScene = this;
  }
  void ResourceManager::loadHidden() {
    if ( !loading() )
      return;

    NextScene = global_s.Application->CurrentScene;

    while ( loading() )
      step();
  }

  void ResourceManager::init() {
    workers( 1 );
  }
  void ResourceManager::destroy() {
    workers( 0 );

    foreach_a( GeneralList ) {
      GeneralList[ a ]->onUnload();
      delete GeneralList[ a ];
    }
  }
  void ResourceManager::step() {
    GeneralMutex.lock();

    if ( Pos < PendingList.Size )
      Current = PendingList[ Pos++ ];
    else if ( PendingList.Size ) {
      PendingList.Size = 0;
      Pos = 0;
      Current = 0;
    } else
      Current = 0;

    GeneralMutex.unlock();

    if ( Current ) 
      Current->onLoad();

    if( loading() ) {
      WasLoading = true;
    } else {
      if( global_s.Application->CurrentScene == this )
        NextScene->activate();

      if( WasLoading && global_s.Application->CurrentScene ) {
        global_s.Application->CurrentScene->GUI.recalc();
      }

      WasLoading = false;
    };
  }
  void ResourceManager::onDraw2D() {

  }

  Resource* ResourceManager::find( const String &ID ) {
    foreach_a( GeneralList ) {
      if( GeneralList[ a ]->Identificator == ID )
        return GeneralList[ a ];
    }

    return 0;
  }

  // ResourceManager::ThreadWorker
  ResourceManager::ThreadWorker::ThreadWorker( UInt16 ID ) : Thread( String( "ResourceManagerThread #" ).pushInt( ID ) ) {
    Current = 0;
  }
  ResourceManager::ThreadWorker::~ThreadWorker() {

  }

  void ResourceManager::ThreadWorker::onThrdStep() {
    global_s.Resources->WorkerMutex.lock();

    if ( Current )
      global_s.Resources->ActiveWorkers --;

    if ( global_s.Resources->WorkerPos < global_s.Resources->WorkerPendingList.Size )
      Current = global_s.Resources->WorkerPendingList[ global_s.Resources->WorkerPos++ ];
    else if ( global_s.Resources->WorkerPendingList.Size ) {
      global_s.Resources->WorkerPendingList.Size = 0;
      global_s.Resources->WorkerPos = 0;
      Current = 0;
    } else
      Current = 0;

    if ( Current )
      global_s.Resources->ActiveWorkers ++;

    global_s.Resources->WorkerMutex.unlock();

    if ( Current ) 
      Current->onLoad();
  }

}

#endif