#pragma once

#include "Engine.Dedicated.h"
#include "Engine.Types.h"
#include "Engine.Container.h"
#include "Engine.Scene.h"
#include "Engine.Thread.h"

#ifdef MODULE_CLIENT

namespace app {

  class Resource {
    
  public:
    enum ePass {
      pEnd,
      pThreaded,
      pMain,
    };

  public:
    UInt8 Pass; // Pass 0 = finished, decreasing
    String Identificator;

  public:
    Resource( UInt8 Passes, const String &Identificator );
    virtual ~Resource();

  public:
    void progress( ePass Type );

    void destroy();

  public:
    virtual void onLoad() = 0;
    virtual void onUnload() {};

  };

  class ResourceManager : public Scene {

  public:
    class ThreadWorker;

  public:
    Container< Resource* > PendingList, WorkerPendingList, GeneralList;
    SizeT Pos, WorkerPos, ActiveWorkers;
    Mutex WorkerMutex, GeneralMutex;

    Resource *Current;

    Container< ThreadWorker* > Workers;

    Scene *NextScene;
    bool WasLoading;

  public:
    ResourceManager();
    ~ResourceManager();

  public:
    void workers( UInt8 Num );

    bool loading();

    void loadStandard();
    void loadHidden();

    void step();
    void init();
    void destroy();

    Resource* find( const String &ID );

  public:
    virtual void onDraw2D();

  };

  class ResourceManager::ThreadWorker : public Thread {

  public:
    Resource *Current;

  public:
    ThreadWorker( UInt16 ID );
    ~ThreadWorker();

  public:
    virtual void onThrdStep();

  };

}

#endif