#include "Engine.Stream.h"
#include "Engine.String.h"
#include "Engine.Exception.h"

namespace app {

  // Stream

  Stream::Stream() : Size( 0 ), Capacity( 8 ), Reader( 0 ) {
    Ptr = ( UInt8* ) malloc( Capacity );
  }
  Stream::Stream( const Stream &Source ) : Size( Source.Size ), Capacity( Source.Capacity ), Reader( 0 ) {
    Ptr = ( UInt8* ) malloc( Capacity );
    memcpy( Ptr, Source.Ptr, Size );
  }
  Stream::~Stream() {
    free( Ptr );
  }

  void Stream::allocate( SizeT N ) {
    if( Size + N > Capacity ) {
      while( Size + N > Capacity )
        Capacity *= 2;

      Ptr = ( UInt8 * ) realloc( Ptr, Capacity );
    }
  }

  void Stream::push( const Streamable &Source, UInt8 Variant /* = 0 */ ) {
    Source.onStreamSerialize( this, Variant );
  }
  void Stream::push( const void *Source, const SizeT Count ) {
    allocate( Count );
    memcpy( Ptr + Size, Source, Count );
    Size += Count;
  }

  void Stream::pull( Streamable &Target, UInt8 Variant /* = 0 */ ) {
    Target.onStreamDeserialize( this, Variant );
  }
  void Stream::pull( void *Target, SizeT Count ) {
    ASSERT_THROW( Reader + Count > Size, "indexOutOfBounds", "location", &IString("Stream::pull( void *Target, Count )") );
    memcpy( Target, Ptr + Reader, Count );
    Reader += Count;
  }

  void Stream::clear() {
    Size = 0;
    Reader = 0;
  }

  void Stream::onStreamSerialize( Stream *Str, UInt8 Variant ) const {
    Str->pushSimple< SizeT >( Size );
    Str->push( Ptr, Size );
  }
  void Stream::onStreamDeserialize( Stream *Str, UInt8 Variant )  {
    clear();
    Str->pullSimple< SizeT >( Size );
    allocate( 0 );
    Str->pull( Ptr, Size );
  }

}