#pragma once

#include "Engine.Types.h"
#include "Engine.Container.h"
#include "Engine.Exception.h"
#include "Engine.String.h"
#include "Engine.Streamable.h"

namespace app {

  class Stream : public Streamable {

  public:
    UInt8 *Ptr;
    SizeT Size, Capacity, Reader;

  public:
    Stream();
    Stream( const Stream &Source );
    ~Stream();

  public:
    void allocate( SizeT N );

  public:
    void push( const Streamable &Source, UInt8 Variant = 0 );
    void push( const void *Source, const SizeT Count );
    template< typename T >
    void pushSimple( const T &Source ) {
      allocate( sizeof( T ) );
      memcpy( Ptr + Size, &Source, sizeof( T ) );
      Size += sizeof( T );
    }

    void pull( Streamable &Target, UInt8 Variant = 0 );
    void pull( void *Target, SizeT Count );
    template< typename T >
    void pullSimple( T &Target ) {
      ASSERT_THROW( ( ( Reader + sizeof( T ) ) > Size ), "indexOutOfBounds", "location", &IString( "Stream::pullSimple" ) );
      memcpy( &Target, Ptr + Reader, sizeof( T ) );
      Reader += sizeof( T );
    }

    void clear();

  public:
    virtual void onStreamSerialize( Stream *Str, UInt8 Variant ) const ;
    virtual void onStreamDeserialize( Stream *Str, UInt8 Variant );


  };

}
