#pragma once

#include "Engine.String.h"
#include "Engine.Exception.h"
#include "Engine.Vector.h"
#include "Engine.Color.h"
#include "Engine.Confix.h"
#include "Engine.Global.h"

namespace app {

  UInt8 strHexToDec1( String &Str, SizeT Pos );
  UInt8 strHexToDec2( String &Str, SizeT Pos );

  template< typename T >
  inline void stringTo( String &Str, T &Var ) {
    ASSERT_THROW( true, "invalidStringCastTo" );
  }


  template<>
  inline void stringTo< int >( String& Str, int &Var ) {
    Var = Str.toInt();
  }
  template<>
  inline void stringTo< short >( String& Str, short &Var ) {
    Var = Str.toInt();
  }
  template<>
  inline void stringTo< char >( String& Str, char &Var ) {
    Var = Str.toInt();
  }


  template<>
  inline void stringTo< unsigned char >( String& Str, unsigned char &Var ) {
    Var = Str.toInt();
  }
  template<>
  inline void stringTo< unsigned short >( String& Str, unsigned short &Var ) {
    Var = Str.toInt();
  }


  template<>
  inline void stringTo< Vec2< short > >( String& Str, Vec2< short > &Var ) {
    if ( !Str.Size ) {
      Var = Vec2< short >();
      return;
    }

    int R = 0;
    char Ch = Str.Ptr[ 0 ];

    bool Neg = false;
    SizeT a = 0;

    if ( Ch == '-' ) {
      Neg = true;
      a = 1;
    }

    while ( a < Str.Size ) {
      Ch = Str.Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' )
        R = R * 10 + Ch - '0';
      else {
        a ++;
        break;
      }

      a ++;
    }

    if ( Neg )
      Var.X = -R;
    else
      Var.X = R;

    // Y

    Neg = false;
    R = 0;

    if ( Ch == '-' ) {
      Neg = true;
      a ++;
    }

    while ( a < Str.Size ) {
      Ch = Str.Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' )
        R = R * 10 + Ch - '0';
      else {
        break;
      }

      a ++;
    }

    if ( Neg )
      Var.Y = -R;
    else
      Var.Y = R;

  }
  template<>
  inline void stringTo< Vec2< int > >( String& Str, Vec2< int > &Var ) {
    if ( !Str.Size ) {
      Var = Vec2< int >();
      return;
    }

    int R = 0;
    char Ch = Str.Ptr[ 0 ];

    bool Neg = false;
    SizeT a = 0;

    if ( Ch == '-' ) {
      Neg = true;
      a = 1;
    }

    while ( a < Str.Size ) {
      Ch = Str.Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' )
        R = R * 10 + Ch - '0';
      else {
        a ++;
        break;
      }

      a ++;
    }

    if ( Neg )
      Var.X = -R;
    else
      Var.X = R;

    // Y

    Neg = false;
    R = 0;

    if ( Ch == '-' ) {
      Neg = true;
      a ++;
    }

    while ( a < Str.Size ) {
      Ch = Str.Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' )
        R = R * 10 + Ch - '0';
      else {
        break;
      }

      a ++;
    }

    if ( Neg )
      Var.Y = -R;
    else
      Var.Y = R;

  }

  template<>
  inline void stringTo< bool >( String& Str, bool &Var ) {
    if ( Str == "true" ) {
      Var = true;
      return;
    }

    if ( Str == "1" ) {
      Var = true;
      return;
    }

    if ( Str == "yes" ) {
      Var = true;
      return;
    }

    if ( Str == "on" ) {
      Var = true;
      return;
    }

    Var = false;
  }


  template<>
  inline void stringTo< float >( String& Str, float &Var ) {
    char *Ch = Str.toCChar();
    Var = ( float ) atof( Ch );
    free( Ch );
  }

  template<>
  inline void stringTo< Color >( String& Str, Color &Var ) {
    if( Str.Size && Str[ 0 ] == '@' ) {
      stringTo< Color >( *global_s.gui.ColorsNode > Str, Var );
      return;
    }

    Var.assign( 0, 0, 0, 0 );

    if ( Str.Size < 6 )
      return;

    Var.R = strHexToDec2( Str, 0 );
    Var.G = strHexToDec2( Str, 2 );
    Var.B = strHexToDec2( Str, 4 );

    if( Str.Size < 8 ) {
      Var.A = 255;
      return;
    }

    Var.A = strHexToDec2( Str, 6 );
  }

  // ----

  template< typename T >
  inline void stringToD( String &Str, T &Var ) {
    ASSERT_THROW( true, "invalidStringCastTo" );
  }

  template<>
  inline void stringToD< short >( String& Str, short &Var ) {
    if( Str.Size == 0 )
      return;

    if( Str.Size == 1 && Str.Ptr[ 0 ] == 'd' )
      return;

    Var = Str.toInt();
  }
  template<>
  inline void stringToD< int >( String& Str, int &Var ) {
    if( Str.Size == 0 )
      return;

    if( Str.Size == 1 && Str.Ptr[ 0 ] == 'd' )
      return;

    Var = Str.toInt();
  }

  // --------------------------

  template< typename T >
  inline void stringFrom( String &Str, T &Var ) {
    ASSERT_THROW( true, "invalidStringCastFrom" );
  }


  template<>
  inline void stringFrom< int >( String& Str, int &Var ) {
    Str.clear().pushInt( Var );
  }
  template<>
  inline void stringFrom< short >( String& Str, short &Var ) {
    Str.clear().pushInt( Var );
  }
  template<>
  inline void stringFrom< char >( String& Str, char &Var ) {
    Str.clear().pushInt( Var );
  }


  template<>
  inline void stringFrom< unsigned char >( String& Str, unsigned char &Var ) {
    Str.clear().pushInt( Var );
  }
  template<>
  inline void stringFrom< unsigned short >( String& Str, unsigned short &Var ) {
    Str.clear().pushInt( Var );
  }


  template<>
  inline void stringFrom< Vec2< short > >( String& Str, Vec2< short > &Var ) {
    Str.clear().pushInt( Var.X ).pushChar( ' ' ).pushInt( Var.Y );
  }
  template<>
  inline void stringFrom< Vec2< int > >( String& Str, Vec2< int > &Var ) {
    Str.clear().pushInt( Var.X ).pushChar( ' ' ).pushInt( Var.Y );
  }


  template<>
  inline void stringFrom< bool >( String& Str, bool &Var ) {
    if ( Var )
      Str = "true";
    else
      Str = "false";
  }


  template<>
  inline void stringFrom< float >( String& Str, float &Var ) {
    char Buf[ 50 ];
    if( ( float )( ( int ) Var ) == Var )
      sprintf_danol( Buf, "%.0f", Var );
    else
      sprintf_danol( Buf, "%g", Var );

    Str = String( Buf );
  }

}
