#include "Engine.String.h"
#include "Engine.Stream.h"

namespace app {

  // IString
  IString::IString() : Ptr( 0 ), Size( 0 ) {

  }
  IString::IString( const IString &Source ) : Ptr( Source.Ptr ), Size( Source.Size ) {

  }
  IString::IString( char *Ptr, SizeT Size ) : Ptr( Ptr ), Size( Size ) {

  }
  IString::IString( const char *Source ) : Ptr( ( char* ) Source ), Size( strlen( Source ) ) {

  }

  IString::~IString() {

  }

  void IString::assign( const char *Source ) {
    Ptr = ( char* ) Source;
    Size = strlen( Source );
  }
  void IString::assign( char *IPtr, SizeT ISize ) {
    Ptr = IPtr;
    Size = ISize;
  }

  char* IString::toCChar() const {
    char *R = ( char* ) malloc( Size + 1 );
    R[ Size ] = 0;
    memcpy( R, Ptr, Size );

    return R;
  }
  void IString::print() const {
    printf( "\n----- String out -----\n%.*s\n----- String out end -----\n\n", Size, Ptr );
  }

  bool IString::contains( char Symbol ) const {
    for( SizeT a = 0; a < Size; a ++ ) {
      if( Ptr[ a ] == Symbol )
        return true;
    }

    return false;
  }

  bool IString::operator==( const char *Source ) const {
    SizeT L = strlen( Source );

    if( L != Size )
      return false;
    else
      return memcmp( Source, Ptr, Size ) == 0;
  }
  bool IString::operator==( const IString &Source ) const {
    if( Source.Size != Size )
      return false;
    else
      return memcmp( Source.Ptr, Ptr, Size ) == 0;
  }

  bool IString::operator!=( const char *Source ) const {
    SizeT L = strlen( Source );

    if( L != Size )
      return true;
    else
      return memcmp( Source, Ptr, Size ) != 0;
  }
  bool IString::operator!=( const IString &Source ) const {
    if( Source.Size != Size )
      return true;
    else
      return memcmp( Source.Ptr, Ptr, Size ) != 0;
  }

  char IString::last( SizeT Pos /* = 0 */ ) const {
    if( Pos >= Size )
      return 0;

    return Ptr[ Size - 1 - Pos ];
  }

  int IString::toInt() const {
    if ( !Size )
      return 0;

    int R = 0;
    char Ch = Ptr[ 0 ];

    bool Neg = false;
    SizeT a = 0;

    if ( Ch == '-' ) {
      Neg = true;
      a = 1;
    }

    while ( a < Size ) {
      Ch = Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' )
        R = R * 10 + Ch - '0';

      a ++;
    }

    if ( Neg )
      return -R;
    else
      return R;
  }
  float IString::toFloat() const {
    if ( !Size )
      return 0;

    float R = 0.0f, O = 0.1f;
    char Ch = Ptr[ 0 ];

    bool Neg = false;
    SizeT a = 0;

    if ( Ch == '-' ) {
      Neg = true;
      a = 1;
    }

    while ( a < Size ) {
      Ch = Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' )
        R = R * 10 + Ch - '0';
      else if( Ch == '.' || Ch == ',' )
        break;

      a ++;
    }
    while ( a < Size ) {
      Ch = Ptr[ a ];

      if ( Ch >= '0' && Ch <= '9' ) {
        R = R + O * ( ( float )( Ch - '0' ) );
        O *= 0.1f;
      }

      a ++;
    }


    if ( Neg )
      return -R;
    else
      return R;
  }

  void IString::onStreamSerialize( Stream *Str, UInt8 Variant ) const {
    Str->pushSimple< SizeT >( this->Size );
    Str->push( this->Ptr, this->Size * sizeof( char ) );
  }

  // String

  String::String() {
    Ptr = 0;
    Size = 0;
  }
  String::String( const char *Source ) {
    Size = strlen( Source );
    Ptr = ( char * ) malloc( Size );

    memcpy( Ptr, Source, Size );
  }
  String::String( const String &Source ) {
    Size = Source.Size;
    Ptr = ( char* ) malloc( Size );

    memcpy( Ptr, Source.Ptr, Size );
  }
  String::String( const IString &Source ) {
    Size = Source.Size;
    Ptr = ( char* ) malloc( Size );

    memcpy( Ptr, Source.Ptr, Size );
  }

  String::~String() {
    free( Ptr );
  }

  void String::assign( const char *Source ) {
    Size = strlen( Source );
    Ptr = ( char* ) realloc( Ptr, Size );
    memcpy( Ptr, Source, Size );
  }
  void String::assign( char *IPtr, SizeT ISize ) {
    Size = ISize;
    Ptr = ( char* ) realloc( Ptr, Size );
    memcpy( Ptr, IPtr, Size );
  }

  String& String::clear() {
    Size = 0;
    free( Ptr );
    Ptr = 0;

    return *this;
  }

  String& String::pushChar( char Ch ) {
    Ptr = ( char* ) realloc( Ptr, ++Size );
    Ptr[ Size - 1 ] = Ch;
    return *this;
  }
  String& String::pushString( const IString& Source ) {
    if( !Source.Size )
      return *this;

    Size += Source.Size;
    Ptr = ( char* ) realloc( Ptr, Size );

    memcpy( Ptr + Size - Source.Size, Source.Ptr, Source.Size );
    return *this;
  }
  String& String::pushCChar( const char *Source ) {
    if( !*Source )
      return *this;

    SizeT Len = strlen( Source );

    Size += Len;
    Ptr = ( char* ) realloc( Ptr, Size );

    memcpy( Ptr + Size - Len, Source, Len );
    return *this;
  }

  String& String::pushInt( int I ) {
    if ( I == 0 ) {
      pushChar( '0' );
      return *this;
    }

    if ( I < 0 ) {
      pushChar( '-' );
      I *= -1;
    }

    int Tmp = 1;

    while ( Tmp <= I )
      Tmp *= 10;

    Tmp /= 10;

    while ( Tmp > 0 ) {
      pushChar( '0' + ( I / Tmp ) );
      I = I % Tmp;
      Tmp /= 10;
    }

    return *this;
  }
  String& String::pushFloat( float Var, float Accuracy /* = 0.01f */ ) {
    Var = mRound< float >( Var / Accuracy ) * Accuracy;

    char Buf[ 50 ];
    if( ( float )( ( int ) Var ) == Var )
      sprintf_danol( Buf, "%.0f", Var );
    else
      sprintf_danol( Buf, "%g", Var );
    pushCChar( Buf );

    return *this;
  }

  void String::operator=( const String &Source ) {
    Ptr = ( char* ) realloc( Ptr, Source.Size );
    Size = Source.Size;

    memcpy( Ptr, Source.Ptr, Size );
  }
  void String::operator=( const IString &Source ) {
    Ptr = ( char* ) realloc( Ptr, Source.Size );
    Size = Source.Size;

    memcpy( Ptr, Source.Ptr, Size );
  }
  void String::operator=( const char *Source ) {
    Size = strlen( Source );
    Ptr = ( char* ) realloc( Ptr, Size );

    memcpy( Ptr, Source, Size );
  }

  void String::putNullOnEnd() {
    Ptr = ( char* ) realloc( Ptr, Size + 1 );
    Ptr[ Size ] = 0;
  }

  char* String::getCChar() {
    putNullOnEnd();
    return Ptr;
  }
  const IString* String::istr() const {
    return this;
  }

  void String::onStreamSerialize( Stream *Str, UInt8 Variant ) const {
    Str->pushSimple< SizeT >( this->Size );
    Str->push( this->Ptr, this->Size * sizeof( char ) );
  }
  void String::onStreamDeserialize( Stream *Str, UInt8 Variant ) {
    this->clear();
    Str->pullSimple< SizeT >( this->Size );
    Ptr = ( char* ) realloc( Ptr, Size );

    Str->pull( this->Ptr, this->Size * sizeof( char ) );
  }

}