#pragma once

#include <stdio.h>
#include <string.h>

#include "Engine.Container.h"
#include "Engine.Streamable.h"

#ifdef __MINGW32__

#define sprintf_danol sprintf

#else

#define sprintf_danol sprintf_s

#endif

namespace app {

  class IString {

  public:
    char *Ptr;
    SizeT Size;

  public:
    IString();
    IString( const IString &Source );
    IString( const char *Source );
    IString( char *Ptr, SizeT Size );

    ~IString();

  public:
    void assign( char *Ptr, SizeT Size );
    void assign( const char *Source );

  public:
    char* toCChar() const; // MUST BE FREED!!
    void print() const;

  public:
    bool contains( char Symbol ) const;

  public:
    bool operator==( const char *Source ) const;
    bool operator==( const IString &Source ) const;

    bool operator!=( const char *Source ) const;
    bool operator!=( const IString &Source ) const;

  public:
    inline char operator[]( const SizeT Pos ) const {
      return Ptr[ Pos ];
    }
    inline char& operator[]( const SizeT Pos ) {
      return Ptr[ Pos ];
    }

    char last( SizeT Pos = 0 ) const;

  public:
    int toInt() const;
    float toFloat() const;

  public:
    virtual void onStreamSerialize( Stream *Str, UInt8 Variant ) const;

  };

  class String : public IString {

  public:
    String();
    String( const char *Source );
    String( const String &Source );
    String( const IString &Source );

    ~String();

  public:
    void assign( char *Ptr, SizeT Size );
    void assign( const char *Source );

    String& clear();

  public:
    String& pushChar( char Ch );
    String& pushString( const IString& Source );
    String& pushCChar( const char * Source );

    String& pushInt( int I );
    String& pushFloat( float F, float Accuracy = 0.01f );

  public:
    void operator=( const char *Source );
    void operator=( const String &Source );
    void operator=( const IString &Source );

  public:
    void putNullOnEnd();

    char* getCChar(); // MUST NOT BE FREED
    const IString* istr() const;

  public:
    virtual void onStreamSerialize( Stream *Str, UInt8 Variant ) const;
    virtual void onStreamDeserialize( Stream *Str, UInt8 Variant );

  };

}