#include "Engine.Thread.h"
#include "Engine.Time.h"
#include "Engine.Hooks.h"
#include "Engine.Exception.h"
#include "Engine.Dedicated.h"

namespace app {

  int runProcedure( void * Ptr ) {
    ( ( Thread* )( Ptr ) )->_runProcedure();    
    return 0;
  }

  Thread::Thread( const String &Name ) : Name( Name ) {
    Status = stStopped;
    CName = Name.toCChar();
  }
  Thread::~Thread() {
    waitForStop();

    free( ( void* ) CName );
  }

  void Thread::run() {
    if ( Status != stStopped )
      return;

    Thrd = SDL_CreateThread( runProcedure, this );
    ASSERT_THROW( !Thrd, "thrdCreateFail", "name", &Name );
  }
  void Thread::run_nothrd() {
    if ( Status != stStopped )
      return;
    _runProcedure();
  }
  void Thread::stop() {
    StatusMutex.lock();
    if ( Status == stRunning )
      Status = stTerminating;
    StatusMutex.unlock();
  }

  void Thread::waitForStop() {
    /*if( UInt32 ID = SDL_GetThreadID( Thrd ) != 0 )
      SDL_WaitThread( Thrd, 0 );*/

    stop();

    bool R;
    while( true ) {
      StatusMutex.lock();
      R = Status == stStopped;
      StatusMutex.unlock();

      if( R ) break;
      time::sleep( 10 );
    }

    /*if( UInt32 ID = SDL_GetThreadID( Thrd ) != 0 )
      SDL_WaitThread( Thrd, 0 );*/
  }
  void Thread::waitForRun() {
    run();

    while( Status !=  stRunning )
      time::sleep( 10 );
  }

  void Thread::_runProcedure() {
    onThrdInit();
    HookManager.call( hookThrdInit );

    Performance.init();

    Status = stRunning;

    while( true ) {
      StatusMutex.lock();
      if( Status != stRunning ) { StatusMutex.unlock(); break; }
      StatusMutex.unlock();

      if( Performance.step() )
        onThrdStep();
    }

    HookManager.call( hookThrdDestroy );
    onThrdDestroy();

    Status = stStopped;
  }

  // Mutex
  Mutex::Mutex() {
    M = SDL_CreateMutex();
  }
  Mutex::~Mutex() {
    SDL_DestroyMutex( M );
  }

  void Mutex::lock() {
    debugOnly(
      ASSERT( !M, "nullMutex" );
      )
    SDL_LockMutex( M );
  }
  void Mutex::unlock() {
    debugOnly(
      ASSERT( !M, "nullMutex" );
    )

    SDL_UnlockMutex( M );
  }

}