#include "Engine.Time.h"
#include "Engine.Math.h"

#include <SDL/SDL_timer.h>

namespace app {

  namespace time {

    type get() {
      return SDL_GetTicks();
    }

    void sleep( UInt16 MS ) {
      SDL_Delay( ( Uint32 ) MS );
    }

  }

  // PerformanceUtility

  PerformanceUtility::PerformanceUtility() {

  }
  PerformanceUtility::~PerformanceUtility() {

  }

  void PerformanceUtility::init() {
    LastTime = time::get();
    LastUpdate = LastTime;
    MinPing = 0;

    FPS = 0;
    FPSC = 0;
    FPSCp = 0;
    Ping = 0;
    OneStep = false;
  }
  bool PerformanceUtility::step() {
    CurrentTime = time::get();

    Ping = CurrentTime - LastTime;
    if ( Ping == 0 )
      FPS = -1;
    else
      FPS = 1000 / Ping;

    FPSCp ++;

    if ( LastUpdate + 1000 < CurrentTime ) {
      LastUpdate = CurrentTime;
      FPSC = FPSCp;
      FPSCp = 0;
    }

    if( OneStep ) {
      OneStep = false;
      LastTime = CurrentTime;
      return true;
    }

    if ( Ping < MinPing ) {
      time::sleep( mMin< time::type >( MinPing - Ping, 1000 ) );
      return false;
    } else
      LastTime = CurrentTime;
      return true;
  }
  void PerformanceUtility::destroy() {

  }

  void PerformanceUtility::limitFPS( UInt16 MaxFPS ) {
    if( MaxFPS == 0 )
      MinPing = -1;
    else
      MinPing = 2000 / MaxFPS;
  }
  void PerformanceUtility::oneStep() {
    OneStep = true;
  }

  UInt16 PerformanceUtility::getFPS() const {
    return FPSC;
  }
  UInt16 PerformanceUtility::getActualFPS() const {
    return FPS;
  }
  time::type PerformanceUtility::getPing() const {
    if( FPSC )
      return 1000 / FPSC;
    else return 99999;
  }
  time::type PerformanceUtility::getTime() const {
    return CurrentTime;
  }

  // FastWatch

  FastWatch::FastWatch() {
    Time = time::get();
  }
  FastWatch::~FastWatch() {

  }

  time::type FastWatch::get() {
    Loc = Time;
    Time = time::get();

    return Time - Loc;
  }

  // StopWatch

  StopWatch::StopWatch() {
    Time = time::get();
  }
  StopWatch::~StopWatch() {

  }

  time::type StopWatch::get() {
    return time::get() - Time;
  }
  void StopWatch::reset() {
    Time = time::get();
  }

  // PerformanceSkipLimiter
  PerformanceSkipLimiter::PerformanceSkipLimiter() {
    
  }
  PerformanceSkipLimiter::~PerformanceSkipLimiter() {

  }

  void PerformanceSkipLimiter::init() {
    StartTime = time::get();
    TargetPing = 1;
    PassedPings = 0;
  }
  void PerformanceSkipLimiter::setFPS( UInt16 FPS ) {
    StartTime = time::get();
    PassedPings = 0;

    if( FPS == 0 )
      TargetPing = 0;
    else
      TargetPing = 1000 / FPS;
  }

  bool PerformanceSkipLimiter::step() {
    if( TargetPing == 0 )
      return false;

    R = ( time::get() - StartTime ) - PassedPings * TargetPing;
    if( R > TargetPing ) {
      PassedPings ++;

      if( R > TargetPing * 30 ) {
        StartTime += TargetPing; // When the debt is too large, decrease it
      }
      return true;
    } else
      return false;
  }

}