#pragma once

#include "Engine.Types.h"
#include "Engine.Math.h"
#include "Engine.Stream.h"

namespace app {

  template< typename T >
  class Vec2 : public Streamable {

  public:
    Vec2() : X( 0 ), Y( 0 ) {
    }
    Vec2( const T U ) : X( U ), Y( U ) {

    }
    Vec2( const T X, const T Y ) : X( X ), Y( Y ) {

    }
    Vec2( const Vec2< T >& Source ) : X( Source.X ), Y( Source.Y ) {

    }

    ~Vec2() {

    }

  public:
    T X, Y;

  public: //sub-vector operators
    T x() const {
      return X;
    }
    T y() const {
      return Y;
    }

    Vec2< T > xx() const {
      return Vec2< T >( X, X );
    }
    Vec2< T > yy() const {
      return Vec2< T >( Y, Y );
    }

    Vec2< T > xy() const {
      return Vec2< T >( X, Y );
    }
    Vec2< T > yx() const {
      return Vec2< T >( Y, X );
    }

    Vec2< T >& base() {
      return *this;
    }
    const Vec2< T >& base() const {
      return *this;
    }

  public: //assigment operators
    void operator=( const T U ) {
      X = U;
      Y = U;
    }
    void operator=( const Vec2< T >& Source ) {
      X = Source.X;
      Y = Source.Y;
    }

    void operator+=( const T U ) {
      X += U;
      Y += U;
    }
    void operator+=( const Vec2< T >& Source ) {
      X += Source.X;
      Y += Source.Y;
    }

    void operator-=( const T U ) {
      X -= U;
      Y -= U;
    }
    void operator-=( const Vec2< T >& Source ) {
      X -= Source.X;
      Y -= Source.Y;
    }

    void operator*=( const T U ) {
      X *= U;
      Y *= U;
    }
    void operator*=( const Vec2< T >& Source ) {
      X *= Source.X;
      Y *= Source.Y;
    }

    void operator/=( const T U ) {
      X /= U;
      Y /= U;
    }
    void operator/=( const Vec2< T >& Source ) {
      X /= Source.X;
      Y /= Source.Y;
    }

    void operator++() {
      X ++;
      Y ++;
    }
    void operator--() {
      X --;
      Y --;
    }

    void assign( const T U ) {
      X = U;
      Y = U;
    }
    void assign( const T Xi, const T Yi ) {
      X = Xi;
      Y = Yi;
    }
    void assign( const Vec2< T >& Source ) {
      X = Source.X;
      Y = Source.Y;
    }
    template< typename T2 >
    void assign( const Vec2< T2 >& Source ) {
      X = ( T ) Source.X;
      Y = ( T ) Source.Y;
    }

  public: //cast operators
    template< typename T2 >
    Vec2< T2 > cast() {
      return Vec2< T2 >( ( T2 ) X, ( T2 ) Y );
    }

  public: //mathematic operators
    Vec2< T > operator+( const T U ) const {
      return Vec2< T >( X + U, Y + U );
    }
    Vec2< T > operator+( const Vec2< T >& Source ) const {
      return Vec2< T >( X + Source.X, Y + Source.Y );
    }

    Vec2< T > operator-( const T U ) const {
      return Vec2< T >( X - U, Y - U );
    }
    Vec2< T > operator-( const Vec2< T >& Source ) const {
      return Vec2< T >( X - Source.X, Y - Source.Y );
    }

    Vec2< T > operator*( const T U ) const {
      return Vec2< T >( X * U, Y * U );
    }
    Vec2< T > operator*( const Vec2< T >& Source ) const {
      return Vec2< T >( X * Source.X, Y * Source.Y );
    }

    Vec2< T > operator/( const T U ) const {
      return Vec2< T >( X / U, Y / U );
    }
    Vec2< T > operator/( const Vec2< T >& Source ) const {
      return Vec2< T >( X / Source.X, Y / Source.Y );
    }

  public: //comparsion operators
    bool operator==( const T U ) const {
      return X == U && Y == U;
    }
    bool operator==( const Vec2< T >& Source ) const {
      return X == Source.X && Y == Source.Y;
    }

    bool operator!=( const T U ) const {
      return X != U || Y != U;
    }
    bool operator!=( const Vec2< T >& Source ) const {
      return X != Source.X || Y != Source.Y;
    }

    bool oneGreater( const T U ) const {
      return X > U || Y > U;
    }
    bool oneGreater( const Vec2< T >& Source ) const {
      return X > Source.X || Y > Source.Y;
    }
    bool oneGreaterOrEqual( const T U ) const {
      return X >= U || Y >= U;
    }
    bool oneGreaterOrEqual( const Vec2< T >& Source ) const {
      return X >= Source.X || Y >= Source.Y;
    }

    bool oneLess( const T U ) const {
      return X < U || Y < U;
    }
    bool oneLess( const Vec2< T >& Source ) const {
      return X < Source.X || Y < Source.Y;
    }
    bool oneLessOrEqual( const T U ) const {
      return X <= U || Y <= U;
    }
    bool oneLessOrEqual( const Vec2< T >& Source ) const {
      return X <= Source.X || Y <= Source.Y;
    }

    bool everyGreater( const T U ) const {
      return X > U && Y > U;
    }
    bool everyGreater( const Vec2< T >& Source ) const {
      return X > Source.X && Y > Source.Y;
    }
    bool everyGreaterOrEqual( const T U ) const {
      return X >= U && Y >= U;
    }
    bool everyGreaterOrEqual( const Vec2< T >& Source ) const {
      return X >= Source.X && Y >= Source.Y;
    }

    bool everyLess( const T U ) const {
      return X < U && Y < U;
    }
    bool everyLess( const Vec2< T >& Source ) const {
      return X < Source.X && Y < Source.Y;
    }
    bool everyLessOrEqual( const T U ) const {
      return X <= U && Y <= U;
    }
    bool everyLessOrEqual( const Vec2< T >& Source ) const {
      return X <= Source.X && Y <= Source.Y;
    }

  public:
    Vec2< T >& mMax( const Vec2< T > Source ) {
      if( Source.X > X )
        X = Source.X;

      if( Source.Y > Y )
        Y = Source.Y;

      return *this;
    }
    Vec2< T >& mMin( const Vec2< T > Source ) {
      if( Source.X < X )
        X = Source.X;

      if( Source.Y < Y )
        Y = Source.Y;

      return *this;
    }
    Vec2< T >& mLimit( const Vec2< T > Min, const Vec2< T > Max ) {
      mMax( Min );
      mMin( Max );
      return *this;
    }

    Vec2< T > mRound() {
      X = app::mRound< T >( X );
      Y = app::mRound< T >( Y );
      return *this;
    }
    Vec2< T > mCeil() {
      X = ceil( X );
      Y = ceil( Y );
      return *this;
    }
    Vec2< T > mFloor() {
      X = floor( X );
      Y = floor( Y );
      return *this;
    }

    static Vec2< T > mMin( const Vec2< T >& S1, const Vec2< T >& S2 ) {
      return Vec2< T >( app::mMin< T >( S1.X, S2.X ), app::mMin< T >( S1.Y, S2.Y ) );
    }
    static Vec2< T > mMax( const Vec2< T >& S1, const Vec2< T >& S2 ) {
      return Vec2< T >( app::mMax< T >( S1.X, S2.X ), app::mMax< T >( S1.Y, S2.Y ) );
    }
    static Vec2< T > mLimit( const Vec2< T >& V, const Vec2< T >& Min, const Vec2< T >& Max ) {
      return Vec2< T >( app::mLimit< T >( V.X, Min.X, Max.X ), app::mLimit< T >( V.Y, Min.Y, Max.Y ) );
    }

    Vec2< T > mAbs() {
      return Vec2< T >( abs( X ), abs( Y ) ); 
    }

   public:
      T minAspect() {
        if( X > Y )
          return Y;
        else
          return X;
      }

      double size() {
        return sqrt( pow( ( double ) X, 2.0 ) + pow( ( double ) Y, 2.0 ) );
      }

  public:
    virtual void onStreamSerialize( Stream *Target, UInt8 Variant ) const {
      Target->pushSimple< T >( X );
      Target->pushSimple< T >( Y );
    }
    virtual void onStreamDeserialize( Stream *Source, UInt8 Variant ) {
      Source->pullSimple< T >( X );
      Source->pullSimple< T >( Y );
    }

  };

  template< typename T >
  class Vec3 /*: public Streamable*/ {

  public:
    Vec3() : X( 0 ), Y( 0 ), Z( 0 ) {
    }
    Vec3( const T U ) : X( U ), Y( U ), Z( U ) {

    }
    Vec3( const T X, const T Y, const T Z ) : X( X ), Y( Y ), Z( Z ) {

    }
    Vec3( const Vec3< T > &Source ) : X( Source.X ), Y( Source.Y ), Z( Source.Z ) {

    }

    ~Vec3() {

    }

  public:
    T X, Y, Z;

  public: //sub-vector operators
    T x() const {
      return X;
    }
    T y() const {
      return Y;
    }
    T z() const {
      return Z;
    }

    Vec2< T > xx() const {
      return Vec3< T >( X, Y );
    }
    Vec2< T > yy() const {
      return Vec3< T >( Y, Y );
    }

    Vec2< T > xy() const {
      return Vec3< T >( X, Y );
    }
    Vec2< T > yx() const {
      return Vec3< T >( Y, X );
    }

    Vec3< T > xxx() const {
      return Vec3( X, X, X );
    }
    Vec3< T > yyy() const {
      return Vec3( Y, Y, Y );
    }
    Vec3< T > zzz() const {
      return Vec3( Z, Z, Z );
    }

    Vec3< T > xyz() const {
      return Vec3( X, Y, Z );
    }
    Vec3< T > zyx() const {
      return Vec3( Z, Y, X );
    }

  public: //assigment operators
    void operator=( const T U ) {
      X = U;
      Y = U;
      Z = U;
    }
    void operator=( const Vec3< T >& Source ) {
      X = Source.X;
      Y = Source.Y;
      Z = Source.Z;
    }

    void operator+=( const T U ) {
      X += U;
      Y += U;
      Z += U;
    }
    void operator+=( const Vec3< T >& Source ) {
      X += Source.X;
      Y += Source.Y;
      Z += Source.Z;
    }

    void operator-=( const T U ) {
      X -= U;
      Y -= U;
      Z -= U;
    }
    void operator-=( const Vec3< T >& Source ) {
      X -= Source.X;
      Y -= Source.Y;
      Z -= Source.Z;
    }

    void operator*=( const T U ) {
      X *= U;
      Y *= U;
      Z *= U;
    }
    void operator*=( const Vec3< T >& Source ) {
      X *= Source.X;
      Y *= Source.Y;
      Z *= Source.Z;
    }

    void operator/=( const T U ) {
      X /= U;
      Y /= U;
      Z /= U;
    }
    void operator/=( const Vec3< T >& Source ) {
      X /= Source.X;
      Y /= Source.Y;
      Z /= Source.Z;
    }

    void operator++() {
      X ++;
      Y ++;
      Z ++;
    }
    void operator--() {
      X --;
      Y --;
      Z --;
    }

    void assign( const T U ) {
      X = U;
      Y = U;
      Z = U;
    }
    void assign( const T Xi, const T Yi, const T Zi ) {
      X = Xi;
      Y = Yi;
      Z = Zi;
    }
    void assign( const Vec3< T >& Source ) {
      X = Source.X;
      Y = Source.Y;
      Z = Source.Z;
    }
    template< typename T2 >
    void assign( const Vec3< T2 >& Source ) {
      X = ( T ) Source.X;
      Y = ( T ) Source.Y;
      Z = ( T ) Source.Z;
    }

  public: //cast operators
    template< typename T2 >
    Vec3< T2 > cast() const {
      return Vec3< T2 >( ( T2 ) X, ( T2 ) Y, ( T2 ) Z );
    }
    template< typename T2 >
    Vec3< T2 > castFloor() const {
      return Vec3< T2 >( ( T2 ) floor( X ), ( T2 ) floor( Y ), ( T2 ) floor( Z ) );
    }
    template< typename T2 >
    Vec3< T2 > castCeil() const {
      return Vec3< T2 >( ( T2 ) ceil( X ), ( T2 ) ceil( Y ), ( T2 ) ceil( Z ) );
    }
    Vec3< T >& castBasic() {
      return *this;
    }

  public: //mathematic operators
    Vec3< T > operator+( const T U ) const {
      return Vec3< T >( X + U, Y + U, Z + U );
    }
    Vec3< T > operator+( const Vec3< T >& Source ) const {
      return Vec3< T >( X + Source.X, Y + Source.Y, Z + Source.Z );
    }

    Vec3< T > operator-( const T U ) const {
      return Vec3< T >( X - U, Y - U, Z - U );
    }
    Vec3< T > operator-( const Vec3< T >& Source ) const {
      return Vec3< T >( X - Source.X, Y - Source.Y, Z - Source.Z );
    }

    Vec3< T > operator*( const T U ) const {
      return Vec3< T >( X * U, Y * U, Z * U );
    }
    Vec3< T > operator*( const Vec3< T >& Source ) const {
      return Vec3< T >( X * Source.X, Y * Source.Y, Z * Source.Z );
    }

    Vec3< T > operator/( const T U ) const {
      return Vec3< T >( X / U, Y / U, Z / U );
    }
    Vec3< T > operator/( const Vec3< T >& Source ) const {
      return Vec3< T >( X / Source.X, Y / Source.Y, Z / Source.Z );
    }

  public: //comparsion operators
    bool operator==( const T U ) const {
      return X == U && Y == U && Z == U;
    }
    bool operator==( const Vec3< T >& Source ) const {
      return X == Source.X && Y == Source.Y && Z == Source.Z;
    }

    bool operator!=( const T U ) const {
      return X != U || Y != U || Z != U;
    }
    bool operator!=( const Vec3< T >& Source ) const {
      return X != Source.X || Y != Source.Y || Z != Source.Z;
    }

    bool oneGreater( const T U ) const {
      return X > U || Y > U || Z > U;
    }
    bool oneGreater( const Vec3< T >& Source ) const {
      return X > Source.X || Y > Source.Y || Z > Source.Z;
    }
    bool oneGreaterOrEqual( const T U ) const {
      return X >= U || Y >= U || Z >= U;
    }
    bool oneGreaterOrEqual( const Vec3< T >& Source ) const {
      return X >= Source.X || Y >= Source.Y || Z >= Source.Z;
    }

    bool oneLess( const T U ) const {
      return X < U || Y < U || Z < U;
    }
    bool oneLess( const Vec3< T >& Source ) const {
      return X < Source.X || Y < Source.Y || Z < Source.Z;
    }
    bool oneLessOrEqual( const T U ) const {
      return X <= U || Y <= U || Z <= U;
    }
    bool oneLessOrEqual( const Vec3< T >& Source ) const {
      return X <= Source.X || Y <= Source.Y || Z <= Source.Z;
    }

    bool everyGreater( const T U ) const {
      return X > U && Y > U && Z > U;
    }
    bool everyGreater( const Vec3< T >& Source ) const {
      return X > Source.X && Y > Source.Y && Z > Source.Z;
    }
    bool everyGreaterOrEqual( const T U ) const {
      return X >= U && Y >= U && Z >= U;
    }
    bool everyGreaterOrEqual( const Vec3< T >& Source ) const {
      return X >= Source.X && Y >= Source.Y && Z >= Source.Z;
    }

    bool everyLess( const T U ) const {
      return X < U && Y < U && Z < U;
    }
    bool everyLess( const Vec3< T >& Source ) const {
      return X < Source.X && Y < Source.Y && Z < Source.Z;
    }
    bool everyLessOrEqual( const T U ) const {
      return X <= U && Y <= U && Z <= U;
    }
    bool everyLessOrEqual( const Vec3< T >& Source ) const {
      return X <= Source.X && Y <= Source.Y && Z <= Source.Z;
    }

  /*public:
    virtual void streamWrite( Stream &Target, UInt8 Type ) const {
      Target.write< T >( X );
      Target.write< T >( Y );
      Target.write< T >( Z );
    }
    virtual void streamRead( Stream &Source, UInt8 Type ) {
      Source.read< T >( X );
      Source.read< T >( Y );
      Source.read< T >( Z );
    }*/

  public:
    T getLength() {
      return ( T ) sqrt( X * X + Y * Y + Z * Z );
    }

    Vec3< T >& normalize() {
      *this /= getLength();
      return *this;
    }

  };

  typedef Vec2< UInt8 > UInt8x2;
  typedef Vec2< UInt16 > UInt16x2;

  typedef Vec2< Int8 > Int8x2;
  typedef Vec2< Int16 > Int16x2;

  typedef Vec2< Float32 > Float32x2;
  typedef Vec2< Float64 > Float64x2;


  typedef Vec3< UInt16 > UInt16x3;

  typedef Vec3< Int16 > Int16x3;

  typedef Vec3< Float32 > Float32x3;
  typedef Vec3< Float64 > Float64x3;


  typedef Vec2< TypeScreen > Vec2Screen;
  typedef Vec2< TypeTex > Vec2Tex;

}