#include "GUI.CustomDrawComboBox.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"
#include "Engine.Rectangle.h"

#include "Resource.Font.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< CustomDrawComboBox >( "CustomDrawComboBox" );
    }

    void onButtonPress( Button &Btn ) {
      CustomDrawComboBox *C = Btn.Parent->getInterface( CustomDrawComboBox );
      ASSERT_RETURN( !C, "comboBoxButtonWrongParentElement" );

      C->cListBox.Visible = true;
      C->cListBox.Size.Y = mLimit( C->cListBox.ElementCount * C->cListBox.ElementSize, C->Skn->ComboBoxMinListBoxSize, C->Skn->ComboBoxMaxListBoxSize );
      C->cListBox.recalc();
    }

    // CustomDrawComboBox
    CustomDrawComboBox::CustomDrawComboBox() {

    }
    CustomDrawComboBox::~CustomDrawComboBox() {


    }

    void CustomDrawComboBox::init( cfx::Node &Source ) {
      Element::init( Source );

      cListBox.Parent = this;
      cListBox.Layer = Layer;
      cListBox.GRoot = GRoot;
      cListBox.init( cfx::Node() );

      cListBox.Visible = false;
      cListBox.Skn = Skn;
      cListBox.Pos = 0;
      cListBox.AllowSelection = true;
      cListBox.ElementSize = ( Source > "ElementSize" ).toInt();
      cListBox.ElementCount = ( Source > "ElementCount" ).toInt();
      if( Source >= "SelectedItem" )
        cListBox.SelectedItem = ( Source > "SelectedItem" ).toInt();

      cButton.Parent = this;
      cButton.Layer = Layer;
      cButton.GRoot = GRoot;
      cButton.init( cfx::Node() );

      cButton.Flags = halRight;
      cButton.Caption = "o";
      cButton.Skn = Skn;
      cButton.onClick = onButtonPress;
    }

    void CustomDrawComboBox::recalc() {
      Element::recalc(); 

      RealSize.Y = cListBox.ElementSize;

      cButton.Size = RealSize.Y;
      
      cListBox.Pos.Y = RealSize.Y - 1;
      cListBox.Size.X = RealSize.X;

      cListBox.recalc();
      cButton.recalc();
    }

    void CustomDrawComboBox::draw() {
      RealSize.Y = cListBox.ElementSize;

      Element::draw();

      Skn->drawPanel( AbsolutePos, RealSize, Skin::ptComboBox, State );

      cListBox.AbsolutePos = AbsolutePos + cListBox.RealPos;
      cButton.AbsolutePos = AbsolutePos + cButton.RealPos;

      if( cListBox.Visible ) {
        cListBox.draw();
        if( ( global_s.Cursor->PrimaryInteraction->isPress() && !Rec2Screen( cListBox.AbsolutePos, cListBox.RealSize ).contains( global_s.Cursor->Pos ) ) || cListBox.SelectedItem != cListBox.PrevSelectedItem )
          cListBox.Visible = false;

        if( cListBox.SelectedItem != prevSelItem )
          cListBox.Visible = false;
      }
      cButton.draw();
      
      if( cListBox.SelectedItem != ( SizeT ) -1 && cListBox.onItemDraw )
        cListBox.onItemDraw( cListBox, cListBox.SelectedItem, AbsolutePos, State );

      if( global_s.Cursor->ClickedObject == this && global_s.Cursor->PrevClickedObject != this ) {
        cListBox.SelectedItem = ( cListBox.SelectedItem < cListBox.ElementCount - 1 ) ? cListBox.SelectedItem + 1 : 0;
        cListBox.PrevSelectedItem = cListBox.SelectedItem;
        if( cListBox.onChange ) cListBox.onChange( cListBox );
      }
              
      prevSelItem = cListBox.SelectedItem;
    }

    interfaceRoot( CustomDrawComboBox,
      interfaceEntry( Element )
      );

  }

}

#endif