#include "GUI.CustomDrawListBox.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"

#include "Resource.Font.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< CustomDrawListBox >( "CustomDrawListBox" );
    }

    // CustomDrawListBox
    CustomDrawListBox::CustomDrawListBox() {
      onItemDraw = 0;
      onChange = 0;
    }
    CustomDrawListBox::~CustomDrawListBox() {


    }

    void CustomDrawListBox::init( cfx::Node &Source ) {
      Element::init( Source );

      ElementCount = ( Source > "ElementCount" ).toInt();
      ElementSize = ( Source > "ElementSize" ).toInt();

      Vertical = ( Source > "Vertical" ) != "false"; 
      AllowSelection = ( Source > "AllowSelection" ) == "true"; 
      if( Source >= "SelectedItem" )
        SelectedItem = ( Source > "SelectedItem" ).toInt();

      SelectedItem = -1;
      
      // ScrollBar
      cScrollBar.Parent = this;
      cScrollBar.GRoot = this->GRoot;
      cScrollBar.Layer = this->Layer;

      cScrollBar.init( cfx::Node() );
      cScrollBar.Skn = Skn;

      cScrollBar.ButtonSpeed = Skn->ListBoxButtonSpeed;
      cScrollBar.WheelSpeed = Skn->ListBoxWheelSpeed;

      cScrollBar.Min = 0;
      cScrollBar.Value = 0;
    }

    void CustomDrawListBox::recalc() {
      Element::recalc(); 

      AbsolutePos += 1;
      RealSize -= 2;

      cScrollBar.Vertical = Vertical;
      if( Vertical ) {
        cScrollBar.Flags = halRight | vszRelative;
        cScrollBar.Size.assign( 0, 100 );

        cScrollBar.BarSize = ( Float32 ) RealSize.Y;
        cScrollBar.Max = mMax< Float32 >( 0, ( Float32 )( ElementSize * ElementCount - RealSize.Y ) );

      } else {
        cScrollBar.Flags = valBottom | hszRelative;
        cScrollBar.Size.assign( 100, 0 );

        cScrollBar.BarSize = ( Float32 ) RealSize.X;
        cScrollBar.Max = mMax< Float32 >( 0, ( Float32 )( ElementSize * ElementCount - RealSize.X ) );

      }
      cScrollBar.recalc();

      AbsolutePos -= 1;
      RealSize += 2;
    }

    void CustomDrawListBox::draw() {
      Element::draw();

      Skn->drawPanel( AbsolutePos, RealSize, Skin::ptListBox, State );

      AbsolutePos += 1;
      RealSize -= 2;

      if( ElementSize < 1 )
        ElementSize = 1;

      // Content
      glPushScissor( AbsolutePos, RealSize );

      Vec2Screen Ps, Sz;

      Sz.assign( RealSize.X - Skn->ListBoxScrollBarSize, ElementSize );

      UInt16 Start = ( UInt16 ) cScrollBar.Value / ElementSize;
      UInt16 End = mMin< UInt16 >( ( ( UInt16 ) cScrollBar.Value + RealSize.Y ) / ElementSize + 1, ElementCount );
      ElementState_e  St;
      for( UInt16 a = Start; a < End; a ++ ) {
        Ps.assign( AbsolutePos.X, AbsolutePos.Y - mMax< TypeScreen >( ( TypeScreen ) cScrollBar.Value, 0 ) + ElementSize * a );
        St = virtualSubElementState( Ps, Vec2Screen( RealSize.X - cScrollBar.Size.X, ElementSize ) );

        if( AllowSelection ) {
          if( St == esClicked )
            SelectedItem = a;
          else if( SelectedItem == a )
            St = esClicked;
        }

        Skn->drawPanel( Ps, Sz, Skin::ptListBoxItem, St );

        if( onItemDraw )
          onItemDraw( *this, a, Ps, St );
      }
      UInt16 Offset = ( UInt16 ) cScrollBar.Value;

      glPopScissor();

      // ScrollBar
      bool B = cScrollBar.State == esHover || cScrollBar.State == esClicked;
      if( Vertical ) {
        if( B && cScrollBar.Size.X != Skn->ListBoxScrollBarSizeHover ) {
          cScrollBar.Size.X = Skn->ListBoxScrollBarSizeHover;
          cScrollBar.recalc();
        } else if( !B && cScrollBar.Size.X != Skn->ListBoxScrollBarSize ) {
          cScrollBar.Size.X = Skn->ListBoxScrollBarSize;
          cScrollBar.recalc();
        }

      } else  {
        if( B && cScrollBar.Size.Y != Skn->ListBoxScrollBarSizeHover ) {
          cScrollBar.Size.Y = Skn->ListBoxScrollBarSizeHover;
          cScrollBar.recalc();
        } else if( !B && cScrollBar.Size.Y != Skn->ListBoxScrollBarSize ) {
          cScrollBar.Size.Y = Skn->ListBoxScrollBarSize;
          cScrollBar.recalc();
        }

      }

      cScrollBar.AbsolutePos = AbsolutePos + cScrollBar.RealPos;

      if( onChange && SelectedItem != PrevSelectedItem )
        onChange( *this );
      PrevSelectedItem = SelectedItem;

      if( State == esHover )
        cScrollBar.doWheelControl();

      cScrollBar.draw();

      AbsolutePos -= 1;
      RealSize += 2;
    }

    interfaceRoot( CustomDrawListBox,
      interfaceEntry( Element )
      );

  }

}

#endif