#include "GUI.ListBox.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"

#include "Resource.Font.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< ListBox >( "ListBox" );
    }

    // ListBox
    void onListBoxItemDraw( CustomDrawListBox &IListBox, UInt16 Item, Vec2Screen Pos, ElementState_e State ) {
      ListBox &LB = *IListBox.getInterface( ListBox );
      String &It = *LB.Items[ Item ];

      LB.Skn->drawLabel( Vec2Screen( Pos.X + 5, Pos.Y + LB.ElementSize / 2 ), It, res::Font::valCenter );
    }

    ListBox::ListBox() {
      onItemDraw = onListBoxItemDraw;
    }
    ListBox::~ListBox() {
      foreach_a( Items )
        delete Items[ a ];
    }

    void ListBox::addItem( const IString& Name ) {
      Items.pushItem( new String( Name ) );
    }

    void ListBox::removeItem( const IString& Name ) {
      foreach_a( Items ) {
        if( *Items[ a ] == Name ) {
          delete Items[ a ];
          Items.removeItemPos( a );
          return;
        }
      }
    }
    void ListBox::removeItem( SizeT Pos ) {
      delete Items[ Pos ];
      Items.removeItemPos( Pos );
    }

    void ListBox::init( cfx::Node &Source ) {
      CustomDrawListBox::init( Source );

      AllowSelection = ( Source > "AllowSelection" ) != "false"; 

      if( ElementSize == 0 )
        ElementSize = Skn->ListBoxItemSize;

      cfx::Node &S = Source >> "Items";
      foreach_a( S.Nodes ) {
        Items.pushItem( new String( S.Nodes[ a ]->Name ) );
      }
    }

    void ListBox::recalc() {
      CustomDrawListBox::recalc();
    }

    void ListBox::draw() {
      ElementCount = Items.Size;

      CustomDrawListBox::draw();
    }

    interfaceRoot( ListBox,
      interfaceEntry( Element )
      interfaceEntry( CustomDrawListBox )
      );

  }

}

#endif