#include "GUI.ScrollBar.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"
#include "Engine.Application.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< ScrollBar >( "ScrollBar" );
    }

    // ScrollBar
    ScrollBar::ScrollBar() {
    
    }
    ScrollBar::~ScrollBar() {


    }

    void ScrollBar::init( cfx::Node &Source ) {
      Element::init( Source );
      IValMinMax::init( Source );

      BarSize = ( Source > "BarSize" ).toFloat();
      ButtonSpeed = ( Source > "ButtonSpeed" ).toFloat();
      WheelSpeed = ( Source > "WheelSpeed" ).toFloat();

      Vertical = ( Source > "Vertical" ) != "false";
    }

    void ScrollBar::recalc() {
      Element::recalc(); 
    }

    void ScrollBar::draw() {
      Element::draw();

      Skn->drawPanel( AbsolutePos, RealSize, Skin::ptScrollBarBack, State );

      if( Max < Min )
        Max = Min;

      Value = mLimit( Value, Min, Max );

      if( Vertical )
        verticalDraw();
      else
        horizontalDraw();
    }
    void ScrollBar::verticalDraw() {
      Vec2Screen Ps, Sz;
      ElementState_e St;
      Float32 SlideSpace, SliderSize;

      // Top button
      Ps = AbsolutePos;
      Sz = Vec2Screen( RealSize.X );
      St = virtualSubElementState( Ps, Sz );
      Skn->drawPanel( Ps, Sz, Skin::ptScrollBarButton, St );

      if( St == esClicked )
        Value = mLimit< Float32 >( Value - ButtonSpeed  / ( Float32 ) global_s.Application->Performance.getFPS(), Min, Max );

      // Bottom button
      Ps = Vec2Screen( AbsolutePos.X, AbsolutePos.Y + RealSize.Y - RealSize.X );
      Sz = Vec2Screen( RealSize.X );
      St = virtualSubElementState( Ps, Sz );
      Skn->drawPanel( Ps, Sz, Skin::ptScrollBarButton, St );

      if( St == esClicked )
        Value = mLimit< Float32 >( Value + ButtonSpeed / ( Float32 ) global_s.Application->Performance.getFPS(), Min, Max );

      // Bar
      if( Min == Max ) {
        SlideSpace = 0;
        Ps = Vec2Screen( AbsolutePos.X, AbsolutePos.Y + RealSize.X );
      } else {
        SlideSpace = ( RealSize.Y - RealSize.X * 2 ) / ( BarSize + ( Max - Min ) ) * ( Max - Min );
        Ps = Vec2Screen( AbsolutePos.X, AbsolutePos.Y + RealSize.X + ( TypeScreen )( ( Value - Min ) / ( Max - Min ) * SlideSpace ) );
      }

      SliderSize = RealSize.Y - RealSize.X * 2 - SlideSpace;
      Sz = Vec2Screen( RealSize.X, ( TypeScreen ) SliderSize );
      St = virtualSubElementState( Ps, Sz );

      if( Dragged && Min != Max ) {
        Value = mLimit< Float32 >( Value + ( Float32 ) global_s.Cursor->RelPos.Y / SlideSpace * ( Max - Min ), Min, Max );
        St = esClicked;
      }

      Skn->drawPanel( Ps, Sz, Skin::ptScrollBarBar, St, Vertical );
      
      if( State != esClicked )
        Dragged = false;
      else if( St == esClicked )
        Dragged = true;

      // Wheel control
      if( State == esHover )
        doWheelControl();
    }
    void ScrollBar::horizontalDraw() {
      Vec2Screen Ps, Sz;
      ElementState_e St;
      Float32 SlideSpace, SliderSize;

      // Top button
      Ps = AbsolutePos;
      Sz = Vec2Screen( RealSize.Y );
      St = virtualSubElementState( Ps, Sz );
      Skn->drawPanel( Ps, Sz, Skin::ptScrollBarButton, St );

      if( St == esClicked )
        Value = mLimit< Float32 >( Value - ButtonSpeed  / ( Float32 ) global_s.Application->Performance.getFPS(), Min, Max );

      // Bottom button
      Ps = Vec2Screen( AbsolutePos.X + RealSize.X - RealSize.Y, AbsolutePos.Y );
      Sz = Vec2Screen( RealSize.Y );
      St = virtualSubElementState( Ps, Sz );
      Skn->drawPanel( Ps, Sz, Skin::ptScrollBarButton, St );

      if( St == esClicked )
        Value = mLimit< Float32 >( Value + ButtonSpeed / ( Float32 ) global_s.Application->Performance.getFPS(), Min, Max );

      // Bar
      if( Min == Max ) {
        SlideSpace = 0;
        Ps = Vec2Screen( AbsolutePos.X + RealSize.Y, AbsolutePos.Y );
      } else {
        SlideSpace = ( RealSize.X - RealSize.Y * 2 ) / ( BarSize + ( Max - Min ) ) * ( Max - Min );
        Ps = Vec2Screen( AbsolutePos.X + RealSize.Y + ( TypeScreen )( ( Value - Min ) / ( Max - Min ) * SlideSpace ), AbsolutePos.Y );
      }
      
      SliderSize = RealSize.X - RealSize.Y * 2 - SlideSpace;
      Sz = Vec2Screen( ( TypeScreen ) SliderSize, RealSize.Y );

      St = virtualSubElementState( Ps, Sz );

      if( Dragged && Min != Max ) {
        Value = mLimit< Float32 >( Value + ( Float32 ) global_s.Cursor->RelPos.X / SlideSpace * ( Max - Min ), Min, Max );
        St = esClicked;
      }

      Skn->drawPanel( Ps, Sz, Skin::ptScrollBarBar, St );

      if( State != esClicked )
        Dragged = false;
      else if( St == esClicked )
        Dragged = true;

      // Wheel control
      if( State == esHover )
        doWheelControl();
    }

    void ScrollBar::doWheelControl() {
      if( global_s.Cursor->ScrollUp->isPress() )
        Value = mLimit< Float32 >( Value - WheelSpeed, Min, Max );

      if( global_s.Cursor->ScrollDown->isPress() )
        Value = mLimit< Float32 >( Value + WheelSpeed, Min, Max );
    }

    interfaceRoot( ScrollBar,
      interfaceEntry( IValMinMax )
      interfaceEntry( Element )
      );

  }

}

#endif