#include "GUI.Slider.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"
#include "Engine.Application.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< Slider >( "Slider" );
    }

    // Slider
    Slider::Slider() {
      onChange = 0;
    }
    Slider::~Slider() {


    }

    void Slider::init( cfx::Node &Source ) {
      Element::init( Source );
      IValMinMax::init( Source );

      Vertical = ( Source > "Vertical" ) == "true";
      PrevValue = Value - 1;
    }

    void Slider::recalc() {
      Element::recalc(); 
      if( Vertical )
        RealSize.X = Skn->SliderSliderSize.Y;
      else
        RealSize.Y = Skn->SliderSliderSize.Y;
    }

    void Slider::draw() {
      Element::draw();

      if( Max < Min )
        Max = Min;

      Value = mLimit( Value, Min, Max );

      if( onChange && Value != PrevValue ) {
        PrevValue = Value;
        onChange( *this );
      }

      if( Vertical )
        verticalDraw();
      else
        horizontalDraw();
    }
    void Slider::verticalDraw() {
      Vec2Screen Ps;

      Skn->drawPanel( Vec2Screen( AbsolutePos.X + ( Size.X  - Skn->SliderBarSize ) / 2, AbsolutePos.Y ), Vec2Screen( Skn->SliderBarSize, RealSize.Y ), Skin::ptSliderBar, State, Vertical );

      // Bar
      if( Min == Max ) {
        Ps = Vec2Screen( AbsolutePos.X + ( RealSize.X - Skn->SliderSliderSize.Y ) / 2, AbsolutePos.Y );
      } else {
        Ps = Vec2Screen( AbsolutePos.X + ( RealSize.X - Skn->SliderSliderSize.Y ) / 2, AbsolutePos.Y + ( TypeScreen )( ( Value - Min ) / ( Max - Min ) * Size.Y ) );
      }

      Skn->drawPanel( Vec2Screen( Ps.X, Ps.Y - Skn->SliderSliderSize.X / 2 ), Skn->SliderSliderSize.yx(), Skin::ptSliderSlider, State, Vertical );

      if( State == esClicked )
        Value = ( Float32 )( global_s.Cursor->Pos.Y - AbsolutePos.Y ) / ( Float32 ) RealSize.Y * ( Max - Min );
    }
    void Slider::horizontalDraw() {
      Vec2Screen Ps;

      Skn->drawPanel( Vec2Screen( AbsolutePos.X, AbsolutePos.Y + ( RealSize.Y  - Skn->SliderBarSize ) / 2 ), Vec2Screen( RealSize.X, Skn->SliderBarSize ), Skin::ptSliderBar, State, Vertical );

      // Bar
      if( Min == Max ) {
        Ps = Vec2Screen( AbsolutePos.X, AbsolutePos.Y + ( RealSize.Y - Skn->SliderSliderSize.Y ) / 2 );
      } else {
        Ps = Vec2Screen( AbsolutePos.X + ( TypeScreen )( ( Value - Min ) / ( Max - Min ) * RealSize.X ), AbsolutePos.Y + ( RealSize.Y - Skn->SliderSliderSize.Y ) / 2 );
      }

      Skn->drawPanel( Vec2Screen( Ps.X - Skn->SliderSliderSize.X / 2, Ps.Y ), Skn->SliderSliderSize, Skin::ptSliderSlider, State, Vertical );

      if( State == esClicked )
        Value = ( Float32 )( global_s.Cursor->Pos.X - AbsolutePos.X ) / ( Float32 ) RealSize.X * ( Max - Min );
    }

    interfaceRoot( Slider,
      interfaceEntry( IValMinMax )
      interfaceEntry( Element )
      );

  }

}

#endif