#include "GUI.Window.h"

#include "Engine.Global.h"
#include "Engine.Display.h"
#include "Engine.Hooks.h"
#include "Engine.GUI.Skin.h"

#include "Resource.Font.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace gui {

    hook( hookRegisterGUIElements ) {
      global_s.gui.ElementList.add< Window >( "Window" );
    }

    // Window
    Window::Window() {

    }
    Window::~Window() {
      
    }

    void Window::init( cfx::Node &Source ) {
      Element::init( Source );
      ICaption::init( Source );
      ILayer::init( Source );
      IContent::init( Source );

      Resizable = ( Source > "Resizable" ) == "true";

      cfx::Node &n = Source >> "MaxSize"; 
      MaxSize.X = n.Value.toInt();
      MaxSize.Y = ( n || ( SizeT ) 0 ).Name.toInt();

      cfx::Node &n2 = Source >> "MinSize"; 
      MinSize.X = n2.Value.toInt();
      MinSize.Y = ( n2 || ( SizeT ) 0 ).Name.toInt();

      onDestroy = 0;
      onDraw = 0;

      recalc();
    }
    bool Window::destroy() {
      if( IDestructible::destroy() )
        return true;

      if( onDestroy )
        onDestroy( *this );

      return false;
    }

    void Window::recalc() {
      Element::recalc(); 
      IContent::recalc();
    }

    void Window::draw() {
      Element::draw();
      ILayer::draw();

      Skn->drawPanel( AbsolutePos, RealSize, Skin::ptWindow, State );

      // Header
      Vec2Screen HeaderPos( AbsolutePos.X, AbsolutePos.Y - Skn->WindowHeaderSize ), HeaderSize( RealSize.X, Skn->WindowHeaderSize + 1 );
      Vec2Screen ButtonPos( Skn->WindowHeaderButtonPos ), ButtonSize( Skn->WindowHeaderButtonSize );
      TypeScreen CaptionOffset = Skn->WindowHeaderCaptionOffset;

      if( global_s.Cursor->Pos.everyGreaterOrEqual( HeaderPos ) && global_s.Cursor->Pos.everyLess( HeaderPos + HeaderSize ) )
        global_s.Cursor->ObjectOnCursorStack.pushItem( this );

      if( ButtonPos.X < 0 )
        ButtonPos.X += HeaderSize.X - ButtonSize.X;
      ButtonPos += HeaderPos;


      ElementState_e BSt = virtualSubElementState( ButtonPos, ButtonSize );
      ElementState_e HSt =  virtualSubElementState( HeaderPos, HeaderSize );
      if( BSt == esClicked || BSt == esHover ) {
        if( GRoot->FutureLayers.Size && GRoot->FutureLayers[ GRoot->FutureLayers.Size - 1 ] == this )
          HSt = esFocused;
        else
          HSt = esNormal;
      } else if( HSt == esNormal && GRoot->FutureLayers.Size && GRoot->FutureLayers[ GRoot->FutureLayers.Size - 1 ] == this )
        HSt = esFocused;

      Skn->drawPanel( HeaderPos, HeaderSize, Skin::ptWindowHeader, HSt );
      Skn->drawPanel( ButtonPos, ButtonSize, Skin::ptWindowCloseButton, BSt );
      if( Skn->WindowHeaderButtonCaption.Size )
        Skn->drawLabel( ButtonPos + ButtonSize / 2, Skn->WindowHeaderButtonCaption, res::Font::halCenter | res::Font::valCenter );

      if( CaptionOffset > 0 )
        Skn->drawLabel( Vec2Screen( HeaderPos.X + CaptionOffset, HeaderPos.Y + HeaderSize.Y / 2 ), Caption, res::Font::valCenter );
      else if( Skn->WindowHeaderCaptionOffset == 0 )
        Skn->drawLabel( Vec2Screen( HeaderPos.X + HeaderSize.X / 2, HeaderPos.Y + HeaderSize.Y / 2 ), Caption, res::Font::valCenter | res::Font::halCenter );
      else
        Skn->drawLabel( Vec2Screen( HeaderPos.X + HeaderSize.X + CaptionOffset, HeaderPos.Y + HeaderSize.Y / 2 ), Caption, res::Font::valCenter | res::Font::halRight );

      if( HSt == esClicked ) {
        Pos += global_s.Cursor->RelPos;
        global_s.Cursor->PrimaryInteractionPos += global_s.Cursor->RelPos;
      }

      IContent::draw();

      if( onDraw )
        onDraw( *this );

      if( Resizable ) {
        ButtonPos = AbsolutePos + RealSize - 10;
        ButtonSize = 10;

        if( global_s.Cursor->Pos.everyGreaterOrEqual( ButtonPos ) && global_s.Cursor->Pos.everyLess( ButtonPos + ButtonSize ) )
          global_s.Cursor->ObjectOnCursorStack.pushItem( this );

        ElementState_e B2St = virtualSubElementState( ButtonPos, ButtonSize );

        Skn->drawPanel( ButtonPos, ButtonSize, Skin::ptButton, B2St );

        if( B2St == esClicked ) {
          if( clickPoint == RealSize )
            clickPoint = RealSize - ( global_s.Cursor->Pos - AbsolutePos );

          Size = global_s.Cursor->Pos - AbsolutePos + clickPoint;
          Size.X = mLimit( Size.X, MinSize.X, MaxSize.X ); 
          Size.Y = mLimit( Size.Y, MinSize.Y, MaxSize.Y ); 

          global_s.Cursor->PrimaryInteractionPos = AbsolutePos + Size - clickPoint;
          recalc();
        } else clickPoint = RealSize;
      }

      if( HSt != esClicked && BSt == esClicked )
        destroy();
      //Skn->drawLabel( AbsolutePos, Caption, TextAlign );
    }

    interfaceRoot( Window,
      interfaceEntry( ICaption )
      interfaceEntry( Element )
      );

  }

}

#endif