#include "Resource.Font.h"
#include "Toolkit.Display.h"

#ifdef MODULE_CLIENT 

namespace app {

  namespace res {

    unsigned short Font::DefaultFontLoad;

    hook( hookPostSDL_OPENGLInit ) {
      linkEngineVariable( "Font" > "DefaultFontLoad",   Font::DefaultFontLoad,  unsigned short );
    }

    // Font
    String Font::unique( const String &Filename, FSizeT Size ) {
      return String().pushCChar( "Font|filename=" ).pushString( Filename ).pushCChar( "|size=" ).pushInt( Size );
    }
    Font* Font::get( const String &Filename, FSizeT Size ) {
      Resource *R = global_s.Resources->find( unique( Filename, Size ) );

      if ( !R )
        return new Font( Filename, Size );
      else
        return ( Font* ) R;
    }

    Font::Font( const String &Filename, FSizeT Size ) : Resource( passError, unique( Filename, Size ) ), Filename( String( "../res/font/" ).pushString( Filename ) ), Size( Size ) {
      Tex_SDL = 0;
      Tex_OGL = -1;
      Fnt_TTF = 0;
      Line = 0;
      LoadPos = 0;
      LineH = 0;

      Lines.pushItem( 0 );

      TexSize = 128;
      for( SizeT a = 0; a < DefaultFontLoad; a ++ )
        ToLoad.pushItem( a );

      progress( pThreaded );
    }
    Font::~Font() {

    }

    Vec2Screen Font::getStringSize( const IString &Text ) {
      if( !Text.Size )
        return Vec2Screen();

      Vec2Screen Sz( 0, LineH );

      foreach_a( Text ) {

        if ( Glyphs.Size <= Text[ a ] )
          continue;

        Sz.X += Glyphs[ Text[ a ] ].W;
      }

      return Sz;
    }

    void Font::draw( Vec2Screen Pos, const IString &Text, Color Col /* = Color( 255, 255, 255 ) */, UInt8 Flags /* = halLeft | valTop */ ) {
      if( !Text.Size || Pass != passDone )
        return;

      unsigned short C;
      Glyph *G;
      SizeT Num = 0;

      if( Flags & alAny ) {
        Vec2Screen Sz = getStringSize( Text );
        if( Flags & halCenter )
          Pos.X -= Sz.X / 2;
        else if( Flags & halRight )
          Pos.X -= Sz.X;

        if( Flags & valCenter )
          Pos.Y -= Sz.Y / 2;
        else if( Flags & valBottom )
          Pos.Y -= Sz.Y;

      }

      if( Flags & efShadow )
        draw( Pos + 1, Text, Color( 0, 0, 0, 128 ) );

      glTextureDrawMode();
      glTexture( Tex_OGL );

      foreach_a( Text ) {
        C = Text[ a ];

        if ( Glyphs.Size <= C )
          continue;

        G = &Glyphs[ C ];

        if ( G->W < 2 )
          continue;

        glVertexBuffer.pushRect( Pos + G->Pos, G->Size );
        glTexCoordBuffer.pushRect( G->TPos / ( float ) TexSize, G->TSize / ( float ) TexSize );
        glColorBuffer.pushColorRect( Col ); 

        Pos.X += G->W;
        Num ++;
      }
    }

    void Font::onLoad() {
      switch( Pass ) {

      case passLoad:
        if ( pLoad() )
          Pass = passError;
        break;

      case passBlit:
        if ( pBlit() )
          Pass = passError;
        break;

      case passRender:
        if ( pRender() )
          Pass = passError;
        break;

      case passError:
        break;

      default:
        ASSERT( true, "wrongPass" );
        break;

      }
    }
    void Font::onUnload() {
      if( Tex_SDL ) {
        SDL_FreeSurface( Tex_SDL );
        Tex_SDL = 0;
      }

      if( Tex_OGL != -1 ) {
        glDeleteTextures( 1, &Tex_OGL );
        Tex_OGL = -1;
      }

      if ( Fnt_TTF ) {
        TTF_CloseFont( Fnt_TTF );
        Fnt_TTF = 0;
      }
    }

    bool Font::pLoad() {      
      Fnt_TTF = TTF_OpenFont( Filename.getCChar(), Size );

      ASSERT_RETURNVAL( !Fnt_TTF, true, "FontLoadFail", "filename", &Filename, "error", &IString( IMG_GetError() ) );

      Tex_SDL = SDL_CreateRGBSurface( SDL_SWSURFACE, TexSize, TexSize, SDL_COLORDEPTH, SDL_RMASK, SDL_GMASK, SDL_BMASK, SDL_AMASK );
      SDL_FillRect( Tex_SDL, 0, SDL_MapRGBA( Tex_SDL->format, 0, 0, 0, 0 ) );

      ASSERT_RETURNVAL( !Tex_SDL, true, "FontSDLTextureCreationFail", "error", &IString( SDL_GetError() ) );

      LineH = TTF_FontLineSkip( Fnt_TTF );
      Ascent = TTF_FontAscent( Fnt_TTF );

      progress( pThreaded );
      return false;
    }
    bool Font::pBlit() {
      unsigned short C;

      SDL_Surface *S;
      SDL_Rect R;
      SDL_Color Cl = { 255, 255, 255, 0 };

      TypeScreen Y = Line * LineH, W;
      SizeT PLine;
      int GmX, GMX, GmY, GMY, GW;

      while ( true ) {
        
        if( LoadPos == ToLoad.Size )
          break;

        C = ToLoad[ LoadPos ];
        allocate( C );

        if( !TTF_GlyphIsProvided( Fnt_TTF, C ) ) {
          LoadPos ++;
          Glyphs[ C ].W = 1;
          continue;
        }

        TTF_GlyphMetrics( Fnt_TTF, C, &GmX, &GMX, &GmY, &GMY, &GW );
        W = GMX - GmX;

        PLine = Line;
        while ( Lines[ Line ] + W >= TexSize ) {
          Line ++;
          if ( Line == Lines.Size )
            Line = 0;
          
          if ( Line == PLine ) {
            Lines.pushItem( 0 );
            Line = Lines.Size - 1;

            if ( ( Lines.Size ) * LineH >= TexSize ) {
              if ( resizeSurface() )
                return true;
            }
          }
        }
        Y = Line * LineH;

        Glyphs[ C ].TPos.assign( ( float ) Lines[ Line ], ( float ) Y );
        Glyphs[ C ].TSize.assign( ( float ) W, ( float )( GMY - GmY ) );

        Glyphs[ C ].Pos.assign( GmX, Ascent - GMY );
        Glyphs[ C ].Size.assign( W, GMY - GmY );

        Glyphs[ C ].W = GW;

        R.x = Lines[ Line ];
        R.y = Y;

        Lines[ Line ] += W;

        S = TTF_RenderGlyph_Blended( Fnt_TTF, C, Cl );
        SDL_SetAlpha( S, 0, 0 );
        SDL_BlitSurface( S, 0, Tex_SDL, &R );
        SDL_FreeSurface( S );

        LoadPos ++;

      }

      if( LoadPos >= ToLoad.Size ) {
        ToLoad.Size = 0;
        LoadPos = 0;
      }

      progress( pMain );
      return false;
    }
    bool Font::pRender() {
      UInt8 Channels = Tex_SDL->format->BytesPerPixel;
      UInt32 Format;
      
      //ASSERT_RETURNVAL( Tex_SDL->format->Rmask != 0x000000ff, true, "surfaceWrongFormat" );

      if ( Channels == 4 ) 
        Format = GL_RGBA;
      else if ( Channels == 3 )
        Format = GL_RGB;
      else
        ASSERT_RETURNVAL( true, true, "surfaceNoTrueColor" );

      glGenTextures( 1, &Tex_OGL );
      ASSERT_RETURNVAL( !Tex_OGL, true, "FontCreateFail" );

      glTexture( Tex_OGL );

      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );

      glTexImage2D( GL_TEXTURE_2D, 0, Channels, Tex_SDL->w, Tex_SDL->h, 0, Format, GL_UNSIGNED_BYTE, Tex_SDL->pixels );

      progress( pEnd );
      return false;
    }

    void Font::allocate( unsigned short Num ) {
      while( Glyphs.Size <= Num ) {
        Glyphs.pushItem( Glyph() );
      }

    }
    bool Font::resizeSurface() {
      TexSize *= 2;
      ASSERT_RETURNVAL( TexSize > global_s.Display->MaxTextureSize, true, "fontOuttaMemory", "filename", &Filename );

      SDL_Surface *TS = SDL_CreateRGBSurface( SDL_SWSURFACE, TexSize, TexSize, SDL_COLORDEPTH, SDL_RMASK, SDL_GMASK, SDL_BMASK, SDL_AMASK );
      ASSERT_RETURNVAL( !Tex_SDL, true, "FontSDLTextureCreationFail", "error", &IString( SDL_GetError() ) );

      SDL_SetAlpha( Tex_SDL, 0, 0 );
      SDL_BlitSurface( Tex_SDL, 0, TS, 0 );
      SDL_FreeSurface( Tex_SDL );

      Tex_SDL = TS;
      Line = 0;

      return false;
    }

  }

}

#endif