#pragma once

#include "Engine.Dedicated.h"
#include "Engine.Resources.h"
#include "Engine.Vector.h"
#include "Engine.Display.h"
#include "Engine.OpenGL.h"
#include "Engine.Color.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace res {

    class Font : public Resource {

      friend class ResourceManager;

    public:
      struct Glyph {
        Vec2Tex TPos, TSize;
        Vec2Screen Pos, Size;
        TypeScreen W;
      };

      typedef UInt16 FSizeT;
      enum ePass {
        passDone,
        passRender,
        passBlit,
        passLoad,
        passError
      };

      enum eFlags : UInt8 {
        valTop    = 0,
        valCenter = 1 << 0,
        valBottom = 1 << 1,

        halLeft   = 0,
        halCenter = 1 << 2,
        halRight  = 1 << 3,

        alAny = valTop | valCenter | valBottom | halLeft | halCenter | halRight,

        efShadow  = 1 << 4,
      };

      static unsigned short DefaultFontLoad;

    public:
      String Filename;
      UInt16 Size, TexSize, LineH, Ascent;
      SizeT LoadPos, Line;

      Container< Glyph > Glyphs;
      Container< UInt16 > Lines;
      Container< unsigned short > ToLoad;

      TTF_Font *Fnt_TTF;
      SDL_Surface *Tex_SDL;
      GLuint Tex_OGL;

    public:
      static Font* get( const String &Filename, FSizeT Size );
      static String unique( const String &Filename, FSizeT Size );

      Font( const String &Filename, FSizeT Size );
      virtual ~Font();

    public:
      Vec2Screen getStringSize( const IString &Text );

      void draw( Vec2Screen Pos, const IString &Text, Color Col = Color( 255, 255, 255 ), UInt8 Flags = halLeft | valTop );

    public:
      virtual void onLoad();
      virtual void onUnload();

    private:
      bool pRender();
      bool pBlit();
      bool pLoad();

      void allocate( unsigned short Num );
      bool resizeSurface();

    };

  }

}

#endif