#pragma once

#include "Engine.Dedicated.h"
#include "Engine.Resources.h"
#include "Engine.OpenGL.h"
#include "Engine.Display.h"
#include "Engine.Vector.h"

#ifdef MODULE_CLIENT

namespace app {

  namespace res {

    class Texture : public Resource {

      friend class ResourceManager;

    public:
      enum ePass {
        passDone,
        passRender,
        passLoad,
        passError,

        passCreate = passLoad,
      };
      enum eCreationType {
        ctLoadFromFile,
        ctJustCreate
      };

    public:
      eCreationType CreationType;
      String Filename;
      Vec2Screen Size;
      SDL_Surface *Tex_SDL;
      GLuint Tex_OGL;
      bool LinMap;

    public:
      static Texture* get( const String &Filename, bool LinearMap = false );
      static String unique( const String &Filename, bool LinearMap );

      static void saveRaw( const String& Filename, Vec2Screen Pos, Vec2Screen Size );

      Texture( const String &Filename, bool LinearMap );
      Texture( Vec2Screen Size, bool LinearMap = false );

      virtual ~Texture();

    public:
      void draw( const Vec2Screen Pos );
      void draw( const Vec2Screen Pos, const Vec2Screen Size );

      void drawTiled( Vec2Screen Pos, Vec2Screen Size, Vec2Screen Tiles );

      void drawPart( Vec2Screen Pos, Vec2Screen Size, Vec2Tex TPos, Vec2Tex TSize );
      void drawPart( Vec2Screen Pos, Vec2Screen Size, Vec2Screen TPos, Vec2Screen TSize );

      void drawPartRotated( Vec2Screen Pos, Vec2Screen Size, Vec2Screen TPos, Vec2Screen TSize, Vec2Screen Center, Float64 Angle );

    private:
      bool pLoad();
      bool pRender();
      bool pCreate();

    public:
      virtual void onLoad();
      virtual void onUnload();

    };

  }

}

#endif